/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004, 2005  by Michael Buesch                     *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "libpwmanager/filecompress.h"
#include "rawfile/rawfile.h"

#include <stdlib.h>
#include <zlib.h>
#define BZ_NO_STDIO
#include <bzlib.h>


#define GZIP_COMPR_DESTLEN(sourceLen)	(sourceLen + (sourceLen / 100 / 10 + 12))
#define GZIP_DECOM_DESTLEN(sourceLen)	(sourceLen * 2)
#define BZIP2_COMPR_DESTLEN(sourceLen)	(sourceLen + (sourceLen / 100 + 600))
#define BZIP2_DECOM_DESTLEN(sourceLen)	(sourceLen * 2)

#define GZIP_COMPRESSION_LEVEL			Z_BEST_COMPRESSION
#define EXPAND_COMPRESS_DESTBUF_BYTES		1024
#define EXPAND_DECOMPRESS_DESTBUF_BYTES		(1024 * 10)
#define BZ_BLOCK_SIZE				9
#define BZ_WORK_FACTOR				30


bool FileCompress::compress(QByteArray *data,
			    unsigned char algo)
{
	switch (algo) {
	case PWM_COMPRESS_NONE:
		return true;
	case PWM_COMPRESS_GZIP:
		return compressGzip(data);
	case PWM_COMPRESS_BZIP2:
		return compressBzip2(data);
	default:
		return false;
	}
	return false;
}

bool FileCompress::decompress(QByteArray *data,
			      unsigned char algo)
{
	switch (algo) {
	case PWM_COMPRESS_NONE:
		return true;
	case PWM_COMPRESS_GZIP:
		return decompressGzip(data);
	case PWM_COMPRESS_BZIP2:
		return decompressBzip2(data);
	default:
		return false;
	}
	return false;
}

bool FileCompress::compressGzip(QByteArray *data)
{
	int ret;
	Bytef *dest;
	const Bytef *source;
	unsigned long sourceLen, destLen;

	source = reinterpret_cast<const Bytef *>(data->data());
	sourceLen = data->size();
	destLen = GZIP_COMPR_DESTLEN(sourceLen);
	dest = static_cast<Bytef *>(malloc(destLen));
	if (!dest)
		return false;
	while (1) {
		ret = compress2(dest,
				static_cast<uLongf *>(&destLen),
				source,
				static_cast<uLong>(sourceLen),
				GZIP_COMPRESSION_LEVEL);
		switch (ret) {
		case Z_OK:
			goto out;
		case Z_BUF_ERROR:
			/* we don't use realloc(), because it may
			 * (in this case) unneccessarily copy the old block
			 * to the new allocated block.
			 */
			free(dest);
			destLen += EXPAND_COMPRESS_DESTBUF_BYTES;
			dest = static_cast<Bytef *>(malloc(destLen));
			if (!dest)
				return false;
			break;
		default:
			free(dest);
			return false;
		}
	}
out:
	data->duplicate(reinterpret_cast<const char *>(dest),
			destLen);
	free(dest);
	return true;
}

bool FileCompress::compressBzip2(QByteArray *data)
{
	int ret;
	char *dest, *source;
	unsigned int sourceLen, destLen;

	sourceLen = data->size();
	destLen = BZIP2_COMPR_DESTLEN(sourceLen);
	source = static_cast<char *>(malloc(sourceLen));
	if (!source)
		return false;
	memcpy(source, data->data(), sourceLen);
	dest = static_cast<char *>(malloc(destLen));
	if (!dest) {
		free(source);
		return false;
	}
	while (1) {
		ret = BZ2_bzBuffToBuffCompress(dest,
					       &destLen,
					       source,
					       sourceLen,
					       BZ_BLOCK_SIZE,
					       0,
					       BZ_WORK_FACTOR);
		switch (ret) {
		case BZ_OK:
			goto out;
		case BZ_OUTBUFF_FULL:
			/* we don't use realloc(), because it may
			 * (in this case) unneccessarily copy the old block
			 * to the new allocated block.
			 */
			free(dest);
			destLen += EXPAND_COMPRESS_DESTBUF_BYTES;
			dest = static_cast<char *>(malloc(destLen));
			if (!dest) {
				free(source);
				return false;
			}
			break;
		default:
			free(source),
			free(dest);
			return false;
		}
	}
out:
	free(source);
	data->duplicate(dest, destLen);
	free(dest);
	return true;
}

bool FileCompress::decompressGzip(QByteArray *data)
{
	int ret;
	Bytef *dest;
	const Bytef *source;
	unsigned long sourceLen, destLen;

	source = reinterpret_cast<const Bytef *>(data->data());
	sourceLen = data->size();
	destLen = GZIP_DECOM_DESTLEN(sourceLen);
	dest = static_cast<Bytef *>(malloc(destLen));
	if (!dest)
		return false;
	while (1) {
		ret = uncompress(dest,
				 static_cast<uLongf *>(&destLen),
				 source,
				 static_cast<uLong>(sourceLen));
		switch (ret) {
		case Z_OK:
			goto out;
		case Z_BUF_ERROR:
			/* we don't use realloc(), because it may
			 * (in this case) unneccessarily copy the old block
			 * to the new allocated block.
			 */
			free(dest);
			destLen += EXPAND_DECOMPRESS_DESTBUF_BYTES;
			dest = static_cast<Bytef *>(malloc(destLen));
			if (!dest)
				return false;
			break;
		default:
			free(dest);
			return false;
		}
	}
out:
	data->duplicate(reinterpret_cast<const char *>(dest),
			destLen);
	free(dest);
	return true;
}

bool FileCompress::decompressBzip2(QByteArray *data)
{
	int ret;
	char *dest, *source;
	unsigned int sourceLen, destLen;

	sourceLen = data->size();
	destLen = BZIP2_DECOM_DESTLEN(sourceLen);
	source = static_cast<char *>(malloc(sourceLen));
	if (!source)
		return false;
	memcpy(source, data->data(), sourceLen);
	dest = static_cast<char *>(malloc(destLen));
	if (!dest) {
		free(source);
		return false;
	}
	while (1) {
		ret = BZ2_bzBuffToBuffDecompress(dest,
						 &destLen,
						 source,
						 sourceLen,
						 0,
						 0);
		switch (ret) {
		case BZ_OK:
			goto out;
		case BZ_OUTBUFF_FULL:
			/* we don't use realloc(), because it may
			 * (in this case) unneccessarily copy the old block
			 * to the new allocated block.
			 */
			free(dest);
			destLen += EXPAND_DECOMPRESS_DESTBUF_BYTES;
			dest = static_cast<char *>(malloc(destLen));
			if (!dest) {
				free(source);
				return false;
			}
			break;
		default:
			free(source),
			free(dest);
			return false;
		}
	}
out:
	free(source);
	data->duplicate(dest, destLen);
	free(dest);
	return true;
}
