/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "libpwmanager/filecrypto.h"
#include "libpwmanager/blowfish.h"
#include "libpwmanager/libgcryptif.h"
#include "rawfile/rawfile.h"
#include "pwmexception.h"

#ifdef CONFIG_PWMANAGER_GCRY
# include <gcrypt.h>
#else
# warning "You're not using libgcrypt! This works, but is less secure!"
# include <stdlib.h>
# include <time.h>
#endif // CONFIG_PWMANAGER_GCRY


FileCrypto::FileCrypto()
{
#ifndef CONFIG_PWMANAGER_GCRY
	srand(time(0));
#endif // CONFIG_PWMANAGER_GCRY
}

FileCrypto::~FileCrypto()
{
}

void FileCrypto::padData(unsigned char *buf,
			 size_t bufLen,
			 size_t boundary)
{
	size_t numPadBytes = boundary - ((bufLen + 1) % boundary);
	buf[bufLen] = 0x01;
	size_t i = 0;
	unsigned char c;
	unsigned char *b;
	while (i < numPadBytes) {
#ifdef CONFIG_PWMANAGER_GCRY
		gcry_randomize(&c,
			       1,
			       GCRY_VERY_STRONG_RANDOM);
#else
		c = rand() % 0xFF;
#endif // CONFIG_PWMANAGER_GCRY
		if (c == 0x01)
			continue;
		b = buf + bufLen + 1 + i;
		*b = c;
		++i;
	}
}

void FileCrypto::unpadData(const unsigned char *buf,
		 	   size_t *bufLen)
{
	size_t pos;
	BUG_ON(*bufLen % 8);
	pos = *bufLen - 1;
	while (buf[pos] != 0x01) {
		BUG_ON(!pos);
		--pos;
	}
	*bufLen = pos;
}

bool FileCrypto::encrypt(QByteArray *data,
			 const QByteArray &passphrase,
			 unsigned char cryptAlgo,
			 unsigned char hashAlgo)
{
	switch (cryptAlgo) {
	case PWM_CRYPT_BLOWFISH: {
		Blowfish bf;
		bf.bf_setkey(reinterpret_cast<const byte *>(passphrase.data()),
			     passphrase.size());
		int inLen = data->size();
		byte *tmp_buf = new byte[getBufLen(inLen, 8)];
		data->resize(getBufLen(inLen, 8));
		padData(reinterpret_cast<unsigned char *>(data->data()),
			inLen, 8);
		bf.bf_encrypt(tmp_buf,
			      reinterpret_cast<const byte *>(data->data()),
			      data->size());
		data->duplicate(reinterpret_cast<const char *>(tmp_buf),
				data->size());
		delete [] tmp_buf;
		break;
	} case PWM_CRYPT_AES128:
		/*... fall through */
	case PWM_CRYPT_AES192:
	case PWM_CRYPT_AES256:
	case PWM_CRYPT_3DES:
	case PWM_CRYPT_TWOFISH:
	case PWM_CRYPT_TWOFISH128: {
TODO();//TODO: implement!
		break;
	} default:
		return false;
	}
	return true;
}

bool FileCrypto::decrypt(QByteArray *data,
			 const QByteArray &passphrase,
			 unsigned char cryptAlgo,
			 unsigned char hashAlgo)
{
	switch (cryptAlgo) {
	case PWM_CRYPT_BLOWFISH: {
		Blowfish bf;
		bf.bf_setkey(reinterpret_cast<const byte *>(passphrase.data()),
			     passphrase.size());
		byte *tmp_buf = new byte[data->size()];
		bf.bf_decrypt(tmp_buf,
			      reinterpret_cast<const byte *>(data->data()),
			      data->size());
		size_t cryptLen = data->size();
		unpadData(reinterpret_cast<const unsigned char *>(tmp_buf),
			  &cryptLen);
		data->duplicate(reinterpret_cast<const char *>(tmp_buf),
				cryptLen);
		delete [] tmp_buf;
		break;
	} case PWM_CRYPT_AES128:
		/*... fall through */
	case PWM_CRYPT_AES192:
	case PWM_CRYPT_AES256:
	case PWM_CRYPT_3DES:
	case PWM_CRYPT_TWOFISH:
	case PWM_CRYPT_TWOFISH128: {
TODO();//TODO: implement and get rid of this PwMerror!
#if 0
		if (!LibGCryptIf::available())
			return false;
		LibGCryptIf gc;
		PwMerror err;
		err = gc.decrypt(&decrypted,
				 &cryptLen,
				 encrypted,
				 cryptLen,
				 reinterpret_cast<const unsigned char *>(pw->latin1()),
				 pw->length(),
				 algo);
		if (err != e_success) {
			delete [] encrypted;
			delete [] decrypted;
			return e_cryptNotImpl;
		}
#endif
		break;
	} default:
		return false;
	}
	return true;
}
