/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "libpwmanager/libpwmanager.h"
#include "libpwmanager/base64.h"
#include "libpwmanager/rawinterface.h"
#include "libpwmanager/privatedata.h"
#include "libpwmanager/uniqueid.h"
#include "randomizer/randomizer.h"
#include "pwmexception.h"

#include <klocale.h>
#include <kapplication.h>

#include <dcopclient.h>

#include <qptrlist.h>


#define DEFAULT_ROOT_NAME	I18N_NOOP("/")
#define DEFAULT_PW_FIELD_ID	I18N_NOOP("Password")
#define DEFAULT_BIN_FIELD_ID	I18N_NOOP("Binary Data")
#define UNIQUEID_LENGTH		(128 / 8) /* 128 bit */


#define WARN_WRITEPROT(root, returnVal)	do {		\
	DataRoot *_r = (root);				\
	if (_r && unlikely(_r->writeProtected)) {	\
		_warn_writeprot(__FILE__,		\
				__LINE__,		\
				__PRETTY_FUNCTION__);	\
		return returnVal ;			\
	}						\
					} while (0)


namespace PwManager
{

static void _warn_writeprot(const char *file, int line, const char *func)
{
	printError(QString("FILE: ")
		   + file + ":" + QString::number(line) + "\n"
		   "FUNCTION: " + func + "\n"
		   "Application BUG detected: "
		   "Attempt to write a protected tree!\n"
		   "This is a bug in the application! You "
		   "must always call DataRoot::beginWrite() "
		   "before you start writing and "
		   "DataRoot::endWrite() after you finished.");
}


/***************************************************************************
 *  class DataMeta
 ***************************************************************************/

DataMeta::DataMeta(DataItem *_di)
 : di (_di)
 , updateInt (0)
 , uniqueid (0)
{
	QDateTime tmp(QDateTime::currentDateTime());
	create = tmp;
	update = tmp;
	valid = tmp;
}

DataMeta::~DataMeta()
{
	delete_ifnot_null(uniqueid);
}

void DataMeta::clear()
{
	WARN_WRITEPROT(di->root, /* void */);
	QDateTime tmp(QDateTime::currentDateTime());
	create = tmp;
	expire = QDateTime();
	update = tmp;
	valid = tmp;
	updateInt = 0;
	uniqueid = 0;
	di->setUpdatedFlags(false);
}

DataMeta & DataMeta::operator= (const DataMeta &m)
{
	di = m.di;
	create = m.create;
	valid = m.valid;
	expire = m.expire;
	update = m.update;
	updateInt = m.updateInt;
	// uniqueid is not cloned here.
	return *this;
}

QDateTime DataMeta::getCreateDate() const
{
	return create;
}

QDateTime DataMeta::getValidDate() const
{
	return valid;
}

void DataMeta::setValidDate(const QDateTime &_valid)
{
	WARN_WRITEPROT(di->root, /* void */);
	valid = _valid;
	di->setUpdatedFlags(false);
}

QDateTime DataMeta::getExpireDate() const
{
	return expire;
}

void DataMeta::setExpireDate(const QDateTime &_expire)
{
	WARN_WRITEPROT(di->root, /* void */);
	expire = _expire;
	di->setUpdatedFlags(false);
}

QDateTime DataMeta::getUpdateDate() const
{
	return update;
}

unsigned long DataMeta::getUpdateInt() const
{
	return updateInt;
}

void DataMeta::setUpdateInt(unsigned long _updateInt)
{
	WARN_WRITEPROT(di->root, /* void */);
	updateInt = _updateInt;
	di->setUpdatedFlags(false);
}

UniqueId DataMeta::getUniqueId() const
{
	return *uniqueid;
}

bool DataMeta::isValid(const QDateTime &dateTime) const
{
	if (valid.isNull())
		return true;
	return (valid < dateTime);
}
bool DataMeta::isExpired(const QDateTime &dateTime) const
{
	if (expire.isNull())
		return false;
	return (expire < dateTime);
}
bool DataMeta::isUpdateIntOver(const QDateTime &dateTime) const
{
	if (updateInt == 0 ||
	    update.isNull())
		return false;
	QDateTime d(update);
	return (d.addSecs(updateInt * 60) < dateTime);
}


/***************************************************************************
 *  class DataItem
 ***************************************************************************/

DataItem::DataItem()
 : type (NoItem)
 , root (0)
 , parentCat (0)
 , rev (0)
 , deleteInParent (true)
{
	meta = new DataMeta(this);
}

DataItem::~DataItem()
{
	//FIXME: still correct here?
	// delete ourself in the parent category.
	if (parentCat && deleteInParent) {
		bool mustWriteProtect = false;
		if (root && unlikely(root->writeProtected)) {
			printWarn("You deleted a PwManager::DataItem (or one "
				  "of its derived classes) which is part of a "
				  "write protected data tree! This time you are lucky "
				  "and I will drop and reaquire write protection for you.");
			root->beginWrite();
			mustWriteProtect = true;
		}
		parentCat->children.remove(this);
		if (root && unlikely(mustWriteProtect))
			root->endWrite();
	}
	delete meta;
}

DataItem & DataItem::operator= (const DataItem &i)
{
	WARN_WRITEPROT(root, *this);
	root = i.root;
	parentCat = i.parentCat;
	rev = i.rev;
	deleteInParent = i.deleteInParent;
	PWM_ASSERT(meta);
	*meta = *(i.meta);
	name = i.name;
	return *this;
}

void DataItem::clear()
{
	WARN_WRITEPROT(root, /* void */);
	name = QString();
	setUpdatedFlags();
}

QString DataItem::getName() const
{
	return name;
}

void DataItem::setName(const QString &_name)
{
	WARN_WRITEPROT(root, /* void */);
	name = _name;
	setUpdatedFlags();
}

DataItem::itemType DataItem::getType() const
{
	return type;
}

DataMeta * DataItem::getMeta()
{
	return meta;
}

const DataMeta * DataItem::getMeta() const
{
	return meta;
}

DataCategory * DataItem::parentCategory()
{
	return parentCat;
}

DataEntry * DataItem::toEntry()
{
	if (unlikely(type != Entry))
		return 0;
	return (dynamic_cast<DataEntry *>(this));
}

DataCategory * DataItem::toCategory()
{
	if (unlikely(type != Category))
		return 0;
	return (dynamic_cast<DataCategory *>(this));
}

unsigned int DataItem::getRev() const
{
	return rev;
}

void DataItem::setUpdatedFlags(bool setDate)
{
	++rev;
	if (root)
		root->treeDirty = true;
	if (setDate)
		meta->update = QDateTime::currentDateTime();
}


/***************************************************************************
 *  class DataEntry
 ***************************************************************************/

DataEntry::DataEntry()
{
	fields.setAutoDelete(true);
}

DataEntry::~DataEntry()
{
}

DataEntry & DataEntry::operator= (const DataEntry &e)
{
	WARN_WRITEPROT(root, *this);
	fields = e.fields;
	DataItem::operator= (e);
	return *this;
}

DataItem * DataEntry::clone() const
{
	DataEntry *e = new DataEntry;
	*e = *this;
	return e;
}

void DataEntry::clear()
{
	WARN_WRITEPROT(root, /* void */);
	fields.clear();
	DataItem::clear();
}

void DataEntry::appendPassword(const QString &pw)
{
	WARN_WRITEPROT(root, /* void */);
	FieldItem *newItem = new FieldItem;
	newItem->type = FieldItem::password;
	newItem->id = i18n(DEFAULT_PW_FIELD_ID);
	QDataStream ds(newItem->data, IO_WriteOnly);
	ds << pw;
	fields.append(newItem);
}

QString DataEntry::getPassword() const
{
	QString ret;
	FieldList::const_iterator i = fields.begin(),
				  end = fields.end();
	const FieldItem *item;
	for ( ; i != end; ++i) {
		item = *i;
		if (item->type == FieldItem::password) {
			QDataStream ds(item->data, IO_ReadOnly);
			ds >> ret;
			break;
		}
	}
	return ret;
}

void DataEntry::appendBinary(const QByteArray &data)
{
	WARN_WRITEPROT(root, /* void */);
	FieldItem *newItem = new FieldItem;
	newItem->type = FieldItem::binary;
	newItem->id = i18n(DEFAULT_BIN_FIELD_ID);
	newItem->data.duplicate(data);
	fields.append(newItem);
}

QByteArray DataEntry::getBinary() const
{
	QByteArray ret;
	FieldList::const_iterator i = fields.begin(),
				  end = fields.end();
	const FieldItem *item;
	for ( ; i != end; ++i) {
		item = *i;
		if (item->type == FieldItem::binary) {
			ret = item->data;
			break;
		}
	}
	return ret;
}

void DataEntry::appendField(const QString &id, const QByteArray &data)
{
	WARN_WRITEPROT(root, /* void */);
	FieldItem *newItem = new FieldItem;
	newItem->id = id;
	newItem->data.duplicate(data);
	fields.append(newItem);
}

DataEntry::FieldList * DataEntry::getFieldList()
{
	return &fields;
}

DataEntry::iterator DataEntry::fieldsBegin()
{
	return fields.begin();
}

DataEntry::iterator DataEntry::fieldsEnd()
{
	return fields.end();
}

DataEntry::const_iterator DataEntry::fieldsBegin() const
{
	return fields.begin();
}

DataEntry::const_iterator DataEntry::fieldsEnd() const
{
	return fields.end();
}


/***************************************************************************
 *  class DataCategory
 ***************************************************************************/

DataCategory::DataCategory()
{
	type = Category;
}

DataCategory::~DataCategory()
{
	QPtrList<DataItem>::iterator i = children.begin(),
				     end = children.end();
	while (i != end) {
		// We need this to avoid recursive delete
		(*i)->deleteInParent = false;
		++i;
	}
	children.setAutoDelete(true);
	children.clear();
}

DataCategory & DataCategory::operator= (const DataCategory &e)
{
	WARN_WRITEPROT(root, *this);
	children = e.children;
	DataItem::operator= (e);
	return *this;
}

DataItem * DataCategory::clone() const
{
	DataCategory *c = new DataCategory;
	*c = *this;
	return c;
}

void DataCategory::clear()
{
	WARN_WRITEPROT(root, /* void */);
	QPtrList<DataItem>::iterator i = children.begin(),
				     end = children.end();
	while (i != end) {
		// We need this to avoid recursive delete
		(*i)->deleteInParent = false;//FIXME: check if we still need this!
		++i;
	}
	children.setAutoDelete(true);
	children.clear();
	children.setAutoDelete(false);
	DataItem::clear();
}

DataItem * DataCategory::appendItem(const DataItem &newItem)
{
	WARN_WRITEPROT(root, 0);
	DataItem *newItemClone = newItem.clone();
	PWM_ASSERT(newItemClone);
	newItemClone->parentCat = this;
	if (root)
		newItemClone->root = root;
	else {
		// hm, "this" seems to be the root
		DataRoot *r = dynamic_cast<DataRoot *>(this);
		BUG_ON(!r);
		newItemClone->root = r;
	}
	newItemClone->meta->uniqueid = new UniqueId(newItemClone->root->uniqueIdContext);
	if (!newItemClone->meta->uniqueid->generate()) {
		delete newItemClone;
		printDebug("No unused UniqueId left :(");
		return 0;
	}
	children.append(newItemClone);
	setUpdatedFlags();
	return newItemClone;
}

void DataCategory::eraseItem(const DataItem *item)
{
	WARN_WRITEPROT(root, /* void */);
	delete item;
	// The item deletes itself out of the "children" list.
	PWM_ASSERT(children.find(item) == -1);
	setUpdatedFlags();
}

void DataCategory::eraseItem(iterator item)
{
	WARN_WRITEPROT(root, /* void */);
	DataItem *p = *item;
	delete p;
	// The item deletes itself out of the "children" list.
	PWM_ASSERT(children.find(p) == -1);
	setUpdatedFlags();
}

DataCategory::iterator DataCategory::childBegin()
{
	return children.begin();
}

DataCategory::iterator DataCategory::childEnd()
{
	return children.end();
}

DataCategory::const_iterator DataCategory::childBegin() const
{
	return children.begin();
}

DataCategory::const_iterator DataCategory::childEnd() const
{
	return children.end();
}

unsigned int DataCategory::count() const
{
	return children.count();
}

/***************************************************************************
 *  class DataRoot
 ***************************************************************************/

DataRoot::DataRoot(PwFile *_pwfile)
 : pwfile (_pwfile)
 , writeProtected (false)
 , treeDirty (false)
{
	name = i18n(DEFAULT_ROOT_NAME);
	writeProtected = true;
	uniqueIdContext = new UniqueIdContext(UNIQUEID_LENGTH);
	/* This is special case. We need to allocate the uniqueid
	 * instance manually here.
	 */
	meta->uniqueid = new UniqueId(uniqueIdContext);
	bool ok = meta->uniqueid->generate();
	/* Generating the unique ID can not (must not) fail
	 * here, because this is the first ID we calculate for
	 * this context.
	 */
	PWM_ASSERT(ok);
}

DataRoot::~DataRoot()
{
	delete uniqueIdContext;
}

void DataRoot::beginWrite()
{
	if (unlikely(!writeProtected)) {
		printError("Called DataRoot::beginWrite() twice!\n"
			   "This is a bug in the application!");
	}
	PWM_ASSERT(!treeDirty);
	writeProtected = false;
}

void DataRoot::endWrite()
{
	if (unlikely(writeProtected)) {
		printError("Called DataRoot::endWrite() twice, or called "
			   "without calling DataRoot::beginWrite first.!\n"
			   "This is a bug in the application!");
	}
	writeProtected = true;
	if (!treeDirty)
		return;
	pwfile->populateTree();
	treeDirty = false;
	/* The daemon doesn't populate the data to ourself.
	 * So request a view update by hand.
	 */
	pwfile->dirty = true;
	uniqueIdContext->clearTrash();
}

//XXX
#if 0
/***************************************************************************
 *  class DataTree
 ***************************************************************************/

DataTree::DataTree(PwFile *_pwfile)
 : DataRoot (_pwfile)
{
}

DataTree::~DataTree()
{
}
#endif


/***************************************************************************
 *  class LibPwManager
 ***************************************************************************/

LibPwManager * LibPwManager::lib (0);

LibPwManager::LibPwManager()
 : lastErr (eUnknown)
{
	p = new LibPwManager_private;
}

LibPwManager::~LibPwManager()
{
	lib = 0;
	delete p;
}

LibPwManager * LibPwManager::initialize()
{
	if (lib)
		return lib;
	LibPwManager *ret = new LibPwManager;
	RawInterface *raw = new RawInterface;

	// probe for DCOP access.
	if (raw->isEnabled(RawInterface::dcopAccess)) {
		if (raw->setAccessType(RawInterface::dcopAccess)) {
			printDebug("Using DCOP access");
			goto out;
		} else
			printWarn("Couldn't register at PwManager Daemon!");
	}

	// probe for DIRECT access.
	if (raw->isEnabled(RawInterface::directAccess)) {
		if (raw->setAccessType(RawInterface::directAccess)) {
			printWarn("Using direct access: You may loose "
				  "data, if you access the "
				  "same file with more then one app at the "
				  "same time!");
			goto out;
		} else
			printWarn("Couldn't do direct file access!");
	}

	// No working interface found. :(
	printError("No working PwManager interface found!");
	delete raw;
	delete ret;
	return 0;

out:
	ret->p->raw = raw;
	lib = ret;
	return ret;
}

QString LibPwManager::getErrorStr(ErrCode err)
{
	const char *errorMap[] = {
		I18N_NOOP("Success"),
		I18N_NOOP("Unknown error"),
		I18N_NOOP("File opening failed"),
		I18N_NOOP("File read/write failed"),
		I18N_NOOP("Wrong file format"),
		I18N_NOOP("The user interrupted the operation"),
		I18N_NOOP("No file path available")
	};
	if (unlikely(err < 0 ||
	    err > static_cast<ErrCode>(array_size(errorMap)) - 1)) {
		BUG();
		return "INTERNAL ERROR";
	}
	return i18n(errorMap[err]);
}

ErrCode LibPwManager::getError()
{
	return lastErr;
}

PwFile * LibPwManager::newFile()
{
	PwFile *ret = new PwFile(this);
	connect(p->raw, SIGNAL(dataDirty(int, const QString &,
					 const FileHeader_t &,
					 const FileData_t &)),
		ret, SLOT(dataDirty_slot(int, const QString &,
					 const FileHeader_t &,
					 const FileData_t &)));
	lastErr = eSuccess;
	return ret;
}

PwFile * LibPwManager::openFile(const QString &path)
{
	PwFile *ret = new PwFile(this);
	connect(p->raw, SIGNAL(dataDirty(int, const QString &,
					 const FileHeader_t &,
					 const FileData_t &)),
		ret, SLOT(dataDirty_slot(int, const QString &,
					 const FileHeader_t &,
					 const FileData_t &)));
	if (!ret->open(path)) {
		lastErr = ret->getError();
		delete ret;
		/* FIXME: this delete crashes with:
#0  0xa785a8b3 in QGList::find (this=0x706f6c85, d=0x805e780, fromStart=true) at tools/qglist.cpp:745
#1  0xa785a318 in QGList::remove (this=0x706f6c85, d=0x805e780) at tools/qglist.cpp:537
#2  0xa7fc3a54 in ~DataItem (this=0x805e780) at qptrlist.h:85
#3  0xa7fc3df1 in ~DataCategory (this=0x805e780) at qptrlist.h:70
#4  0xa7fc3ed8 in ~DataRoot (this=0x0) at libpwmanager.cpp:574
#5  0xa7fc3f0d in ~DataTree (this=0x805e780) at libpwmanager.cpp:625
#6  0xa7fc7bd8 in ~PwFile (this=0x805e3c0) at pwfile.cpp:255
#7  0xa7fbf6c5 in PwManager::LibPwManager::openFile (this=0x805cdc0, path=@0x706f6c85, 
    parent=0x706f6c85) at libpwmanager.cpp:757
#8  0x0804a620 in CmdHandler::cmd_open (this=0xaffff298, args=@0x805d6bc) at basic_string.h:476
#9  0x0804a7a1 in CmdParser::handleCommands (this=0xaffff280) at stl_iterator.h:614
#10 0x0804debf in PwMTxtArgs::parseArgs (this=0xaffff200) at pwmtxtargs.cpp:66
#11 0x0804d9ef in main (argc=1886350469, argv=0x706f6c85) at txtmain.h:28
*/
		return 0;
	}
	lastErr = eSuccess;
	return ret;
}

void LibPwManager::complain(int complainId)
{
	p->raw->complain(complainId);
}

QWidget * LibPwManager::getParentWidget() const
{
	return p->parentWidget;
}

void LibPwManager::setParentWidget(QWidget *w)
{
	p->parentWidget = w;
}


} // namespace PwManager

#include "libpwmanager.moc"
//#include "daemon/daemondcop_stub.cpp"
