/***************************************************************************
 *                                                                         *
 *   PwManager -- The  Personal Wallet Manager                             *
 *                                                                         *
 *   Public PwManager interface.                                           *
 *                                                                         *
 *   copyright (C) 2003, 2004, 2005  by Michael Buesch                     *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 *   http://passwordmanager.sourceforge.net/gplv2.txt                      *
 *                                                                         *
 ***************************************************************************/

#ifndef LIB_PWMANAGER_H_
#define LIB_PWMANAGER_H_

#if !defined(__cplusplus) && !defined(c_plusplus)
# error "libpwmanager.h is for C++ only."
#endif

#include <qobject.h>
#include <qptrlist.h>
#include <qdatetime.h>

#include <stdint.h>


namespace PwManager
{

class DataMeta;
class DataItem;
class DataEntry;
class DataCategory;
class DataRoot;
class PwFile;
class LibPwManager;
class Serializer;
class RawInterface;
class FileHeader_t;
class FileData_t;
class UniqueId;
class UniqueIdContext;

/** Global error codes. */
enum ErrCode
{
	eSuccess = 0,
	eUnknown = 1,
	eOpen = 2,
	eIO = 3,
	eFormat = 4,
	eUser = 5,
	eNoFilepath = 6
};

//TODO: support for a user defined runtime data class, which is not stored in the file.

/** Unique DataItem ID. */
class UniqueId
{
public:
	/** Constructor. */
	UniqueId(UniqueIdContext *_context);
	/** Copy constructor.
	  * The copy shares the data.
	  */
	UniqueId(const UniqueId &copySource);
	/** Destructor. */
	~UniqueId();

	/** Assign operator.
	  * The copy shares the data.
	  */
	UniqueId & operator= (const UniqueId &copySource);

	/** Returns the ID.
	  * It is unique through out the UniqueIdContext.
	  */
	QByteArray getId() const;

	/** Generate a new ID.
	  * Returns true on success or false if all IDs are
	  * already in use in the current context.
	  */
	bool generate();
	/** Assign a given ID.
	  * Returns false, if the ID already exists in the
	  * current context.
	  */
	bool assign(const QByteArray &idData);

protected:
	/** Pointer to the context.
	  * @internal
	  */
	UniqueIdContext *context;
	/** Pointer to the QByteArray containing
	  * the ID data.
	  * @internal
	  */
	QByteArray *id;
};

/** Meta data for a DataItem */
class DataMeta
{
	friend class DataRoot;
	friend class DataItem;
	friend class DataCategory;
	friend class Serializer;

protected:
	/** Constructor.
	  * @internal
	  */
	DataMeta(DataItem *_di);
	/** = operator */
	DataMeta & operator= (const DataMeta &m);

public:
	/** Destructor. */
	~DataMeta();

	/** Get the creation date.
	  * @return Returns the QDateTime of the time at which
	  *         the constructor of this class was called.
	  *         (constructor called == item created)
	  */
	QDateTime getCreateDate() const;
	/** Get the valid date.
	  * @return Returns the QDateTime of the "is valid as of xxx"
	  *         Date. The QDateTime represents the Date at which
	  *         this item becomes/became a valid item.
	  */
	QDateTime getValidDate() const;
	/** Set the valid date.
	  * See getValidDate()
	  * @param _valid The QDateTime for the
	  *               "is valid as of xxx" date.
	  */
	void setValidDate(const QDateTime &_valid);
	/** Get the expire date.
	  * @return Returns the QDateTime of the "expires at xxx"
	  *         Date. The QDateTime represents the Date at which
	  *         this item expires/expired.
	  */
	QDateTime getExpireDate() const;
	/** Set the expire date.
	  * See getExpireDate()
	  * @param _expire The QDateTime for the "expires at xxx" date.
	  */
	void setExpireDate(const QDateTime &_expire);
	/** Get the "last updated at" date.
	  * @return Returns the QDateTime at which this item's data got
	  *         last updated.
	  */
	QDateTime getUpdateDate() const;
	/** Get the update interval (in minutes).
	  * @return Returns the update interval (in minutes) in which
	  *         the user has to update the "data" in this item.
	  *         In case of a DataPasswordEntry, this is the password.
	  *         In case of a DataBinaryEntry, this is the binary data.
	  *         This has currently no meaning in case of a DataCategory.
	  */
	unsigned long getUpdateInt() const;
	/** Set the update interval (in minutes).
	  * See getUpdateDate()
	  * @param _updateInt The update intervall in minutes.
	  */
	void setUpdateInt(unsigned long _updateInt);
	/** Get the Unique-ID.
	  * @return Returns the Unique-ID for this item.
	  *         The item can be identified in the whole data tree
	  *         with this ID, as it's unique in the tree.
	  */
	UniqueId getUniqueId() const;

	/** Reset all data to the default values.
	  * Set the creation date to current date.
	  */
	void clear();
	/** Helper function: Checks if the item is
	  * valid at "dateTime".
	  */
	bool isValid(const QDateTime &dateTime = QDateTime::currentDateTime()) const;
	/** Helper function: Checks if the item has expired
	  * at "dateTime".
	  */
	bool isExpired(const QDateTime &dateTime = QDateTime::currentDateTime()) const;
	/** Helper function: Checks if the update interval
	  * of the item is over at "dateTime".
	  */
	bool isUpdateIntOver(const QDateTime &dateTime = QDateTime::currentDateTime()) const;

protected:
	/** Pointer to the DataItem.
	  * @internal
	  */
	DataItem	*di;
	/** Creation date of the PwMDataItem to which
	  * this meta data belongs.
	  * @internal
	  */
	QDateTime	create;
	/** Becomes valid on this date.
	  * @internal
	  */
	QDateTime	valid;
	/** Expire date.
	  * @internal
	  */
	QDateTime	expire;
	/** Update date (last updated at this date).
	  * @internal
	  */
	QDateTime	update;
	/** Update interval (in minutes). Time since the
	  * last update to remind the user to update the item.
	  * 0 disables.
	  * @internal
	  */
	unsigned long	updateInt;
	/** Unique ID value.
	  * @internal
	  */
	UniqueId	*uniqueid;
};


/** This is the pure virtual and abstract representation of an item
  * in the in-memory document tree.
  */
class DataItem
{
	friend class DataMeta;
	friend class DataCategory;

public:
	/** A DataItem has an item-type.
	  * This enum lists all possible type values.
	  */
	enum itemType
	{
		NoItem		= 0,
		Entry		= 1,
		Category	= 2
	};

public:
	/** Constructor.
	  */
	DataItem();
	/** Destructor */
	virtual ~DataItem();

	/** = operator */
	DataItem & operator= (const DataItem &i);//FIXME: virtual???
	/** Clones this DataItem. This allocates a new DataItem with "new",
	  * clones all internal data and returns a pointer to the new item.
	  * You are responsible to delete the returned pointer.
	  * @return Returns a pointer to the cloned DataItem.
	  */
	virtual DataItem * clone() const = 0;

	/** Get the human readable name string.
	  * @return Returns a human readable and descriptive name string.
	  */
	QString getName() const;
	/** Set the human readable name string.
	  * @param _name The human readable and descriptive name string.
	  */
	void setName(const QString &_name);

	/** Returns the type of this item. */
	itemType getType() const;
	/** Returns a pointer to the item meta data. */
	DataMeta * getMeta();
	const DataMeta * getMeta() const;
	/** returns a pointer to the category item this
	  * items belongs to. If this is a DataRoot item,
	  * this returns a NULL pointer.
	  * If this item doesn't belong to a category yet,
	  * it returns a NULL pointer, too.
	  */
	DataCategory * parentCategory();
	/** Convert this DataItem to a DataEntry.
	  * Returns a NULL pointer if this DataItem is not
	  * of the itemType::passwordItem or itemType::binaryItem
	  */
	DataEntry * toEntry();
	/** Convert this DataItem to a DataCategory.
	  * Returns a NULL pointer if this DataItem is not
	  * of the itemType::categoryItem
	  */
	DataCategory * toCategory();
	/** clears all data in this item.
	  * This function is pure virtual here and does _NOT_
	  * delete the meta data. If you want to clear
	  * (reset to default) the meta data, call DataMeta::clear()
	  */
	virtual void clear();
	/** Get the data revision counter. This counter can be used
	  * to easily, efficiently determine if this data item
	  * has changed since some time.
	  * This counter is incremented on every update.
	  */
	unsigned int getRev() const;

protected:
	/** Set internal flags that the data of this entry got updated.
	  * @internal
	  */
	void setUpdatedFlags(bool setDate = true);

protected:
	/** The type of this item. See enum itemType.
	  * @internal
	  */
	itemType	type;
	/** Meta data for this data item.
	  * @internal
	  */
	DataMeta	*meta;
	/** Pointer to the root of the tree which
	  * this item belongs to. If this _is_ the root,
	  * this is a NULL pointer.
	  * @internal
	  */
	DataRoot *root;
	/** Pointer to the parent category.
	  * @internal
	  */
	DataCategory *parentCat;
	/** Data revision counter.
	  * @internal
	  */
	unsigned int rev;
	/** Delete "this" in the parent list, if our
	  * destructor is called?
	  * @internal
	  */
	bool deleteInParent;
	/** Human readable and descriptive name string.
	  * @internal
	  */
	QString name;
};

class DataEntry : public DataItem
{
	friend class Serializer;

public:
	/** Data field type. */
	class FieldItem
	{
	public:
		/** Constructor. */
		FieldItem()
		 : type (custom)
			{ }
		/** clear all data in the field. */
		void clear()
			{
				type = custom;
				id = QString::null;
				data.resize(0);
			}
		/** Field type */
		enum type_t {
			custom = 0,
			password = 1,
			binary = 2
		} type;
		/** Field ID */
		QString id;
		/** Field data */
		QByteArray data;
	};

	class FieldList : public QPtrList<FieldItem>
	{
	public:
		FieldList() { }
		FieldItem * newItem(FieldItem *d)
			{
//FIXME: debug. remove.
//cout << "newItem" << endl;
				FieldItem *ret = new FieldItem;
				ret->type = d->type;
				ret->id = d->id;
				ret->data.duplicate(d->data);
				return ret;
			}
	};

public:
	typedef QPtrList<FieldItem>::iterator		iterator;
	typedef QPtrList<FieldItem>::const_iterator	const_iterator;

public:
	/** Constructor.
	  */
	DataEntry();
	/** Destructor. */
	virtual ~DataEntry();

	/** = operator */
	DataEntry & operator= (const DataEntry &e);
	/** Reimplemented from DataItem. */
	virtual DataItem * clone() const;

	/** Clears all data in this item. */
	virtual void clear();

	/** Append a new password data field. */
	void appendPassword(const QString &pw);
	/** Get the password string field.
	  * @return Returns the first data field with the "password" id.
	  *         Returns QString::null if there is no such field.
	  */
	QString getPassword() const;
	/** Append a new binary data field. */
	void appendBinary(const QByteArray &data);
	/** Get the binary data field.
	  * @return Returns the first data field with the "binary" id.
	  *         Returns an empty array if there is no such field.
	  */
	QByteArray getBinary() const;

	/** Append a new custom data field. */
	void appendField(const QString &id, const QByteArray &data);
	/** Get the current data field list. */
	FieldList * getFieldList();
	/** Returns the begin-iterator for the field list. */
	iterator fieldsBegin();
	const_iterator fieldsBegin() const;
	/** Returns the end-iterator for the field list. */
	iterator fieldsEnd();
	const_iterator fieldsEnd() const;

protected:
	/** List of data fields.
	  * @internal
	  */
	FieldList fields;
};

/** representation of a category as dataItem */
class DataCategory : public DataItem
{
	friend class DataItem;
	friend class DataRoot;
	friend class DataEntry;

public:
	/** Iterator to iterate through all children in this category. */
	typedef QPtrList<DataItem>::iterator		iterator;
	/** Same as iterator, but for const operations. */
	typedef QPtrList<DataItem>::const_iterator	const_iterator;

public:
	/** Constructor.
	  */
	DataCategory();
	/** Destructor. */
	virtual ~DataCategory();

	/** = operator */
	DataCategory & operator= (const DataCategory &c);
	/** Reimplemented from DataItem. */
	virtual DataItem * clone() const;

	/** Clears all data in this item. */
	virtual void clear();

	/** Append a new DataItem to this category item.
	  *///FIXME: this can fail. describe why and what happens.
	DataItem * appendItem(const DataItem &newItem);
	/** Erase this item. */
	void eraseItem(const DataItem *item);
	/** Erase this item. */
	void eraseItem(iterator item);
	/** Returns the begin-iterator for all children DataItems. */
	iterator childBegin();
	const_iterator childBegin() const;
	/** Returns the end-iterator for all children DataItems. */
	iterator childEnd();
	const_iterator childEnd() const;
	/** Returns the number of children items. */
	unsigned int count() const;

protected:
	/** List of all children in this category.
	  * @internal
	  */
	QPtrList<DataItem> children;
};

/** The root-element of the data tree. */
class DataRoot : public DataCategory
{
	friend class PwFile;
	friend class DataMeta;
	friend class DataItem;
	friend class DataEntry;
	friend class DataPasswordEntry;
	friend class DataBinaryEntry;
	friend class DataCategory;

protected:
	/** Constructor.
	  * @internal
	  */
	DataRoot(PwFile *_pwfile);
	/** Destructor.
	  * @internal
	  */
	virtual ~DataRoot();

public:
	/** Removes the write protection on the whole
	  * tree and allows writing.
	  * Call this before you start writing.
	  */
	void beginWrite();
	/** Resets the write protection on the whole tree
	  * and populates/applies the changes.
	  * Call this after writing.
	  */
	void endWrite();

protected:
	/** Pointer to PwFile object.
	  * @internal
	  */
	PwFile *pwfile;
	/** Write protection flag for this tree.
	  * @internal
	  */
	bool writeProtected;
	/** Data tree dirty flag.  FIXME: still neccessary?
	  * This is true if something in the tree changed since
	  * the last endWrite() call. endWrite() will flush the
	  * to the rawinterface and reset this to false.
	  * @internal
	  */
	bool treeDirty;
	/** Context for generating the random and unique item IDs.
	  * @internal
	  */
	UniqueIdContext *uniqueIdContext;
};

//FIXME
#if 0
/** Wrapper class for accessing single objects in the whole tree.
  * Use this if you don't want to walk through the whole tree, but
  * only access some single objects in it.
  * If you want do do advanced operation on the tree (iterate through
  * it, etc...), use the functions of the underlying class DataRoot .
  */
class DataTree : public DataRoot
{
	friend class PwFile;

protected:
	/** Constructor.
	  * @internal
	  */
	DataTree(PwFile *_pwfile);

public:
	/** Destructor. */
	~DataTree();

//TODO
	/** Get the entry at a specific "entry path".
	  * The "entry path" is the path of the entry in the
	  * whole DataTree. "/" is the DataRoot.
	  * @param path The path string.
	  * @return Returns true on success and false if the entry
	  *         could not be read.
	  */
//	bool getEntryAtPath(const QString &path);
	/** Store the given DataEntry in the given "entry path".
	  * @param entry The DataEntry to write at the path.
	  * @param path The path string.
	  * @return Returns true on success and false if the entry
	  *         could not be written.
	  */
//	bool setEntryAtPath(const QString &path, const DataEntry &entry)
};
#endif

/** This is the In Memory representation of an opened file. */
class PwFile : public QObject
{
	Q_OBJECT

	friend class LibPwManager;
	friend class DataRoot;

protected:
	/** Constructor
	  * @internal
	  */
	PwFile(LibPwManager *_lib);
	/** Open a file.
	  * @internal
	  */
	bool open(const QString &path);
	/** Populate our modified tree through DCOP.
	  * @internal
	  */
	void populateTree();
	/** Update the internal tree.
	  * @internal
	  */
	bool updateTree(const FileHeader_t &header,
			FileData_t &data);

public:
	/** Destructor */
	~PwFile();

	/** Check if the PwFile is dirty.
	  * @return Returns true, if there's dirty data in memory,
	  *         which is not yet sync()ed.
	  */
	bool isDirty() const;
	/** Synchronize the data on disk.
	  * @return Returns true on success, else false.
	  */
	bool sync();
	/** Get the file path.
	  * @return Returns the file path.
	  *         If there's no path available (not saved, yet),
	  *         it returns QString::null
	  */
	QString getPath();
	/** Set the file path to a new path name.
	  * @param newPath The new path.
	  */
	bool setPath(const QString &newPath);
	/** Get a pointer to the root element of the data tree.
	  * @return Returns a pointer to the root-element.
	  */
	DataRoot * getDataTree();
	/** Get last error.
	  * @return Returns the last error code. This is something
	  *         like "errno" in the libc.
	  */
	ErrCode getError();

protected slots:
	/** Data dirty slot.
	  * @internal
	  */
	void dataDirty_slot(int handle,
			    const QString &path,
			    const FileHeader_t &header,
			    const FileData_t &data);

protected:
	/** Last error code.
	  * @internal
	  */
	ErrCode lastErr;
	/** File path.
	  * @internal
	  */
	QString path;
	/** The root of all data categories.
	  * @internal
	  */
	DataRoot *root;
	/** If true, this PwFile contains unsynced data.
	  * @internal
	  */
	bool dirty;

	/** Data type of the internal data.
	  * @internal
	  */
	class PwFile_private;
	/** Internal data.
	  * @internal
	  */
	PwFile_private *p;
};

/** This is the main public PwManager interface class.
  *
  * Every application may use this class to access the PwManager database.
  * It is almost transparent, if the data is transfered over DCOP
  * or something else, to the application.
  *
  * @author Michael Buesch <mbuesch@freenet.de>
  * @short Public PwManager interface
  */
class LibPwManager : public QObject
{
	Q_OBJECT

	friend class PwFile;

protected:
	/** Constructor
	  * @internal
	  */
	LibPwManager();

public:
	/** Destructor */
	~LibPwManager();

	/** Initialize the LibPwManager interface.
	  * Initialize the interface only once and do all operations
	  * on the returned pointer.
	  * @return Returns a pointer to a new allocated
	  *         LibPwManager instance, if successfull.
	  *         You are responsible for deleting it, if you
	  *         don't use it any longer!
	  *         Returns a NULL pointer if an error occured.
	  */
	static LibPwManager * initialize();
	/** Convert error Code to a string.
	  * @param err Error code.
	  * @return Returns the last error as a human readable
	  *         and i18n()ed string.
	  */
	static QString getErrorStr(ErrCode err);


	/** Get last error.
	  * @return Returns the last error code. This is something
	  *         like "errno" in the libc.
	  */
	ErrCode getError();
	/** Create a new PwManagerFile.
	  * @return Returns a pointer to the document on success
	  *         and a NULL Pointer on failure.
	  *         You are responsible for deleting it, if you
	  *         don't use it any longer!
	  */
	PwFile * newFile();
	/** Open a file.
	  * @param path The path of the file to open, or the
	  *             plain filename, if you want to open the file
	  *             in the default document path.
	  * @return Returns a pointer to the document on success
	  *         and a NULL Pointer on failure.
	  *         You are responsible for deleting it, if you
	  *         don't use it any longer!
	  */
	PwFile * openFile(const QString &path);

	/** Get parent widget.
	  * @return Returns a pointer to the widget, which is
	  *         used as parent for all dialog windows.
	  *         Can be NULL.
	  */
	QWidget * getParentWidget() const;
	/** Set parent widget.
	  * @param w Pointer to the widget, which should be used as
	  *          parent for all dialog windows. NULL is default.
	  */
	void setParentWidget(QWidget *w);

protected:
	/** If some PwFile object wants to complain about the
	  * data it got via dataDirty_slot(), it calls this
	  * function. That means it is only valid to call this
	  * if we are in PwFile::dataDirty_slot() context.
	  * @internal
	  */
	void complain(int complainId);

protected:
	/** LibPwManager instance.
	  * @internal
	  */
	static LibPwManager *lib;
	/** Last error code.
	  * @internal
	  */
	ErrCode lastErr;
	/** Data type of the internal data.
	  * @internal
	  */
	class LibPwManager_private;
	/** Internal data.
	  * @internal
	  */
	LibPwManager_private *p;
};

} // namespace PwManager

#endif // LIB_PWMANAGER_H_
