/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "libpwmanager/libpwmanager.h"
#include "libpwmanager/serializer.h"
#include "libpwmanager/filecrypto.h"
#include "libpwmanager/filecompress.h"
#include "libpwmanager/datahash.h"
#include "libpwmanager/rawinterface.h"
#include "libpwmanager/privatedata.h"
#include "rawfile/rawfile.h"
#include "pwmexception.h"

#include <qcstring.h>

#include <kmessagebox.h>
#include <klocale.h>


namespace PwManager
{

class PwFile::PwFile_private
{
public:
	PwFile_private(LibPwManager *_lib,
		       PwFile *_pwfile,
		       RawInterface *_raw)
	 : handle (-1)
	 , lib (_lib)
	 , raw (_raw)
	 , pwfile (_pwfile)
		{ }

	/** display the "invalid file format" message. */
	void msg_invalidFileFormat(QWidget *parent);
	/** Request the Passphrase from the user. */
	bool requestPassphrase(QByteArray *p, const FileHeader_t &header);
	/** Request the user to enter a new Passphrase. */
	bool newPassphrase();
	/** check the passphrase for correctness. */
	bool checkPassphrase(const QByteArray &p, const FileHeader_t &header);
	/** decrypt the data */
	bool decrypt(QByteArray *data, const FileHeader_t &header);
	/** encrypt the data */
	bool encrypt(QByteArray *data, const FileHeader_t &header);
	/** decompress the data */
	bool decompress(QByteArray *data, const FileHeader_t &header);
	/** compress the data */
	bool compress(QByteArray *data, const FileHeader_t &header);
	/** check the data hash. */
	bool validateDataStream(const QByteArray &data,
				const FileHeader_t &header);
	/** close the handle / file. */
	void closeHandle();

public:
	/** Handle number for DCOP transfer. */
	int handle;
	/** Pointer to the main interface class. */
	LibPwManager *lib;
	/** Current header information. */
	FileHeader_t hdr;
	/** Current passphrase. */
	QByteArray passphrase;
	/** Filepath of the current file. */
	QString path;
	/** Raw data interface. */
	RawInterface *raw;

private:
	/** Pointer to the PwFile instance. */
	PwFile *pwfile;
};

void PwFile::PwFile_private::msg_invalidFileFormat(QWidget *parent)
{
	KMessageBox::error(parent,
			   i18n("This file is not a valid PwManager "
				"data-file, or it is corrupt."),
			   i18n("Invalid Format"));
}

bool PwFile::PwFile_private::requestPassphrase(QByteArray *p,
					       const FileHeader_t &header)
{
#if 0
	switch (header.mpwType) {
	case 0x00: {
		// normal passphrase
		GetPassphraseImpl wnd(path, parent);
		if (!wnd.exec()) {
			pwfile->lastErr = eUser;
			return false;
		}
		QString pf(wnd.getPassphrase());
		p->duplicate(pf.latin1(), pf.length());
		break;
	} case 0x01: {
		// chipcard (keycard)
		//TODO
		TODO();
		break;
	} default:
		msg_invalidFileFormat(parent);
		pwfile->lastErr = eFormat;
		return false;
	}
	return true;
#endif
}

bool PwFile::PwFile_private::checkPassphrase(const QByteArray &p,
					     const FileHeader_t &header)
{
#if 0
	DataHash h;
	QByteArray hash;
	if (!h.hash(p, &hash, header.keyHashId)) {
		msg_invalidFileFormat(parent);//FIXME: correct message?
		pwfile->lastErr = eFormat;
		return false;
	}
	if (hash != header.keyHash) {
		//TODO: message
		return false;
	}
	return true;
#endif
}

bool PwFile::PwFile_private::newPassphrase()
{
#if 0
	PWM_ASSERT(handle > 0);
	PWM_ASSERT(hdr.valid);
	PWM_ASSERT(path.length());
	if (!passphrase.isEmpty()) {
		// Let the user type in the old passphrase.
		QByteArray oldPass;
		if (!requestPassphrase(&oldPass, hdr))
			return false;
		if (oldPass != passphrase) {
			//TODO: message here!
			return false;
		}
	}

	SetPassphraseImpl wnd(path, parent);
	if (!wnd.exec())
		return false;
	QString pw(wnd.getPassphrase());
	passphrase.duplicate(pw.ascii(), pw.length());
	return true;
#endif
}

bool PwFile::PwFile_private::validateDataStream(const QByteArray &data,
						const FileHeader_t &header)
{
	DataHash h;
	QByteArray hash;
	if (!h.hash(data, &hash, header.dataHashId)) {
		msg_invalidFileFormat(lib->getParentWidget());//TODO: correct message?
		pwfile->lastErr = eFormat;
		return false;
	}
	if (hash != header.dataHash) {
		msg_invalidFileFormat(lib->getParentWidget());
		pwfile->lastErr = eFormat;
		return false;
	}
	return true;
}

bool PwFile::PwFile_private::decrypt(QByteArray *data,
				     const FileHeader_t &header)
{
	FileCrypto c;
	if (!c.decrypt(data,
		       passphrase,
		       header.cryptAlgoId,
		       header.dataHashId)) {
		msg_invalidFileFormat(lib->getParentWidget());
		pwfile->lastErr = eFormat;
		return false;
	}
	return true;
}

bool PwFile::PwFile_private::encrypt(QByteArray *data,
				     const FileHeader_t &header)
{
	FileCrypto c;
	if (!c.encrypt(data,
		       passphrase,
		       header.cryptAlgoId,
		       header.dataHashId)) {
		return false;
	}
	return true;
}

bool PwFile::PwFile_private::decompress(QByteArray *data,
					const FileHeader_t &header)
{
	FileCompress c;
	if (!c.decompress(data, header.comprAlgoId)) {
		msg_invalidFileFormat(lib->getParentWidget());
		pwfile->lastErr = eFormat;
		return false;
	}
	return true;
}

bool PwFile::PwFile_private::compress(QByteArray *data,
				      const FileHeader_t &header)
{
	FileCompress c;
	if (!c.compress(data, header.comprAlgoId))
		return false;
	return true;
}

void PwFile::PwFile_private::closeHandle()
{
	if (handle < 0)
		return;
	raw->closeDoc(handle);
	handle = -1;
}



PwFile::PwFile(LibPwManager *_lib)
 : lastErr (eUnknown)
 , dirty (false)
{
	p = new PwFile_private(_lib, this, _lib->p->raw);
	root = new DataRoot(this);
}

PwFile::~PwFile()
{
	/* We do NOT sync any data here. The LibPwManager user has
	 * to sync it by himself, before he deletes PwFile.
	 */
	PWM_ASSERT(root->writeProtected);
	PWM_ASSERT(!root->treeDirty);
	p->closeHandle();
	delete root;
	delete p;
}

ErrCode PwFile::getError()
{
	return lastErr;
}

QString PwFile::getPath()
{
	return path;
}

bool PwFile::setPath(const QString &newPath)
{
	if (path == newPath)
		return true;
	/* We do NOT sync here!
	 * The user must sync before he changes the path.
	 */
	path = newPath;
	p->closeHandle();
	if (!open(path)) {
		// lastErr is already set by open()
		return false;
	}
	return true;
}

bool PwFile::open(const QString &path)
{
	int ret;
	p->path = path;
	ret = p->raw->openDoc(path);
	if (ret < 0) {
		//TODO: get exact error.
cout << "raw e " << ret << endl;
		lastErr = eOpen;
		return false;
	}
	p->handle = ret;

	FileHeader_t header = p->raw->getHeader(p->handle);
	if (!header.valid) {
		lastErr = eIO;
		return false;
	}
	FileData_t data = p->raw->getData(p->handle);
	if (!data.valid) {
		lastErr = eIO;
		return false;
	}

	if (!updateTree(header, data)) {
		p->raw->closeDoc(p->handle);
		// lastErr is already set by updateTree()
		return false;
	}
	return true;
}

bool PwFile::isDirty() const
{
	return dirty;
}

bool PwFile::sync()
{
	if (p->handle < 0) {
		// file is not open / path not set.
		lastErr = eNoFilepath;
		return false;
	}
	PWM_ASSERT(path.length());
	if (p->passphrase.isEmpty()) {
		/* We don't have a passphrase, yet (This
		 * is the first time we sync this).
		 * Now request one from the user.
		 */
		if (!p->newPassphrase()) {
			lastErr = eUser;
			return false;
		}
		/* We didn't do a populateTree(), because
		 * we had no passphrase until now. So
		 * do it now. It's last chance and we must
		 * do it now.
		 */
		populateTree();
	}
	if (!p->raw->sync(p->handle)) {
		lastErr = eIO;
		return false;
	}
	return true;
}

DataRoot * PwFile::getDataTree()
{
	return root;
}

void PwFile::dataDirty_slot(int handle,
			    const QString &path,
			    const FileHeader_t &header,
			    const FileData_t &data)
{
	if (handle != p->handle)
		return; // this is not for us.
	PWM_ASSERT(path == p->path);
	PWM_ASSERT(header.valid);
	PWM_ASSERT(data.valid);
	/* We did already some sanity checks on the header
	 * and the data in the Callback class (DCOPCallback).
	 * Now check if something changed in the header, compared
	 * to the old header. For example the user may have
	 * changed the passphrase in another app.
	 */
	//TODO
	if (p->hdr.keyHashId != header.keyHashId ||
	    p->hdr.keyHash != header.keyHash) {
		/* New key hash. This means we have a
		 * new passphrase. Ask for it now!
		 */
		//TODO
	}
	FileData_t _data = data;
	if (!updateTree(header, _data))
		p->lib->complain(CB_COMPLAIN_CORRUPT);
}

void PwFile::populateTree()
{
	if (p->passphrase.isEmpty()) {
		/* We haven't got a passphrase. That means
		 * we haven't saved the doc, yet. So don't
		 * populate it, because it's not possible.
		 */
		printDebug("PwFile::populateTree(): no passphrase!");
		return;
	}
	BUG_ON(p->handle < 0);
	Serializer s;
	FileData_t data;
	if (!s.serialize(&data.data, *root)) {
		printWarn("Serializing failed!");
		return;
	}

	// now update the header.
	PWM_ASSERT(p->hdr.valid);
	PWM_ASSERT(p->hdr.magic.size() == qstrlen(PWM_FILE_MAGIC));
	PWM_ASSERT(!qstrcmp(p->hdr.magic.data(), PWM_FILE_MAGIC));
	PWM_ASSERT(p->hdr.verMajor == PWM_FILE_VER_MAJOR);
	PWM_ASSERT(RawFile::validHashId(p->hdr.keyHashId));
	PWM_ASSERT(RawFile::validHashId(p->hdr.dataHashId));
	PWM_ASSERT(RawFile::validCryptId(p->hdr.cryptAlgoId));
	PWM_ASSERT(RawFile::validComprId(p->hdr.comprAlgoId));
	PWM_ASSERT(p->hdr.reserved.size() == PWM_FILE_RESERVED_BYTES);
	PWM_ASSERT(QCString(p->hdr.reserved) ==
		   QCString().fill(0x00, PWM_FILE_RESERVED_BYTES));
	DataHash h;
	if (!h.hash(data.data, &p->hdr.dataHash, p->hdr.dataHashId)) {
		printWarn("Data hashing failed!");
		return;
	}
	if (!h.hash(p->passphrase, &p->hdr.keyHash, p->hdr.keyHashId)) {
		printWarn("Key hashing failed!");
		return;
	}

	if (!p->compress(&(data.data), p->hdr)) {
		printWarn("Compressing failed!");
		return;
	}
	if (!p->encrypt(&(data.data), p->hdr)) {
		printWarn("Encrypting failed!");
		return;
	}
	data.valid = 1;
	if (!p->raw->populateChanges(p->handle,
				     p->hdr,
				     data)) {
		printWarn("Populating changes failed!");
		return;
	}
}

bool PwFile::updateTree(const FileHeader_t &header,
			FileData_t &data)
{//TODO: remove debugging code.
//TODO: set all error codes correctly.
	PWM_ASSERT(header.valid);
	PWM_ASSERT(data.valid);

	bool needNewPassphrase;
	if (p->passphrase.isEmpty()) {
		needNewPassphrase = true;
	} else {
		if (!p->checkPassphrase(p->passphrase, header))
			needNewPassphrase = true;
		else
			needNewPassphrase = false;
	}
	if (needNewPassphrase) {
		QByteArray pass;
		if (!p->requestPassphrase(&pass, header)) {
cout << "req pw failed" << endl;
			return false;
		}
		if (!p->checkPassphrase(pass, header)) {
cout << "wrong pass" << endl;
			return false;
		}
		p->passphrase.duplicate(pass);
	}

	if (!p->decrypt(&(data.data), header)) {
cout << "decrypt failed" << endl;
		return false;
	}

	if (!p->decompress(&(data.data), header)) {
cout << "decompress failed" << endl;
		return false;
	}
	if (!p->validateDataStream(data.data, header)) {
cout << "invalid data" << endl;
		return false;
	}
	Serializer s;
	root->writeProtected = false;
//FIXME: don't write data directly here. Merge it.
return false;
	if (!s.deSerialize(data.data, root)) {
		//FIXME: make sure that no data is changed if we fail!
cout << "deserialize failed" << endl;
		root->writeProtected = true;
		return false;
	}
	root->writeProtected = true;
	root->treeDirty = false;
	p->hdr = header;
	return true;
}

} // namespace PwManager
