/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004, 2005  by Michael Buesch                     *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#ifndef LIBPWMANAGER_PWMEXCEPTION_H_
#define LIBPWMANAGER_PWMEXCEPTION_H_

#include "globalstuff.h"

/** compile BUG()checking code? (0/1)*/
#define COMPILE_BUGCHECKING	1


#if COMPILE_BUGCHECKING != 0
/* This is an internal function to reduce code-overhead
 * of the BUG(), WARN(), TOD0() and FiXME() macros. Please use
 * these macros instead of calling this function directly.
 */
void _pwmanagerFatal(unsigned int id,
		     const char *file,
		     int line,
		     const char *additional = 0);
extern bool _pwmanagerFatalInProgress;
#endif // COMPILE_BUGCHECKING != 0

/** Use PWM_ASSERT(condition) for debugging assertions.
  * "condition" is eaten up and replaced with a NOP
  * when debugging is disabled.
  *
  * PWM_ASSERT_NOEAT(condition) is the same as PWM_ASSERT(condition),
  * but it does _not_ eat up "condition" and ensures that
  * condition is always evaluated.
  */
#ifdef PWM_ASSERT
# undef PWM_ASSERT
#endif
#ifdef PWM_ASSERT_NOEAT
# undef PWM_ASSERT_NOEAT
#endif
#if defined(PWM_DEBUG) && COMPILE_BUGCHECKING != 0
# define PWM_ASSERT(x)	do {							\
			if (unlikely(!(x)))					\
				_pwmanagerFatal(4, __FILE__, __LINE__, #x);	\
			} while (0)
# define PWM_ASSERT_NOEAT(x)	do { PWM_ASSERT(x); } while (0)
#else // PWM_DEBUG
# define PWM_ASSERT(x)		do { } while (0)
# define PWM_ASSERT_NOEAT(x)	do { if (x) { /* nothing */ } } while (0)
#endif // PWM_DEBUG

/** Insert a BUG() into code paths which clearly show
  * a bug in the code and which should, under normal
  * circumstances, never execute.
  */
#ifdef BUG
# undef BUG
#endif
#if COMPILE_BUGCHECKING != 0
# define BUG()	do { _pwmanagerFatal(0, __FILE__, __LINE__); } while (0)
#else // COMPILE_BUGCHECKING != 0
# define BUG()	do { } while (0)
#endif // COMPILE_BUGCHECKING != 0

/** Use BUG_ON(condition) to print a bug-message if "condition"
  * is true. This is also enabled in non-debugging code.
  */
#ifdef BUG_ON
# undef BUG_ON
#endif
#if COMPILE_BUGCHECKING != 0
# define BUG_ON(x)	do { if (unlikely(x))	BUG(); } while (0)
#else // COMPILE_BUGCHECKING != 0
# define BUG_ON(x)	do { if (x) { /* nothing */ } } while (0)
#endif // COMPILE_BUGCHECKING != 0

/** Insert a WARN() into code-paths which should not
  * execute normally, but if they do it's non-fatal.
  */
#ifdef WARN
# undef WARN
#endif
#if COMPILE_BUGCHECKING != 0
# define WARN()	do { _pwmanagerFatal(1, __FILE__, __LINE__); } while (0)
#else // COMPILE_BUGCHECKING != 0
# define WARN()	do { } while (0)
#endif // COMPILE_BUGCHECKING != 0

/** Same as BUG_ON() but prints a warning-message */
#ifdef WARN_ON
# undef WARN_ON
#endif
#if COMPILE_BUGCHECKING != 0
# define WARN_ON(x)	do { if (unlikely(x))	WARN(); } while (0)
#else // COMPILE_BUGCHECKING != 0
# define WARN_ON(x)	do { if (x) { /* nothing */ } } while (0)
#endif // COMPILE_BUGCHECKING != 0

/** Insert this into code which is incomplete */
#ifdef TODO
# undef TODO
#endif
#if COMPILE_BUGCHECKING != 0
# define TODO()	do { _pwmanagerFatal(2, __FILE__, __LINE__); } while (0)
#else // COMPILE_BUGCHECKING != 0
# define TODO() do { } while (0)
#endif // COMPILE_BUGCHECKING != 0

/** Insert this into code which likely contains bugs */
#ifdef FIXME
# undef FIXME
#endif
#if COMPILE_BUGCHECKING != 0
# define FIXME()	do { _pwmanagerFatal(3, __FILE__, __LINE__); } while (0)
#else // COMPILE_BUGCHECKING != 0
# define FIXME()	do { } while (0)
#endif // COMPILE_BUGCHECKING != 0


// application exit codes.
#define PWM_EXIT_SUCCESS	0
#define PWM_EXIT_INITFAIL	1
// Note that the signum is added to PWM_EXIT_SIG
#define PWM_EXIT_SIG		40
#define PWM_EXIT_BUG		(PWM_EXIT_SIG + SIGILL)


/** PwM error codes */
enum PwMerror {
	e_success = 0,

	// file access errors
	e_filename,
	e_readFile,
	e_writeFile,
	e_openFile,
	e_accessFile, // permission error, etc...
	e_fileGeneric,
	e_alreadyOpen,

	// other file errors
	e_fileVer,
	e_fileFormat,		// format error
	e_unsupportedFormat,	// completely unsupported format
	e_setFilePointer,
	e_fileBackup,
	e_fileCorrupt,		// file data has correct format,
				// but is corrupt (checksum error, etc)

	// password errors
	e_wrongPw,
	e_getPw,
	e_weakPw,
	e_noPw,

	// action not implemented errors
	e_hashNotImpl,
	e_cryptNotImpl,

	// argument/parameter errors
	e_incompleteArg,
	e_invalidArg,

	// misc
	e_writeHeader,
	e_serializeDta,
	e_enc,
	e_entryExists,
	e_categoryExists,
	e_maxAllowedEntr,	// no more entries can be added.
	e_outOfMem,
	e_lock,			// error while (un)locking
	e_docNotSaved,		// doc wasn't saved to a file, yet.
	e_docIsEmpty,
	e_binEntry,
	e_normalEntry,

	e_generic
};

class QString;

void _pwmanagerPrintInfo(const QString &msg);
void _pwmanagerPrintWarn(const QString &msg);
void _pwmanagerPrintError(const QString &msg);

#ifdef PWM_DEBUG
  void _pwmanagerPrintDebug(const QString &msg);
# define printDebug(x)	_pwmanagerPrintDebug(x)
#else
# define printDebug(x)	do { } while (0)
#endif

#define printInfo(x)	_pwmanagerPrintInfo(x)
#define printWarn(x)	_pwmanagerPrintWarn(x)
#define printError(x)	_pwmanagerPrintError(x)

#endif // LIBPWMANAGER_PWMEXCEPTION_H_
