/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/ 

#include "libpwmanager/rawinterface.h"

#include <kapplication.h>

#include <stdint.h>

#ifdef ENABLE_RAW_DCOP
//# include "daemon/daemondcop_stub.h"
//# include "daemon/daemoninit.h"
# include <dcopclient.h>
#endif

#ifdef ENABLE_RAW_DIRECT
# include "randomizer/randomizer.h"
#endif


namespace PwManager
{


RawInterface::RawInterface()
 : accessType (noAccess)
#ifdef ENABLE_RAW_DCOP
 , dc (0)
 , stub (0)
#endif
#ifdef ENABLE_RAW_DIRECT
 , handleMap (0)
#endif
{
#ifdef ENABLE_RAW_DCOP
	connect(&dcopCallback, SIGNAL(dataDirty(int, const QString &,
				      const FileHeader_t &,
				      const FileData_t &)),
		this, SIGNAL(dataDirty(int, const QString &,
				       const FileHeader_t &,
				       const FileData_t &)));
#endif // ENABLE_RAW_DCOP
}

RawInterface::~RawInterface()
{
	cleanup();
#ifdef ENABLE_RAW_DCOP
	delete_ifnot_null(stub);
#endif
#ifdef ENABLE_RAW_DIRECT
	delete_ifnot_null(handleMap);
#endif
}

#ifdef ENABLE_RAW_DCOP
bool RawInterface::initDCOPClient()
{
	if (!dc) {
		if (kapp)
			dc = kapp->dcopClient();
		else
			dc = new DCOPClient;
	}
	if (dc->isRegistered())
		return true;
	if (dc->registerAs(qApp->argv()[0], true).isNull())
		return false;
	return true;
}

bool RawInterface::initDCOPStub()
{
	if (stub)
		return true;
	stub = new DaemonDcop_stub(PWM_DAEMON_DCOPAPPID,
				   PWM_DAEMON_DCOPOBJ);
	return true;
}
#endif // ENABLE_RAW_DCOP

void RawInterface::cleanup()
{
	switch (accessType) {
	case noAccess:
		return;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		PWM_ASSERT(dc);
		PWM_ASSERT(stub);
		stub->unregisterApp();
		break;
#endif // ENABLE_RAW_DCOP
#ifdef ENABLE_RAW_DIRECT
	case directAccess:
		delete_ifnot_null(handleMap);
		break;
#endif // ENABLE_RAW_DIRECT
	default:
		BUG();
	}
	accessType = noAccess;
#ifdef ENABLE_RAW_DCOP
	if (!kapp) {
		/* have to delete the dcop client,
		 * as we created it by ourself.
		 */
		PWM_ASSERT(dc);
		delete_and_null(dc);
	}
#endif // ENABLE_RAW_DCOP
}

bool RawInterface::setAccessType(DataAccess _accessType)
{
	cleanup();
	switch (_accessType) {
	case noAccess:
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		if (!initDCOPClient())
			return false;
		if (!initDCOPStub())
			return false;
		if (!stub->registerApp())
			return false;
		break;
#endif // ENABLE_RAW_DCOP
#ifdef ENABLE_RAW_DIRECT
	case directAccess:
		handleMap = new QMap<int, RawFile *>;
		break;
#endif // ENABLE_RAW_DIRECT
	default:
		return false;
	}
	accessType = _accessType;
	return true;
}

void RawInterface::complain(int complainId)
{
	PARAM_UNUSED(complainId);
	switch (accessType) {
	default:
	case noAccess:
	case directAccess:
		// nobody has to complain in direct access. fullstop.
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		dcopCallback.complain(complainId);
		break;
#endif
	}
}

int RawInterface::openDoc(const QString &document)
{
	switch (accessType) {
	default:
	case noAccess:
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		return stub->openDoc(document);
#endif // ENABLE_RAW_DCOP
#ifdef ENABLE_RAW_DIRECT
	case directAccess:
	//	RawFile
	//	if (raw->open(document))
	//		return -1;
		//TODO: handle
		break;
#endif // ENABLE_RAW_DIRECT
	}
	return -1;
}

//TODO: implement the rest DIRECT file access!

void RawInterface::closeDoc(int handle)
{
	switch (accessType) {
	default:
	case noAccess:
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		stub->closeDoc(handle);
		break;
#endif // ENABLE_RAW_DCOP
	case directAccess:
		//TODO
		break;
	}
}

FileHeader_t RawInterface::getHeader(int handle)
{
	switch (accessType) {
	default:
	case noAccess:
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		return stub->getHeader(handle);
#endif // ENABLE_RAW_DCOP
	case directAccess:
		//TODO
		return FileHeader_t();
	}
	return FileHeader_t();

}

FileData_t RawInterface::getData(int handle)
{
	switch (accessType) {
	default:
	case noAccess:
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		return stub->getData(handle);
#endif // ENABLE_RAW_DCOP
	case directAccess:
		//TODO
		return FileData_t();
	}
	return FileData_t();
}

bool RawInterface::populateChanges(int handle,
				   const FileHeader_t &header,
				   const FileData_t &data)
{
	switch (accessType) {
	default:
	case noAccess:
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		return stub->populateChanges(handle,
					     header,
					     data);
#endif // ENABLE_RAW_DCOP
	case directAccess:
		//TODO
		return false;
	}
	return false;
}

bool RawInterface::sync(int handle)
{
	switch (accessType) {
	default:
	case noAccess:
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		return stub->sync(handle);
#endif // ENABLE_RAW_DCOP
	case directAccess:
		//TODO
		return false;
	}
	return false;
}

bool RawInterface::isEnabled(DataAccess interface)
{
	switch (interface) {
	case noAccess:
		return true;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		if (!initDCOPClient())
			return false;
		if (!initDCOPStub())
			return false;
		return stub->isEnabled();
#endif
	case directAccess:
		return true;
	default:
		return false;
	}
}

QString RawInterface::getDefaultDocPath() const
{
	switch (accessType) {
	default:
	case noAccess:
		break;
#ifdef ENABLE_RAW_DCOP
	case dcopAccess:
		return stub->getDefaultDocPath();
#endif // ENABLE_RAW_DCOP
	case directAccess:
		//TODO
		return QString::null;
	}
	return QString::null;
}

#ifdef ENABLE_RAW_DIRECT
int RawInterface::genHandle()
{
	int ret;
	BUG_ON((sizeof(int) != 4) && (sizeof(int) != 8));
	const int intMin = (sizeof(int) == 4) ? INT32_MIN : INT64_MIN;
	Randomizer rnd;
	do {
		rnd >> ret;
		if (ret < 0) {
			if (unlikely(ret == intMin))
				continue;
			ret = -ret;
		}
	} while (handleMap->find(ret) != handleMap->end());
	return ret;
}
#endif // ENABLE_RAW_DIRECT

} // namespace PwManager

#ifdef ENABLE_RAW_DCOP
//# include "daemon/libpwmanager/rawinterface_skel.cpp"
#endif
#include "rawinterface.moc"
