/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/ 

#ifndef LIBPWMANAGER_RAWINTERFACE_H_
#define LIBPWMANAGER_RAWINTERFACE_H_

#include "pwmexception.h"
#include "rawfile/rawfile.h"

#include <qmap.h>

// XXX: always disable DCOP for now. We don't support it, yet.
#undef ENABLE_RAW_DCOP
#define ENABLE_RAW_DIRECT

#include "dcopcallback.h"


#if !defined(ENABLE_RAW_DCOP) && !defined(ENABLE_RAW_DIRECT)
# error "No raw file interface configured!"
#endif

#include <qobject.h>
#ifdef ENABLE_RAW_DCOP
# include <dcopobject.h>
#endif

/* Return values for the callback interfaces. */
/* No complain. */
#define CB_COMPLAIN_NONE	(0)
/* A sanity check failed on some input data. */
#define CB_COMPLAIN_INVAL	(-1)
/* The input data is corrupt (couldn't be decrypted, etc...). */
#define CB_COMPLAIN_CORRUPT	(-2)
/* The user complained manually. */
#define CB_COMPLAIN_USER	(-3)
/* You are not allowed to call this callback
 * (you are not the pwmanager-daemon)
 */
#define CB_COMPLAIN_PEER	(-4)



class DCOPClient;
class DaemonDcop_stub;

namespace PwManager
{

/** This is the Raw Data Stream interface class.
  * It provides a constant API for all kinds of
  * raw data access. (currently DCOP and DIRECT)
  */
class RawInterface : public QObject
{
	Q_OBJECT

public:
	enum DataAccess
	{
		noAccess,
		directAccess,
		dcopAccess
	};

public:
	RawInterface();
	virtual ~RawInterface();

	/** returns the access type. */
	DataAccess getAccessType()
			{ return accessType; }
	/** set the access type. */
	bool setAccessType(DataAccess _accessType);
	/** complain about data we got from RawInterface. */
	void complain(int complainId);

	/** open "document". */
	int openDoc(const QString &document);
	/** close the document with "handle". */
	void closeDoc(int handle);
	/** returns the header of an open document. */
	FileHeader_t getHeader(int handle);
	/** returns the encrypted data of an open document. */
	FileData_t getData(int handle);
	/** populate changed data to the daemon. */
	bool populateChanges(int handle,
			     const FileHeader_t &header,
			     const FileData_t &data);
	/** sync all populated changes to disk. */
	bool sync(int handle);
	/** interface enabled? */
	bool isEnabled(DataAccess interface);
	/** returns the default document path */
	QString getDefaultDocPath() const;

signals:
	/** data dirty signal.
	  * If this signal is emitted, the raw data is dirty,
	  * and any user of it should re-read and update it.
	  */
	void dataDirty(int handle,
		       const QString &path,
		       const FileHeader_t &header,
		       const FileData_t &data);

protected:
	/** cleanup anything, based on the current accessType. */
	void cleanup();

protected:
	/** current access type. */
	DataAccess accessType;

#ifdef ENABLE_RAW_DCOP
protected:
	/** set up the dcop connection. */
	bool initDCOPClient();
	/** set up the st} // namespace PwManagerub object. */
	bool initDCOPStub();

protected:
	/** DCOPClient object */
	DCOPClient *dc;
	/** DCOP stub object */
	DaemonDcop_stub *stub;
	/** DCOP callback interface */
	DCOPCallback dcopCallback;
#endif // ENABLE_RAW_DCOP

#ifdef ENABLE_RAW_DIRECT
protected:
	/** handle map for direct file access. */
	QMap<int, RawFile *> *handleMap;
	/** generate a handle. */
	int genHandle();
#endif // ENABLE_RAW_DIRECT
};

} // namespace PwManager

#endif // LIBPWMANAGER_RAWINTERFACE_H_
