/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "libpwmanager/serializer.h"
#include "libpwmanager/libpwmanager.h"
#include "pwmexception.h"

#include <qregexp.h>

#include <utility>


/* An example listing of generated complete XML code follows:

<p v="3">
    <s>//FIXME: Handle this server-side!
        <t1 n="target name">QDateTime ISO format</t1>
        <t2 n="target name">QDateTime ISO format</t2>
        ...
    </s>
    <c0 i="/">
        <c0 i="This is the category ID string (name)">
            <e0 i="This is the entry ID string (name)">
                <d0 i="This is the data field ID string (title)" t="0">Data field data</d0>
                ...
                <m>
                    <c>QDateTime ISO format</c>
                    <v>QDateTime ISO format</v>          
                    <e>QDateTime ISO format</e>
                    <u>QDateTime ISO format</u>
                    <i>4</i>
                    <n>UNIQUEID</n>
                </m>
            </e0>
	    ...
            <m>
                <c>QDateTime ISO format</c>
                <v>QDateTime ISO format</v>          
                <e>QDateTime ISO format</e>
                <u>QDateTime ISO format</u>
                <i>4</i>
                <n>UNIQUEID</n>
            </m>
        </c0>
	...
        <m>
            <c>QDateTime ISO format</c>
            <v>QDateTime ISO format</v>          
            <e>QDateTime ISO format</e>
            <u>QDateTime ISO format</u>
            <i>4</i>
            <n>UNIQUEID</n>
        </m>
    </c0>
</p>

*/


/* enable/disable serializer debugging (0/1) */
#define SERIALIZER_DEBUG	1
/* enable/disable base64 encoding (0/1) */
#define SERIALIZER_BASE64	0
/* base value used when converting integers to strings */
#define STRING_NUM_BASE		16


#define ROOT_MAGIC		"p"
#define VER_STR			"v"
#define VER_NR			"3"
#define CAT_PREFIX		"c"
#define CAT_ID			"i"
#define ENTRY_PREFIX		"e"
#define ENTRY_ID		"i"
#define ENTRY_FIELD_PREFIX	"d"
#define ENTRY_FIELD_ID		"i"
#define ENTRY_FIELD_TYPE	"t"
#define ENTRY_META		"m"
#define META_CREATE_DATE	"c"
#define META_VALID_DATE		"v"
#define META_EXPIRE_DATE	"e"
#define META_UPDATE_DATE	"u"
#define META_UPDATE_INT		"i"
#define META_UNIQUEID		"n"


#if SERIALIZER_DEBUG != 0
# warning "SERIALIZER_DEBUG enabled!"
# ifndef PWM_DEBUG
#  undef SERIALIZER_DEBUG
#  define SERIALIZER_DEBUG	0
#  warning "PWM_DEBUG not defined. Debugging disabled!"
# endif
#endif
#if SERIALIZER_DEBUG != 0
# define XML_ENDL		"\n"
#else
# define XML_ENDL		/* nothing */
#endif
#if SERIALIZER_BASE64 == 0
# warning "Base64 en/decoding disabled!"
#endif


using std::vector;
using std::string;
using std::map;


namespace PwManager
{

Serializer::Serializer()
 : tree (0)
{
}

Serializer::~Serializer()
{
}

bool Serializer::deSerialize(const QByteArray &buffer,
			     DataRoot *data)
{
	// parse the XML
#if SERIALIZER_DEBUG != 0
	printDebug("<BEGIN Serializer::deSerialize() dump>\n");
	cout << buffer << endl;
	printDebug("<END Serializer::deSerialize() dump>");
#endif // DEBUG

	vector<SimpleXmlParser::XmlElement> _tree;
	tree = &_tree;
	SimpleXmlParser xml;
	if (!xml.parseBuffer(string(buffer.data(), buffer.size()), tree)) {
		printDebug("Serializer: parse error");
		return false;
	}

	curRoot = data;
	// validate the root magic data.
	vector<SimpleXmlParser::XmlElement>::iterator i = tree->begin(),
						      end = tree->end();
	for ( ; i != end; ++i) {
		if ((*i).name == ROOT_MAGIC) {
			// found the root-magic.
			map<string, string>::iterator props;
			props = (*i).props.find(VER_STR);
			if (props == (*i).props.end()) {
				printDebug("Serializer: no version property");
				return false;
			}
			if ((*props).second != VER_NR) {
				printDebug("Serializer: wrong version");
				return false;
			}
			// now go through the category tree.
			vector<SimpleXmlParser::XmlElement>::iterator ci, cend;
			ci = (*i).children.begin();
			cend = (*i).children.end();
			for ( ; ci != cend; ++ci) {
				if (!readCategory(&(*ci), data))
					return false;
			}
			return true;
		}
	}
	printDebug("Serializer: ROOT_MAGIC not found");
	return false;
}

bool Serializer::readCategory(SimpleXmlParser::XmlElement *e,
			      DataItem *d)
{
	DataCategory *c = d->toCategory();
	PWM_ASSERT(c);

	map<string, string>::iterator prop;
	prop = e->props.find(CAT_ID);
	if (prop == e->props.end()) {
		printWarn("Serializer: No category id. Skipping...");
		return true;
	}
	decode(&(*prop).second);
	c->setName((*prop).second.c_str());

	QString nodeName;
	string tmp;
	vector<SimpleXmlParser::XmlElement>::iterator i = e->children.begin(),
						      end = e->children.end();
	for ( ; i != end; ++i) {
		nodeName = (*i).name.c_str();
		if (nodeName.contains(QRegExp("^" CAT_PREFIX "\\d+$")) == 1) {
			DataCategory newCat;
			if (!readCategory(&(*e), &newCat))
				return false;
			c->appendItem(newCat);
		} else if (nodeName.contains(QRegExp("^" ENTRY_PREFIX "\\d+$")) == 1) {
			DataEntry newEntry;
			if (!readEntry(&(*e), &newEntry))
				return false;
			c->appendItem(newEntry);
		} else if (nodeName == ENTRY_META) {
			if (!readMeta(&(*e), c->getMeta()))
				return false;
		} else
			printDebug("Serializer: not category, entry or meta");
	}
}

bool Serializer::readEntry(SimpleXmlParser::XmlElement *e,
			   DataItem *d)
{
	DataEntry *de = d->toEntry();
	PWM_ASSERT(de);

	map<string, string>::iterator prop;
	map<string, string>::iterator propEnd;
	prop = e->props.find(ENTRY_ID);
	if (prop == e->props.end()) {
		printWarn("Serializer: No entry id. Skipping...");
		return true;
	}
	decode(&(*prop).second);
	de->setName((*prop).second.c_str());

	// iterate through the data fields
	QString nodeName;
	vector<SimpleXmlParser::XmlElement>::iterator i = e->children.begin(),
						      end = e->children.end();
	for ( ; i != end; ++i) {
		nodeName = (*i).name.c_str();
		if (nodeName.contains(QRegExp("^" ENTRY_FIELD_PREFIX "\\d+$")) == 1) {
			DataEntry::FieldItem *newField = new DataEntry::FieldItem;
			propEnd = (*i).props.end();
			// read the ID property
			prop = (*i).props.find(ENTRY_FIELD_ID);
			if (prop == propEnd) {
				printDebug("Serializer: no field ID");
				delete newField;
				continue;
			}
			decode(&(*prop).second);
			newField->id = (*prop).second.c_str();

			// read the type property
			prop = (*i).props.find(ENTRY_FIELD_TYPE);
			if (prop == propEnd) {
				printDebug("Serializer: no field type");
				// fall back to "custom" type
			} else {
				// don't need to decode() here.
				bool ok = false;
				int type = QString((*prop).second.c_str()).toInt(&ok, STRING_NUM_BASE);
				if (!ok || type < 0 || type > 2)
					printDebug("Serializer: invalid field type");
				else
					newField->type = static_cast<DataEntry::FieldItem::type_t>(type);
			}

			// read the data
			decode(&(*i).content);
			newField->data.duplicate((*i).content.c_str(), (*i).content.size());
			de->fields.append(newField);
		} else if (nodeName == ENTRY_META) {
			if (!readMeta(&(*e), de->getMeta()))
				return false;
		} else
			printDebug("Serializer: not category, entry or meta");
	}
	return true;
}

bool Serializer::readMeta(SimpleXmlParser::XmlElement *e,
			  DataMeta *d)
{
	QString nodeName;
	QString nodeVal;
	vector<SimpleXmlParser::XmlElement>::iterator i = e->children.begin(),
						      end = e->children.end();
	for ( ; i != end; ++i) {
		nodeName = (*i).name.c_str();
		nodeVal = (*i).content.c_str();
		if (nodeVal.isEmpty()) {
			printDebug("Serializer: Meta node empty");
			continue;
		}
		if (nodeName == META_CREATE_DATE) {
			d->create = QDateTime::fromString(nodeVal, Qt::ISODate);
		} else if (nodeName == META_VALID_DATE) {
			d->valid = QDateTime::fromString(nodeVal, Qt::ISODate);
		} else if (nodeName == META_EXPIRE_DATE) {
			d->expire = QDateTime::fromString(nodeVal, Qt::ISODate);
		} else if (nodeName == META_UPDATE_DATE) {
			d->update = QDateTime::fromString(nodeVal, Qt::ISODate);
		} else if (nodeName == META_UPDATE_INT) {
			bool ok = false;
			d->updateInt = nodeVal.toULong(&ok, STRING_NUM_BASE);
			if (!ok) {
				printDebug("Serializer: Meta invalid updateInt value");
				// updateInt is already set to 0. So continue.
			}
		} else if (nodeName == META_UNIQUEID) {
		TODO();//TODO
#if 0
			bool ok = false;
			d->uniqueid = nodeVal.toUInt(&ok, STRING_NUM_BASE);
			if (!ok) {
				printDebug("Serializer: Meta no unique id. "
					   "Generating new uniqueid");
				TODO();
				//TODO: generate a new uniqueid
			}
#endif
		} else {
			printDebug(QString("Serializer: invalid meta: ")
				   + nodeName);
		}
	}
	return true;
}

bool Serializer::serialize(QByteArray *buffer,
			   const DataRoot &data)
{
	QCString _xmlData;
	xmlData = &_xmlData;

	tagBegin_root(VER_NR);
	tagBegin_cat("0", data.getName().ascii());
	if (!writeCategory(data))
		return false;
	if (!writeMeta(*(data.getMeta())))
		return false;
	tagEnd_cat("0");
	tagEnd_root();
	buffer->duplicate(_xmlData);

#if SERIALIZER_DEBUG != 0
	printDebug("<BEGIN Serializer::serialize() dump>\n");
	cout << _xmlData << endl;
	printDebug("<END Serializer::serialize() dump>");
#endif // DEBUG
	return true;
}

bool Serializer::writeCategory(const DataCategory &d)
{
	unsigned int catRef = 0;
	unsigned int entryRef = 0;
	DataCategory::const_iterator i = d.childBegin(),
				     end = d.childEnd();
	DataItem *item;
	for ( ; i != end; ++i) {
		item = *i;
		switch (item->getType()) {
		case DataItem::Entry: {
			DataEntry *entry = item->toEntry();
			PWM_ASSERT(entry);
			QCString ref(QString::number(entryRef++).ascii());
			tagBegin_entry(ref, entry->getName().ascii());
			if (!writeEntry(*entry))
				return false;
			if (!writeMeta(*(entry->getMeta())))
				return false;
			tagEnd_entry(ref);
			break;
		} case DataItem::Category: {
			DataCategory *cat = item->toCategory();
			PWM_ASSERT(cat);
			QCString ref(QString::number(catRef++).ascii());
			tagBegin_cat(ref, cat->getName().ascii());
			if (!writeCategory(*cat))
				return false;
			if (!writeMeta(*(cat->getMeta())))
				return false;
			tagEnd_cat(ref);
			break;
		} default:
			BUG();
			return false;
		}
	}
	return true;
}

bool Serializer::writeEntry(const DataEntry &d)
{
	unsigned int fieldRef = 0;
	DataEntry::const_iterator i = d.fieldsBegin(),
				  end = d.fieldsEnd();
	DataEntry::FieldItem *item;
	for ( ; i != end; ++i) {
		item = *i;
		tag_field(QString::number(fieldRef++).ascii(),
			  item->id.ascii(),
			  QString::number(item->type, STRING_NUM_BASE).ascii(),
			  encode(item->data));
	}
	return true;
}

bool Serializer::writeMeta(const DataMeta &d)
{//TODO
	tagBegin_meta();
	//TODO
	tagEnd_meta();
}

QCString Serializer::encode(const QByteArray &s)
{
#if SERIALIZER_BASE64 != 0
	//TODO
#else // SERIALIZER_BASE64
	return QCString(s.data(), s.size());
#endif // SERIALIZER_BASE64
}

void Serializer::decode(std::string *s)
{
#if SERIALIZER_BASE64 != 0
	//TODO
#else // SERIALIZER_BASE64
	PARAM_UNUSED(s);
#endif // SERIALIZER_BASE64
}

void Serializer::tagBegin_root(const QCString &version)
{
	*xmlData += "<" ROOT_MAGIC " " VER_STR "=\""
		 + version
		 + "\">" XML_ENDL;
}

void Serializer::tagBegin_cat(const QCString &refCnt,
			      const QCString &id)
{
	*xmlData += "<" CAT_PREFIX
		 + refCnt
		 + " " CAT_ID "=\""
		 + encode(id)
		 + "\">" XML_ENDL;
}

void Serializer::tagBegin_entry(const QCString &refCnt,
				const QCString &id)
{
	*xmlData += "<" ENTRY_PREFIX
		 + refCnt
		 + " " ENTRY_ID "=\""
		 + encode(id)
		 + "\">" XML_ENDL;
}

void Serializer::tag_field(const QCString &refCnt,
			   const QCString &id,
			   const QCString &type,
			   const QCString &content)
{
	*xmlData += "<" ENTRY_FIELD_PREFIX
		 + refCnt
		 + " " ENTRY_FIELD_ID "=\""
		 + encode(id)
		 + "\" " ENTRY_FIELD_TYPE "=\""
		 + type
		 + "\">"
		 + encode(content)
		 + "</" ENTRY_FIELD_PREFIX
		 + refCnt
		 + ">" XML_ENDL;
}

void Serializer::tagBegin_meta()
{
	*xmlData += "<" ENTRY_META ">" XML_ENDL;
}

void Serializer::tag_metaCreate(const QDateTime &date)
{
	if (!date.isValid())
		return;
	*xmlData += "<" META_CREATE_DATE ">"
		 + QCString(date.toString(Qt::ISODate).ascii())
		 + "</" META_CREATE_DATE ">" XML_ENDL;
}

void Serializer::tag_metaValid(const QDateTime &date)
{
	if (!date.isValid())
		return;
	*xmlData += "<" META_VALID_DATE ">"
		 + QCString(date.toString(Qt::ISODate).ascii())
		 + "</" META_VALID_DATE ">" XML_ENDL;
}

void Serializer::tag_metaExpire(const QDateTime &date)
{
	if (!date.isValid())
		return;
	*xmlData += "<" META_EXPIRE_DATE ">"
		 + QCString(date.toString(Qt::ISODate).ascii())
		 + "</" META_EXPIRE_DATE ">" XML_ENDL;
}

void Serializer::tag_metaUpdate(const QDateTime &date)
{
	if (!date.isValid())
		return;
	*xmlData += "<" META_UPDATE_DATE ">"
		 + QCString(date.toString(Qt::ISODate).ascii())
		 + "</" META_UPDATE_DATE ">" XML_ENDL;
}

void Serializer::tag_metaUpdateInt(unsigned long updateInt)
{
	*xmlData += "<" META_UPDATE_INT ">"
		 + QCString(QString::number(updateInt, STRING_NUM_BASE).ascii())
		 + "</" META_UPDATE_INT ">" XML_ENDL;
}

void Serializer::tag_metaUniqueId(unsigned int uniqueid)
{
	*xmlData += "<" META_UPDATE_INT ">"
		 + QCString(QString::number(uniqueid, STRING_NUM_BASE).ascii())
		 + "</" META_UPDATE_INT ">" XML_ENDL;

}

void Serializer::tagEnd_root()
{
	*xmlData += "</" ROOT_MAGIC ">" XML_ENDL;
}

void Serializer::tagEnd_cat(const QCString &refCnt)
{
	*xmlData += "</" CAT_PREFIX
		 + refCnt
		 + ">" XML_ENDL;
}

void Serializer::tagEnd_entry(const QCString &refCnt)
{
	*xmlData += "</" ENTRY_PREFIX
		 + refCnt
		 + ">" XML_ENDL;
}

void Serializer::tagEnd_meta()
{
	*xmlData += "</" ENTRY_META ">" XML_ENDL;
}


#if 0
bool Serializer::checkValid()
{
	QDomElement root(domDoc->documentElement());
	if (root.nodeName() != ROOT_MAGIC) {
		printDebug("Serializer: wrong magic");
		return false;
	}
	if (root.attribute(VER_STR) != VER_NR) {
		printDebug("Serializer: wrong version");
		return false;
	}
	return true;
}

QCString Serializer::getXml()
{
#if SERIALIZER_DEBUG != 0
	printDebug("<BEGIN Serializer::getXml() dump>\n");
	cout << domDoc->toCString(8) << endl;
	printDebug("<END Serializer::getXml() dump>");
#endif // DEBUG

	QCString ret(domDoc->toCString(0));
	ret.replace('\n', "");
	return ret;
}

bool Serializer::serialize(const DataRoot *data)
{
	QDomElement docRoot(domDoc->createElement(ROOT_MAGIC));
	docRoot.setAttribute(VER_STR, VER_NR);
	/* create the root-category and write all
	 * other subcategories recursively.
	 */
	QString elementId(CAT_PREFIX);
	elementId += "0";
	QDomElement element = domDoc->createElement(elementId);
	element.setAttribute(CAT_ID, data->getId());
	if (!writeCategory(data, &element))
		return false;
	QDomElement meta = domDoc->createElement(ENTRY_META);
	if (!writeMeta(data->getMeta_const(), &meta))
		return false;
	element.appendChild(meta);
	docRoot.appendChild(element);
	//TODO: write sync information here?
	domDoc->appendChild(docRoot);
	return true;
}

bool Serializer::deSerialize(DataRoot *data)
{
	QDomElement docRoot(domDoc->documentElement());
	QDomNode node(docRoot.firstChild());
	QString nodeName;
	bool haveRootCat = false;
	bool ret = false;

	data->sanityChecks(false);//FIXME: should we _really_ disable them?
	data->clear();
	for ( ; !node.isNull(); node = node.nextSibling()) {
		/* We have to search for the root category.
		 * Note that currently there can be only one
		 * root category.
		 */
		nodeName = node.nodeName();
		if (nodeName.contains(QRegExp("^" CAT_PREFIX "\\d+$")) == 1) {
			if (haveRootCat)
				continue;
			if (!readCategory(node, data, data)) {
				ret = false;
				goto out;
			}
			haveRootCat = true;
		}
		//TODO: check for sync information here?
	}
	if (haveRootCat)
		ret = true;
out:
	data->sanityChecks(true);
	return ret;
}

bool Serializer::readCategory(const QDomNode &n,
			      DataItem *data,
			      DataRoot *root)
{
	QDomNode node(n.firstChild());
	QDomElement element;
	QString nodeName, id;
	DataEntry newEntry;
	DataCategory newCategory;
	DataCategory *data_cat = data->toCategory();
	PWM_ASSERT(data_cat);

	for ( ; !node.isNull(); node = node.nextSibling()) {
		if (!node.isElement()) {
			printDebug("Serializer: node is not element");
			continue;
		}
		element = node.toElement();
		nodeName = node.nodeName();
		if (nodeName.contains(QRegExp("^" CAT_PREFIX "\\d+$")) == 1) {
			id = element.attribute(CAT_ID);
			if (id.isNull()) {
				printDebug("Serializer: no category id");
				continue;
			}
			newCategory.clear();
			newCategory.setId(id);
			if (!readCategory(node, &newCategory, root))
				return false;
			data_cat->appendItem(newCategory);
		} else if (nodeName.contains(QRegExp("^" ENTRY_PREFIX "\\d+$")) == 1) {
			id = element.attribute(ENTRY_ID);
			if (id.isNull()) {
				printDebug("Serializer: no entry id");
				continue;
			}
			newEntry.clear();
			if (!readEntry(node, &newEntry, root))
				return false;
			data_cat->appendItem(newEntry);
		} else if (nodeName == ENTRY_META) {
			if (!readMeta(node, data->getMeta()))
				return false;
		} else
			printDebug("Serializer: not category, entry or meta");
	}
	return true;
}

bool Serializer::readEntry(const QDomNode &n,
			   DataItem *data,
			   DataRoot *root)
{
	QDomNode node(n.firstChild());
	QDomElement element;
	QString nodeName, tmp;
	DataEntry::FieldItem *newField;
	DataEntry *data_entry = data->toEntry();
	PWM_ASSERT(data_entry);

	for ( ; !node.isNull(); node = node.nextSibling()) {
		if (!node.isElement()) {
			printDebug("Serializer: Entry not element");
			continue;
		}
		element = node.toElement();
		nodeName = node.nodeName();
		if (nodeName.contains(QRegExp("^" ENTRY_FIELD_PREFIX "\\d+$")) == 1) {
			newField = new DataEntry::FieldItem;
			tmp = element.attribute(ENTRY_FIELD_TYPE);
			if (!tmp.isEmpty()) {
				bool ok = false;
				int type = tmp.toInt(&ok, STRING_NUM_BASE);
				if (!ok || type < 0 || type > 2)
					printDebug("Serializer: invalid field type");
				else
					newField->type = static_cast<DataEntry::FieldItem::type_t>(type);
			}
			else
				printDebug("Serializer: empty or null field type");
			tmp = element.attribute(ENTRY_FIELD_ID);
			if (!tmp.isEmpty())
				newField->id = tmp;
			else
				printDebug("Serializer: empty or null field id");
			tmp = element.text();
			//TODO: base64
			newField->data.duplicate(tmp.ascii(), tmp.length());
			data_entry->fields.append(newField);
		} else if (nodeName == ENTRY_META) {
			if (!readMeta(node, data->getMeta()))
				return false;
		} else
			printDebug("Serializer: not entry field or meta");

	}
	return true;
}

bool Serializer::readMeta(const QDomNode &n,
			  DataMeta *data)
{
	QDomNode node(n.firstChild());
	QDomElement element;
	QString nodeName, val;

	data->clear();
	for ( ; !node.isNull(); node = node.nextSibling()) {
		if (!node.isElement()) {
			printDebug("Serializer: Meta node is not element");
			continue;
		}
		element = node.toElement();
		nodeName = node.nodeName();
		val = element.text();
		if (val.isEmpty()) {
			printDebug("Serializer: Meta empty or null node text");
			continue;
		}
		if (nodeName == META_CREATE_DATE) {
			data->create = QDateTime::fromString(val, Qt::ISODate);
		} else if (nodeName == META_VALID_DATE) {
			data->valid = QDateTime::fromString(val, Qt::ISODate);
		} else if (nodeName == META_EXPIRE_DATE) {
			data->expire = QDateTime::fromString(val, Qt::ISODate);
		} else if (nodeName == META_UPDATE_DATE) {
			data->update = QDateTime::fromString(val, Qt::ISODate);
		} else if (nodeName == META_UPDATE_INT) {
			bool ok = false;
			data->updateInt = val.toULong(&ok, STRING_NUM_BASE);
			if (!ok) {
				printDebug("Serializer: Meta invalid updateInt value");
				// updateInt is already set to 0. So continue.
			}
		} else if (nodeName == META_UNIQUEID) {
			bool ok = false;
			data->uniqueid = val.toUInt(&ok, STRING_NUM_BASE);
			if (!ok) {
				printDebug("Serializer: Meta no unique id. "
					   "Generating new uniqueid");
				TODO();
				//TODO: generate a new uniqueid
			}
		} else {
			printDebug(string("Serializer: invalid meta: ")
				   + nodeName.latin1());
		}
	}
	return true;
}

bool Serializer::writeCategory(const DataCategory *data,
			       QDomElement *e)
{
	unsigned int catCnt = 0;
	unsigned int entrCnt = 0;
	QString elementId;
	QDomElement element, meta;
	DataCategory::const_iterator i = data->childBegin(),
				     end = data->childEnd();
	DataItem *item;
	for ( ; i != end; ++i) {
		item = *i;
		switch (item->getType()) {
		case DataItem::Entry:
			elementId = ENTRY_PREFIX;
			elementId += QString::number(entrCnt++);
			element = domDoc->createElement(elementId);
			element.setAttribute(ENTRY_ID, data->getId());
			if (!writeEntry(item->toEntry(), &element))
				return false;
			break;
		case DataItem::Category:
			elementId = CAT_PREFIX;
			elementId += QString::number(catCnt++);
			element = domDoc->createElement(elementId);
			element.setAttribute(CAT_ID, data->getId());
			if (!writeCategory(item->toCategory(), &element))
				return false;
			break;
		default:
			BUG();
			return false;
		}
		meta = domDoc->createElement(ENTRY_META);
		if (!writeMeta(item->getMeta_const(), &meta))
			return false;
		element.appendChild(meta);
		e->appendChild(element);
	}
	return true;
}

bool Serializer::writeEntry(const DataEntry *data,
			    QDomElement *e)
{
	QDomText text;
	QDomElement element;
	QString elementId;
	unsigned int fieldCnt = 0;

	DataEntry::const_iterator i = data->fieldsBegin_const(),
				  end = data->fieldsEnd_const();
	DataEntry::FieldItem *item;
	for ( ; i != end; ++i) {
		item = *i;
		elementId = ENTRY_FIELD_PREFIX;
		elementId += QString::number(fieldCnt++);
		element = domDoc->createElement(elementId);
		element.setAttribute(ENTRY_FIELD_ID, item->id);
		element.setAttribute(ENTRY_FIELD_TYPE, QString::number(item->type));
		text = domDoc->createTextNode(item->data);//FIXME: base64
		element.appendChild(text);
		e->appendChild(element);
	}
	return true;
}

bool Serializer::writeMeta(const DataMeta *data,
			   QDomElement *e)
{
	QDomText text;
	QDomElement tag;
	QDateTime date;

	date = data->getCreateDate();
	if (date.isValid()) {
		tag = domDoc->createElement(META_CREATE_DATE);
		text = domDoc->createTextNode(date.toString(Qt::ISODate));
		tag.appendChild(text);
		e->appendChild(tag);
	}

	date = data->getValidDate();
	if (date.isValid()) {
		tag = domDoc->createElement(META_VALID_DATE);
		text = domDoc->createTextNode(date.toString(Qt::ISODate));
		tag.appendChild(text);
		e->appendChild(tag);
	}

	date = data->getExpireDate();
	if (date.isValid()) {
		tag = domDoc->createElement(META_EXPIRE_DATE);
		text = domDoc->createTextNode(date.toString(Qt::ISODate));
		tag.appendChild(text);
		e->appendChild(tag);
	}

	date = data->getUpdateDate();
	if (date.isValid()) {
		tag = domDoc->createElement(META_UPDATE_DATE);
		text = domDoc->createTextNode(date.toString(Qt::ISODate));
		tag.appendChild(text);
		e->appendChild(tag);
	}

	tag = domDoc->createElement(META_UPDATE_INT);
	text = domDoc->createTextNode(QString::number(data->getUpdateInt(), STRING_NUM_BASE));
	tag.appendChild(text);
	e->appendChild(tag);

	tag = domDoc->createElement(META_UNIQUEID);
	text = domDoc->createTextNode(QString::number(data->getUniqueId(), STRING_NUM_BASE));
	tag.appendChild(text);
	e->appendChild(tag);

	return true;
}
#endif

} // namespace PwManager
