/***************************************************************************
 *                                                                         *
 *  Simple XML Parser.                                                     *
 *     copyright (C)  Tim Hockin <thockin@hockin.org>                      *
 *                                                                         *
 *  Heavily modified / ported to C++                                       *
 *     copyright (C) 2004, 2005  Michael Buesch <mbuesch@freenet.de>       *
 *                                                                         *
 * Released under the LGPL version 2.1 (http://www.gnu.org)                *
 * This program is free software; you can redistribute it and/or modify    *
 * it under the terms of the GNU Lesser General Public License version 2.1 *
 * as published by the Free Software Foundation.                           *
 *                                                                         *
 ***************************************************************************/

#include "libpwmanager/simplexmlparser.h"

#include <ctype.h>
#include <utility>


#ifdef DPRINTF
# undef DPRINTF
#endif
#if SIMPLEXMLPARSER_DEBUG != 0
# warning "SimpleXmlParser debugging enabled!"
# include <stdio.h>
# define DPRINTF(f, a...)	fprintf(stderr, f, ##a)
#else
# define DPRINTF(f, a...)	do { } while (0)
#endif


using std::string;
using std::vector;
using std::map;
using std::pair;


bool SimpleXmlParser::isNameStart(char c)
{
	return (isalpha(c) || c == '_' || c == ':');
}

bool SimpleXmlParser::isName(char c)
{
	return (isNameStart(c) ||
		isdigit(c) ||
		c == '.' ||
		c == '-');
}

bool SimpleXmlParser::nextChar(char *c)
{
	if (bufPos >= bufLen)
		return false;
	*c = (*buffer)[bufPos++];
	return true;
}

bool SimpleXmlParser::parseBuffer(const string &buf,
				  vector<SimpleXmlParser::XmlElement> *ret)
{
	char c;
	XmlElement newElement;
	bool canProlog = true;

	ret->clear();
	setLineno(1);
	buffer = &buf;
	bufPos = 0;
	bufLen = buffer->size();

	while (nextChar(&c)) {
		if (c == '\n')
			incLineno();
		if (isspace(c))
			continue;
		else if (c == '<') {
			if (!nextChar(&c)) {
				DPRINTF("ERROR: premature EOF (line %d)\n", getLineno());
				return false;
			}
			/* handle a prolog ("<?") */
			if (c == '?' && canProlog) {
				prolog();
				continue;
			}
			canProlog = false;

			/* handle a comment ("<!") */
			if (c == '!') {
				comment();
				continue;
			}
			if (c == '\n')
				incLineno();

			if (!parseElement(c, &newElement))
				return false;
			ret->push_back(newElement);
		} else {
			DPRINTF("invalid char '%c' in state top (line %d)\n",
				c, getLineno());
		}
	}
	return true;
}

void SimpleXmlParser::saveProp(XmlElement *e,
			       string *propname,
			       string *propval)
{
	map<string, string>::iterator i = e->props.find(*propname);
	if (i != e->props.end())
		e->props.erase(i);
	e->props.insert(pair<string, string>(*propname, *propval));
	propname->clear();
	propval->clear();
}

bool SimpleXmlParser::parseElement(char first,
				   XmlElement *element)
{
	char c;
	state_t state = ST_TAGSTART;
	XmlElement newElement;
	string propname;
	string propval;
	string endName;

	element->name.clear();
	element->content.clear();
	element->props.clear();
	element->children.clear();

	/* we got a '<' */
	if (isNameStart(first)) {
		element->name = tolower(first);
		state = ST_START_STAG;
	} else
		state = ST_TAGSTART;

	while (nextChar(&c)) {
		if (c == '\n')
			incLineno();
		switch (state) {
			/* start */
			case ST_TAGSTART:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (isNameStart(c)) {
					element->name += tolower(c);
					state = ST_START_STAG;
				} else
					goto out_inval_char;
				break;
			/* we got the start of a keyword for an stag */
			case ST_START_STAG:
				if (isName(c)) {
					element->name += tolower(c);
					// state = CURRENT_STATE;
				} else if (isspace(c))
					state = ST_STAG_NAME_DONE;
				else if (c == '>')
					state = ST_END_STAG;
				else if (c == '/')
					state = ST_END_SETAG;
				else
					goto out_inval_char;
				break;
			/* keyword done - look for next prop or other */
			case ST_STAG_NAME_DONE:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (c == '>')
					state = ST_END_STAG;
				else if (c == '/')
					state = ST_END_SETAG;
				else if (isNameStart(c)) {
					/* must be a propname */
					propname += tolower(c);
					state = ST_PROPNAME;
				} else
					goto out_inval_char;
				break;
			/* found start of a propname */
			case ST_PROPNAME:
				if (isName(c)) {
					propname += tolower(c);
					// state = CURRENT_STATE;
				} else if (c == '=')
					state = ST_GOTEQ;
				else if (isspace(c))
					state = ST_GETEQ;
				else
					goto out_inval_char;
				break;
			/* got propname, get '=' */
			case ST_GETEQ:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (c == '=')
					state = ST_GOTEQ;
				else
					goto out_inval_char;
				break;
			/* got '=' - find propval */
			case ST_GOTEQ:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (c == '"')
					state = ST_PROPVAL;
				else if (c == '\'')
					state = ST_SQ_PROPVAL;
				else
					goto out_inval_char;
				break;
			/* got a quoted propval */
			case ST_PROPVAL:
				if (c == '&') {
					char lc = unescape();
					if (lc > 0)
						propval += lc;
				} else if (c == '"') {
					saveProp(element, &propname, &propval);
					state = ST_STAG_NAME_DONE;
				} else if (c == '<')
					goto out_inval_char;
				else {
					propval += c;
					// state = CURRENT_STATE;
				}
				break;
			/* got a single-quoted propval */
			case ST_SQ_PROPVAL:
				if (c == '&') {
					char lc = unescape();
					if (lc > 0)
						propval += lc;
				} else if (c == '\'') {
					saveProp(element, &propname, &propval);
					state = ST_STAG_NAME_DONE;
				} else if (c == '<') {
					/* spec says this is an error */
					goto out_inval_char;
				} else {
					propval += c;
					// state = CURRENT_STATE;
				}
				break;
			/* spec says we must have a space between props */
			case ST_POST_PROPVAL:
				if (isspace(c))
					state = ST_STAG_NAME_DONE;
				else if (c == '>')
					state = ST_END_STAG;
				else if (c == '/')
					state = ST_END_SETAG;
				else
					goto out_inval_char;
				break;
			/* got a '/', suggesting an empty element */
			case ST_END_SETAG:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (c == '>')
					goto out_success;
				else
					goto out_inval_char;
				break;
			/* time to look for content */
			case ST_END_STAG:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (c == '&') {
					char lc = unescape();
					if (lc > 0)
						element->content += lc;
					state = ST_CONTENT;
				} else if (c == '<')
					state = ST_MAYBE_COMMENT;
				else {
					element->content += c;
					state = ST_CONTENT;
				}
				break;
			/* processing content */
			case ST_CONTENT:
				if (isspace(c)) {
					element->content += c;
					// state = CURRENT_STATE;
				} else if (c == '&') {
					char lc = unescape();
					if (lc > 0)
						element->content += lc;
				} else if (c == '<')
					state = ST_MAYBE_COMMENT;
				else {
					element->content += c;
					// state = CURRENT_STATE;
				}
				break;
			/* got a '<' - check for a comment */
			case ST_MAYBE_COMMENT:
				if (c == '!')
					comment();
				else if (isspace(c))
					state = ST_NEWTAG_START;
				else if (c == '/')
					state = ST_START_ETAG;
				else {
					if (!parseElement(c, &newElement))
						return false;
					element->children.push_back(newElement);
					state = ST_CONTENT;
				}
				break;
			/* got a '<' in content - figure it out */
			case ST_NEWTAG_START:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (c == '/')
					state = ST_START_ETAG;
				else {
					if (!parseElement(c, &newElement))
						return false;
					element->children.push_back(newElement);
					state = ST_CONTENT;
				}
				break;
			/* got a </ */
			case ST_START_ETAG:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (isNameStart(c)) {
					endName += c;
					state = ST_ETAG_NAME;
				} else
					goto out_inval_char;
				break;
			/* reading all of </name */
			case ST_ETAG_NAME:
				if (isName(c)) {
					endName += c;
					// state = CURRENT_STATE;
				} else if (isspace(c))
					state = ST_END_ETAG;
				else if (c == '>') {
					if (!strcasecmp(element->name.c_str(), endName.c_str()))
						goto out_success;
					else {
						/* not a match! */
						DPRINTF("endName (%s) != element->name (%s)\n",
							endName.c_str(), element->name.c_str());
						state = ST_CONTENT;
						endName.clear();
					}
				} else
					goto out_inval_char;
				break;
			/* just get '>' and wrap up */
			case ST_END_ETAG:
				if (isspace(c)) {
					// state = CURRENT_STATE;
				} else if (c == '>') {
					if (!strcasecmp(element->name.c_str(), endName.c_str()))
						goto out_success;
					else {
						/* not a match! */
						DPRINTF("endName (%s) != element->name (%s)\n",
							endName.c_str(), element->name.c_str());
						state = ST_CONTENT;
						endName.clear();
					}
				} else
					goto out_inval_char;
				break;
		}
	}
	DPRINTF("ERROR: premature EOF (line %d)\n", getLineno());
	return false;

out_inval_char:
	DPRINTF("invalid char '%c' in state %d (line %d)\n",
		c, state, getLineno());
	return false;

out_success:
	return true;
}

void SimpleXmlParser::prolog()
{
	char cur = ' ';
	char prev = ' ';

	while (nextChar(&cur)) {
		if (cur == '\n')
			incLineno();
		else if (cur == '>' && prev == '?')
			return;
		prev = cur;
	}
}

void SimpleXmlParser::comment()
{
	char cur = ' ';
	char prev = ' ';
	char prev2 = ' ';

	while (nextChar(&cur)) {
		if (cur == '\n')
			incLineno();
		else if (cur == '>' && prev == '-' && prev2 == '-')
			return;
		prev2 = prev;
		prev = cur;
	}
}

char SimpleXmlParser::unescape()
{
	char buf[16];
	unsigned int i = 0;
	const char *escapes[] = {
		"lt;",
		"gt;",
		"amp;",
		"apos;",
		"quot;",
		"nbsp;",
		NULL
	};
	const char esc_vals[] = {
		'<',
		'>',
		'&',
		'\'',
		'"',
		' ',
		'\0'
	};

	/* we assume we got a '&' before we got here */
	if (!nextChar(buf + i)) {
		DPRINTF("ERROR: premature EOF (line %d)\n", getLineno());
		return -1;
	}
	if (buf[i] == '\n')
		incLineno();
	while ((buf[i] != ';') && (!isspace(buf[i])) && (i < sizeof(buf)-1)) {
		i++;
		if (!nextChar(buf + i)) {
			DPRINTF("ERROR: premature EOF (line %d)\n", getLineno());
			return -1;
		}
		if (buf[i] == '\n')
			incLineno();
	}
	buf[i + 1] = '\0';

	/* got an escape */
	if (buf[i] == ';') {
		i = 0;
		/* got a &# style escape */
		if (buf[0] == '#') {
			char charesc = -1;
			int base = 10;
			i++;
			if (tolower(buf[i]) == 'x') {
				base = 16;
				i++;
			}
			while (buf[i] != ';') {
				int decval;

				/* convert buf[i]) */
				if (isdigit(buf[i])) {
					decval = buf[i] - '0';
				} else if (isxdigit(buf[i])) {
					decval = (tolower(buf[i]) - 'a') + 10;
				} else {
					return -1;
				}
				charesc *= base;
				charesc += decval;
			}
			return charesc;
		}
		/* got a 'normal' escape */
		while (escapes[i]) {
			if (!strcasecmp(buf, escapes[i])) {
				return esc_vals[i];
			}
			i++;
		}
	}

	return -1;
}

#if SIMPLEXMLPARSER_DEBUG != 0
string SimpleXmlParser::dumpElements(const vector<XmlElement> &e,
				     int indent)
{
	string ret;
	int j;
	vector<XmlElement>::const_iterator i = e.begin(),
					   end = e.end();
	map<string, string>::const_iterator map_i, map_end;
	for ( ; i != end; ++i) {
		for (j = 0; j < indent; j++)
			ret += "  ";
		ret += "\"";
		ret += (*i).name;
		ret += "\" => \"";
		ret += (*i).content;
		ret += "\"\n";
		map_i = (*i).props.begin();
		map_end = (*i).props.end();
		for ( ; map_i != map_end; ++ map_i) {
			for (j = 0; j < indent; j++)
				ret += "  ";
			ret += "property: \"";
			ret += (*map_i).first;
			ret += "\" \"";
			ret += (*map_i).second;
			ret += "\"\n";
		}
		ret += dumpElements((*i).children, indent + 1);
	}
	return ret;
}
#endif // SIMPLEXMLPARSER_DEBUG
