/***************************************************************************
 *                                                                         *
 *   copyright (C) 2005 by Michael Buesch                                  *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "libpwmanager/uniqueid.h"
#include "libpwmanager/libpwmanager.h"
#include "randomizer/randomizer.h"
#include "pwmexception.h"


using namespace PwManager;


UniqueId::UniqueId(UniqueIdContext *_context)
 : context (_context)
 , id (0)
{
}

UniqueId::UniqueId(const UniqueId &copySource)
{
	context = copySource.context;
	id = copySource.id;
}

UniqueId::~UniqueId()
{
}

UniqueId & UniqueId::operator= (const UniqueId &copySource)
{
	context = copySource.context;
	id = copySource.id;
	return *this;
}

QByteArray UniqueId::getId() const
{
	if (!id)
		return QByteArray();
	return *id;
}

bool UniqueId::generate()
{
	QByteArray *oldId = id;
	id = context->get();
	if (!id)
		return false;
	if (oldId)
		context->put(oldId);
	return true;
}

bool UniqueId::assign(const QByteArray &idData)
{
	QByteArray *oldId = id;
	QByteArray *d = new QByteArray;
	d->duplicate(idData);
	bool ret = context->assign(d);
	if (!ret)
		delete d;
	id = d;
	if (oldId)
		context->put(oldId);
	return ret;
}


UniqueIdContext::UniqueIdContext(unsigned int _length)
 : length (_length)
{
}

UniqueIdContext::~UniqueIdContext()
{
}

QByteArray * UniqueIdContext::get()
{//TODO: check if set is full. Also check if we are still able to generate an unique ID given the ID-length.
	unsigned int setSize = idSet.size();
	if (setSize >= idSet.max_size()) {
		/* can't insert more elements, even if
		 * there are still unused IDs.
		 */
		return 0;
	}

	id_t newid;
	newid.id->resize(length);
	Randomizer rnd;
	std::set<id_t>::iterator setEnd = idSet.end();
	std::set<id_t>::iterator trashEnd = idTrash.end();
	do {
		rnd >> *newid.id;
	} while (idTrash.find(newid) != trashEnd ||
		 !idSet.insert(newid).second);
	return newid.id;
}

bool UniqueIdContext::assign(QByteArray *d)
{
	id_t id(d);
	if (idTrash.find(id) != idTrash.end())
		return false;
	bool ok = idSet.insert(id).second;
	return ok;
}

void UniqueIdContext::put(QByteArray *d)
{
	std::set<id_t>::iterator i;
	i = idSet.find(id_t(d));
	PWM_ASSERT(i != idSet.end());
	bool ok = idTrash.insert(*i).second;
	PWM_ASSERT(ok);
	idSet.erase(i);
	/* Do not delete the QByteArray here.
	 * It's deleted when we clear the trash.
	 */
}

void UniqueIdContext::clearTrash()
{
	std::set<id_t>::iterator i = idTrash.begin(),
				 end = idTrash.end();
	for ( ; i != end; ++i)
		delete (*i).id;
	idTrash.clear();
}
