/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004, 2005  by Michael Buesch                     *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "mainwindow.h"
#include "pwmexception.h"
#include "pwminit.h"
#include "pwmprint.h"
#include "configuration.h"

#include <kpopupmenu.h>
#include <ktoolbar.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <khtml_part.h>
#include <khtmlview.h>

#include <qlabel.h>

#ifdef CONFIG_PWMANAGER_SMARTCARD
# include "masterkey/smartkey/smartkey.h"
#endif


// Default window size
#define DEFAULT_SIZE		(QSize(700, 400))
// Relative path to the splash screen HTML file
#define SPLASH_PATH		"pwmanager/html/splash.html"


/* The following enums define the Button-IDs for
 * the menu bar and tool bar buttons and sub-menus.
 */
enum {
	BUTTON_POPUP_FILE_NEW = 0,
	BUTTON_POPUP_FILE_OPEN,
	BUTTON_POPUP_FILE_CLOSE,
	BUTTON_POPUP_FILE_LOCKDOC,
	BUTTON_POPUP_FILE_SAVE,
	BUTTON_POPUP_FILE_SAVEAS,
	BUTTON_POPUP_FILE_EXPORT,
	BUTTON_POPUP_FILE_IMPORT,
	BUTTON_POPUP_FILE_PRINT,
	BUTTON_POPUP_FILE_QUIT,
};
enum {
	BUTTON_POPUP_MANAGE_CHMASTERKEY = 0,
};
enum {
	BUTTON_POPUP_SMARTCARD_GETID = 0,
	BUTTON_POPUP_SMARTCARD_GEN,
	BUTTON_POPUP_SMARTCARD_ERASE,
	BUTTON_POPUP_SMARTCARD_TOFILE,
	BUTTON_POPUP_SMARTCARD_FROMFILE,
};
enum {
	BUTTON_POPUP_VIEW_FIND = 0,
};
enum {
	BUTTON_POPUP_TOOLS_PWGEN = 0,
};
enum {
	BUTTON_POPUP_SETTINGS_CONFIG = 0,
};
enum {
	BUTTON_POPUP_EXPORT_TEXT = 0,
	BUTTON_POPUP_EXPORT_GPASMAN,
	BUTTON_POPUP_EXPORT_PWSAFE,
	BUTTON_POPUP_EXPORT_CSV,
#ifdef CONFIG_KWALLETIF
	BUTTON_POPUP_EXPORT_KWALLET,
#endif
};
enum {
	BUTTON_POPUP_IMPORT_TEXT = 0,
	BUTTON_POPUP_IMPORT_GPASMAN,
	BUTTON_POPUP_IMPORT_PWSAFE,
	BUTTON_POPUP_IMPORT_CSV,
#ifdef CONFIG_KWALLETIF
	BUTTON_POPUP_IMPORT_KWALLET,
#endif
};
enum {
	BUTTON_TOOL_NEW = 0,
	BUTTON_TOOL_OPEN,
	BUTTON_TOOL_SAVE,
	BUTTON_TOOL_SAVEAS,
	BUTTON_TOOL_PRINT,
	BUTTON_TOOL_LOCKDOC,
};


MainWindowSplash::MainWindowSplash(QWidget *parent,
				   const char *name,
				   WFlags f)
 : QWidget (parent, name, f)
 , html (0)
 , alt (0)
{
	QString splashPath(::locate("data", SPLASH_PATH));
	if (splashPath.isEmpty()) {
		QString noDocString(i18n("(No Document. Click \"File / New\" to create an empty Document.)"));
		QLabel *label = new QLabel(QString("<QT>"
						   "<H1> " PROG_NAME " " PACKAGE_VER "</H1><BR>"
						   "<SMALL>")
						 + noDocString
						 + "</SMALL>"
						   "</QT>",
					   this);
		alt = label;
		label->setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
		return;
	}
	html = new KHTMLPart(this);
	html->openURL(splashPath);
}

MainWindowSplash::~MainWindowSplash()
{
	delete_ifnot_null(html);
}

void MainWindowSplash::resizeEvent(QResizeEvent *)
{
	if (html)
		html->view()->resize(size());
	else if (alt)
		alt->resize(size());
}


QPtrList<MainWindow> MainWindow::list;

MainWindow::MainWindow(QWidget *parent,
		       const char *name,
		       WFlags f)
 : KMainWindow (parent, name, f)
 , central (0)
{
	list.append(this);
	initMenuBar();
	initToolBar();
	initMetrics();
	statusBar()->show();
	statusBarMsg(i18n("Ready."));
}

MainWindow::~MainWindow()
{
	list.remove(this);
	delete_ifnot_null(central);
}

bool MainWindow::initialize(PwMDoc *doc,
			    const QString &loadFile,
			    bool showWndMinimized)
{
	if (!doc && loadFile.isEmpty()) {
		initSplash();
	} else {
		initView(doc);
		if (!loadFile.isEmpty()) {
			if (!doc->load(loadFile, this))
				initSplash();
		}
	}
	updateCaption();
	if (showWndMinimized)
		showMinimized();
	else
		show();
	return true;
}

void MainWindow::initView(PwMDoc *doc)
{
	delete_ifnot_null(central);
	if (!doc)
		doc = new PwMDoc(pwminit);
	connect(doc, SIGNAL(dataChanged(PwMDoc *)),
		this, SLOT(updateButtons()));
	PwMView *view = new PwMView(doc, this);
	central = view;
	setCentralWidget(central);
	view->show();
	updateButtons();
}

void MainWindow::initSplash()
{
	delete_ifnot_null(central);
	central = new MainWindowSplash(this);
	setCentralWidget(central);
	central->show();
	updateButtons();
}

void MainWindow::addMButton(QPopupMenu *menu, int id,
			    const char *member, const QString &text,
			    const char *icon)
{
	if (icon) {
		menu->insertItem(QIconSet(iconLoader.loadIcon(icon, KIcon::Small)),
				 text, this, member, 0, id);
	} else
		menu->insertItem(text, this, member, 0, id);
}

void MainWindow::initMenuBar()
{
	KIconLoader icons;
	KMenuBar *mb = menuBar();

	filePopup = new KPopupMenu(this);
	importPopup = new KPopupMenu(filePopup);
	exportPopup = new KPopupMenu(filePopup);
	managePopup = new KPopupMenu(this);
#ifdef CONFIG_PWMANAGER_SMARTCARD
	smartcardPopup = new KPopupMenu(this);
#endif
	viewPopup = new KPopupMenu(this);
	toolsPopup = new KPopupMenu(this);
	settingsPopup = new KPopupMenu(this);

// "file" popup menu
	addMButton(filePopup, BUTTON_POPUP_FILE_NEW,
		   SLOT(new_slot()),
		   i18n("&New"), "filenew");
	addMButton(filePopup, BUTTON_POPUP_FILE_OPEN,
		   SLOT(open_slot()),
		   i18n("&Open..."), "fileopen");
	addMButton(filePopup, BUTTON_POPUP_FILE_CLOSE,
		   SLOT(close_slot()),
		   i18n("&Close"), "fileclose");
	addMButton(filePopup, BUTTON_POPUP_FILE_LOCKDOC,
		   SLOT(lockDocument_slot()));
	filePopup->insertSeparator();
	addMButton(filePopup, BUTTON_POPUP_FILE_SAVE,
		   SLOT(save_slot()),
		   i18n("&Save"), "filesave");
	addMButton(filePopup, BUTTON_POPUP_FILE_SAVEAS,
		   SLOT(saveAs_slot()),
		   i18n("Save &As..."), "filesaveas");
	filePopup->insertSeparator();
	// "file/export" popup menu
	addMButton(exportPopup, BUTTON_POPUP_EXPORT_TEXT,
		   SLOT(exportText()),
		   i18n("&Text file..."), "fileexport");
	addMButton(exportPopup, BUTTON_POPUP_EXPORT_GPASMAN,
		   SLOT(exportGpasman()),
		   i18n("&GPasman / KPasman..."), "fileexport");
	addMButton(exportPopup, BUTTON_POPUP_EXPORT_PWSAFE,
		   SLOT(exportPwSafe()),
		   i18n("&PwSafe and compatible..."), "fileexport");
	addMButton(exportPopup, BUTTON_POPUP_EXPORT_CSV,
		   SLOT(exportCsv()),
		   i18n("&Comma-Separated Values (CSV)..."), "fileexport");
#ifdef CONFIG_KWALLETIF
	addMButton(exportPopup, BUTTON_POPUP_EXPORT_KWALLET,
		   SLOT(exportKWallet()),
		   i18n("&KWallet..."), "fileexport");
#endif
	filePopup->insertItem(QIconSet(iconLoader.loadIcon("fileexport", KIcon::Small)),
			      i18n("E&xport"), exportPopup,
			      BUTTON_POPUP_FILE_EXPORT);
	// "file/import" popup menu
	addMButton(importPopup, BUTTON_POPUP_IMPORT_TEXT,
		   SLOT(importText()),
		   i18n("&Text file..."), "fileimport");
	addMButton(importPopup, BUTTON_POPUP_IMPORT_GPASMAN,
		   SLOT(importGpasman()),
		   i18n("&GPasman / KPasman..."), "fileimport");
	addMButton(importPopup, BUTTON_POPUP_IMPORT_PWSAFE,
		   SLOT(importPwSafe()),
		   i18n("&PwSafe and compatible..."), "fileimport");
	addMButton(importPopup, BUTTON_POPUP_IMPORT_CSV,
		   SLOT(importCsv()),
		   i18n("&Comma-Separated Values (CSV)..."), "fileimport");
#ifdef CONFIG_KWALLETIF
	addMButton(importPopup, BUTTON_POPUP_IMPORT_KWALLET,
		   SLOT(importKWallet()),
		   i18n("&KWallet..."), "fileimport");
#endif
	filePopup->insertItem(QIconSet(iconLoader.loadIcon("fileimport", KIcon::Small)),
			      i18n("I&mport"), importPopup,
			      BUTTON_POPUP_FILE_IMPORT);
	filePopup->insertSeparator();
	addMButton(filePopup, BUTTON_POPUP_FILE_PRINT,
		   SLOT(print_slot()),
		   i18n("&Print..."), "fileprint");
	filePopup->insertSeparator();
	addMButton(filePopup, BUTTON_POPUP_FILE_QUIT,
		   SLOT(quit_slot()),
		   i18n("&Quit"), "exit");
	mb->insertItem(i18n("&File"), filePopup);
// "manage" popup menu
	addMButton(managePopup, BUTTON_POPUP_MANAGE_CHMASTERKEY,
		   SLOT(changeMasterKey_slot()),
		   i18n("Change &Master Key..."), "rotate");
	mb->insertItem(i18n("&Manage"), managePopup);
#ifdef CONFIG_PWMANAGER_SMARTCARD
// "smartcard" popup menu
	addMButton(smartcardPopup, BUTTON_POPUP_SMARTCARD_GETID,
		   SLOT(readSmartKeyId_slot()),
		   i18n("Read SmartKey &ID..."), "identity");
	smartcardPopup->insertSeparator();
	addMButton(smartcardPopup, BUTTON_POPUP_SMARTCARD_GEN,
		   SLOT(genSmartKey_slot()),
		   i18n("&Generate new SmartKey..."), "gear");
	addMButton(smartcardPopup, BUTTON_POPUP_SMARTCARD_ERASE,
		   SLOT(eraseSmartCard_slot()),
		   i18n("&Erase SmartCard..."), "eraser");
	smartcardPopup->insertSeparator();
	addMButton(smartcardPopup, BUTTON_POPUP_SMARTCARD_TOFILE,
		   SLOT(exportSmartKey_slot()),
		   i18n("E&xport SmartKey to KeyFile..."), "fileexport");
	addMButton(smartcardPopup, BUTTON_POPUP_SMARTCARD_FROMFILE,
		   SLOT(importSmartKey_slot()),
		   i18n("I&mport SmartKey from KeyFile..."), "fileimport");
	mb->insertItem(i18n("SmartCard"), smartcardPopup);
#endif // CONFIG_PWMANAGER_SMARTCARD
// "view" popup menu
	mb->insertItem(i18n("&View"), viewPopup);
// "tools" popup menu
	addMButton(toolsPopup, BUTTON_POPUP_TOOLS_PWGEN,
		   SLOT(genPassword_slot()),
		   i18n("&Generate Password..."), "math_sqrt");
	mb->insertItem(i18n("&Tools"), toolsPopup);
// "settings" popup menu
	addMButton(settingsPopup, BUTTON_POPUP_SETTINGS_CONFIG,
		   SLOT(config_slot()),
		   i18n("&Configure PwManager..."), "configure");
	mb->insertItem(i18n("&Settings"), settingsPopup);
// "help" popup menu
	helpPopup = helpMenu(QString::null, false);
	mb->insertItem(i18n("&Help"), helpPopup);
}

void MainWindow::addTButton(KToolBar *tb, int id,
			    const char *member, const QString &text,
			    const char *icon)
{
	tb->insertButton(iconLoader.loadIcon(icon, KIcon::MainToolbar),
			 id, SIGNAL(clicked(int)), this,
			 member, true, text);
}

void MainWindow::initToolBar()
{
	KIconLoader icons;
	KToolBar *tb = toolBar();

	addTButton(tb, BUTTON_TOOL_NEW,
		   SLOT(new_slot()),
		   i18n("New"), "filenew");
	addTButton(tb, BUTTON_TOOL_OPEN,
		   SLOT(open_slot()),
		   i18n("Open..."), "fileopen");
	tb->insertSeparator();
	addTButton(tb, BUTTON_TOOL_SAVE,
		   SLOT(save_slot()),
		   i18n("Save"), "filesave");
	addTButton(tb, BUTTON_TOOL_SAVEAS,
		   SLOT(saveAs_slot()),
		   i18n("Save As..."), "filesaveas");
	addTButton(tb, BUTTON_TOOL_PRINT,
		   SLOT(print_slot()),
		   i18n("Print..."), "fileprint");
	tb->insertSeparator();
	addTButton(tb, BUTTON_TOOL_LOCKDOC,
		   SLOT(lockDocument_slot()),
		   i18n("Lock / Unlock the Document"), "encrypted");
}

void MainWindow::initMetrics()
{
	QSize s = conf()->confWndMainWndSize();
	if (s.isValid())
		resize(s);
	else
		resize(DEFAULT_SIZE);
}

void MainWindow::updateCaption()
{
	PwMDoc *doc = getDoc();
	if (doc) {
		setPlainCaption(doc->getTitle()
				+ (doc->getPwFile()->isDirty() ? QString(" *") : QString::null)
				+ " - " PROG_NAME " " PACKAGE_VER);
	} else {
		setPlainCaption(PROG_NAME " " PACKAGE_VER);
	}
}

void MainWindow::updateButtons()
{
cout << "update buttons" << endl;//TODO: remove this cout.
	KIconLoader icons;
	PwMDoc *doc = getDoc();
	KToolBar *tb = toolBar();

	PWM_ASSERT(tb);

	/* A bool for every button / menu follows.
	 * If it's true, the button should be enabled.
	 */
	bool popup_file_new		= true;
	bool popup_file_open		= true;
	bool popup_file_close		= true;
	bool popup_file_lockdoc		= true;
	bool popup_file_save		= true;
	bool popup_file_saveas		= true;
	bool popup_file_export		= true;
	bool popup_file_import		= true;
	bool popup_file_print		= true;
	bool popup_file_quit		= true;
	bool popup_manage_chmasterkey	= true;
#ifdef CONFIG_PWMANAGER_SMARTCARD
	bool popup_smartcard_getid	= true;
	bool popup_smartcard_gen	= true;
	bool popup_smartcard_erase	= true;
	bool popup_smartcard_tofile	= true;
	bool popup_smartcard_fromfile	= true;
#endif // CONFIG_PWMANAGER_SMARTCARD
	bool popup_tools_pwgen		= true;
	bool popup_options_config	= true;
	bool popup_export_text		= true;
	bool popup_export_gpasman	= true;
	bool popup_export_pwsafe	= true;
	bool popup_export_csv		= true;
#ifdef CONFIG_KWALLETIF
	bool popup_export_kwallet	= true;
	bool popup_import_kwallet	= true;
#endif // CONFIG_KWALLETIF
	bool popup_import_text		= true;
	bool popup_import_gpasman	= true;
	bool popup_import_pwsafe	= true;
	bool popup_import_csv		= true;
	bool tool_new			= true;
	bool tool_open			= true;
	bool tool_save			= true;
	bool tool_saveas		= true;
	bool tool_print			= true;
	bool tool_lockdoc		= true;

	if (!doc) {
		filePopup->changeItem(BUTTON_POPUP_FILE_LOCKDOC,
				      QIconSet(icons.loadIcon("encrypted", KIcon::Small)),
				      i18n("&Lock"));
		tb->setButtonIcon(BUTTON_TOOL_LOCKDOC, "encrypted");

		popup_file_close = false;
		popup_file_lockdoc = false;
		popup_file_save = false;
		popup_file_saveas = false;
		popup_file_export = false;
		popup_file_print = false;
		popup_manage_chmasterkey = false;
		tool_save = false;
		tool_saveas = false;
		tool_print = false;
		tool_lockdoc = false;
		goto out;
	}

	switch (doc->getContext()) {
	case PwMDoc::ctx_file:
		filePopup->changeItem(BUTTON_POPUP_FILE_LOCKDOC,
				      QIconSet(icons.loadIcon("encrypted", KIcon::Small)),
				      i18n("&Lock"));

		tb->setButtonIcon(BUTTON_TOOL_LOCKDOC, "encrypted");
		break;
	case PwMDoc::ctx_locked:
		filePopup->changeItem(BUTTON_POPUP_FILE_LOCKDOC,
				      QIconSet(icons.loadIcon("decrypted", KIcon::Small)),
				      i18n("&Unlock"));

		tb->setButtonIcon(BUTTON_TOOL_LOCKDOC, "decrypted");
		break;
	case PwMDoc::ctx_noFile:
		filePopup->changeItem(BUTTON_POPUP_FILE_LOCKDOC,
				      QIconSet(icons.loadIcon("encrypted", KIcon::Small)),
				      i18n("&Lock"));
		popup_file_lockdoc = false;

		tb->setButtonIcon(BUTTON_TOOL_LOCKDOC, "encrypted");
		tool_lockdoc = false;
		break;
	}

	if (!doc->getDataTree()->count()) {
		popup_file_save = false;
		popup_file_saveas = false;
		popup_file_lockdoc = false;
		popup_file_export = false;
		popup_file_print = false;
		popup_manage_chmasterkey = false;
		tool_save = false;
		tool_saveas = false;
		tool_print = false;
		tool_lockdoc = false;
	}
	if (!doc->getPwFile()->isDirty()) {
		popup_file_save = false;
		tool_save = false;
	}

out:
	/* now enable / disable the buttons / menus,
	 * depending on the bool state.
	 */
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_NEW, popup_file_new);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_OPEN, popup_file_open);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_CLOSE, popup_file_close);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_LOCKDOC, popup_file_lockdoc);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_SAVE, popup_file_save);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_SAVEAS, popup_file_saveas);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_EXPORT, popup_file_export);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_IMPORT, popup_file_import);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_PRINT, popup_file_print);
	filePopup->setItemEnabled(BUTTON_POPUP_FILE_QUIT, popup_file_quit);
	managePopup->setItemEnabled(BUTTON_POPUP_MANAGE_CHMASTERKEY, popup_manage_chmasterkey);
#ifdef CONFIG_PWMANAGER_SMARTCARD
	smartcardPopup->setItemEnabled(BUTTON_POPUP_SMARTCARD_GETID, popup_smartcard_getid);
	smartcardPopup->setItemEnabled(BUTTON_POPUP_SMARTCARD_GEN, popup_smartcard_gen);
	smartcardPopup->setItemEnabled(BUTTON_POPUP_SMARTCARD_ERASE, popup_smartcard_erase);
	smartcardPopup->setItemEnabled(BUTTON_POPUP_SMARTCARD_TOFILE, popup_smartcard_tofile);
	smartcardPopup->setItemEnabled(BUTTON_POPUP_SMARTCARD_FROMFILE, popup_smartcard_fromfile);
#endif // CONFIG_PWMANAGER_SMARTCARD
	toolsPopup->setItemEnabled(BUTTON_POPUP_TOOLS_PWGEN, popup_tools_pwgen);
	settingsPopup->setItemEnabled(BUTTON_POPUP_SETTINGS_CONFIG, popup_options_config);
	exportPopup->setItemEnabled(BUTTON_POPUP_EXPORT_TEXT, popup_export_text);
	exportPopup->setItemEnabled(BUTTON_POPUP_EXPORT_GPASMAN, popup_export_gpasman);
	exportPopup->setItemEnabled(BUTTON_POPUP_EXPORT_PWSAFE, popup_export_pwsafe);
	exportPopup->setItemEnabled(BUTTON_POPUP_EXPORT_CSV, popup_export_csv);
#ifdef CONFIG_KWALLETIF
	exportPopup->setItemEnabled(BUTTON_POPUP_EXPORT_KWALLET, popup_export_kwallet);
	importPopup->setItemEnabled(BUTTON_POPUP_IMPORT_KWALLET, popup_import_kwallet);
#endif // CONFIG_KWALLETIF
	importPopup->setItemEnabled(BUTTON_POPUP_IMPORT_TEXT, popup_import_text);
	importPopup->setItemEnabled(BUTTON_POPUP_IMPORT_GPASMAN, popup_import_gpasman);
	importPopup->setItemEnabled(BUTTON_POPUP_IMPORT_PWSAFE, popup_import_pwsafe);
	importPopup->setItemEnabled(BUTTON_POPUP_IMPORT_CSV, popup_import_csv);
	tb->setItemEnabled(BUTTON_TOOL_NEW, tool_new);
	tb->setItemEnabled(BUTTON_TOOL_OPEN, tool_open);
	tb->setItemEnabled(BUTTON_TOOL_SAVE, tool_save);
	tb->setItemEnabled(BUTTON_TOOL_SAVEAS, tool_saveas);
	tb->setItemEnabled(BUTTON_TOOL_PRINT, tool_print);
	tb->setItemEnabled(BUTTON_TOOL_LOCKDOC, tool_lockdoc);
}

void MainWindow::statusBarMsg(const QString &msg, int timeout)
{
	statusBar()->message(msg, timeout * 1000);
}

void MainWindow::new_slot()
{
	PwMDoc *doc = getDoc();
	if (doc) {
		MainWindow *mainwnd = new MainWindow;
		mainwnd->initialize();
		mainwnd->initView();
	} else
		initView();
	updateCaption();
}

void MainWindow::open_slot()
{
	PwMDoc *doc = getDoc();
	if (doc) {
		// Open in a new MainWindow.
		MainWindow *mainwnd = new MainWindow;
		mainwnd->initialize();
		mainwnd->initView();
		doc = mainwnd->getDoc();
		if (!doc->load(QString(), this)) {
			delete mainwnd;
			return;
		}
		mainwnd->updateCaption();
		mainwnd->updateButtons();
		goto out_success;
	}
	initView();
	doc = getDoc();
	PWM_ASSERT(doc);
	if (!doc->load(QString(), this)) {
		initSplash();
		return;
	}
	updateCaption();
	updateButtons();
out_success:
	statusBarMsg(i18n("Successfully opened file."));
}

void MainWindow::close_slot()
{
	unsigned int mainwndCnt = list.count();
	PWM_ASSERT(mainwndCnt != 0);
	if (mainwndCnt == 1) {
		/* This is the last window.
		 * Don't close the MainWindow, but display
		 * the splash instead of the view widget.
		 */
		initSplash();
	} else
		close();
	updateCaption();
}

void MainWindow::lockDocument_slot()
{
	PwMDoc *doc = getDoc();
	if (!doc) {
		PWM_ASSERT(false);
		return;
	}
	switch (doc->getContext()) {
	case PwMDoc::ctx_file:
		doc->lockDocument();
		break;
	case PwMDoc::ctx_locked:
		doc->unlockDocument();
		break;
	case PwMDoc::ctx_noFile:
		PWM_ASSERT(false);
	}
	updateCaption();
	updateButtons();
}

void MainWindow::quit_slot()
{
	pwminit->shutdownApp();
}

void MainWindow::save_slot()
{
	PwMDoc *doc = getDoc();
	if (!doc) {
		PWM_ASSERT(false);
		return;
	}
	if (!doc->save(QString(), this)) {
		statusBarMsg(i18n("Failed to save data."));
		return;
	}
	updateCaption();
	updateButtons();
	statusBarMsg(i18n("Successfully saved data."));
}

void MainWindow::saveAs_slot()
{
	PwMDoc *doc = getDoc();
	if (!doc) {
		PWM_ASSERT(false);
		return;
	}
	if (!doc->saveAs(this)) {
		statusBarMsg(i18n("Failed to save data."));
		return;
	}
	updateCaption();
	updateButtons();
	statusBarMsg(i18n("Successfully saved data."));
}

void MainWindow::exportText()
{//TODO
}

void MainWindow::exportGpasman()
{//TODO
}

void MainWindow::exportPwSafe()
{//TODO
}

void MainWindow::exportKWallet()
{//TODO
}

void MainWindow::exportCsv()
{//TODO
}

bool MainWindow::importText()
{//TODO
}

bool MainWindow::importGpasman()
{//TODO
}

bool MainWindow::importPwSafe()
{//TODO
}

bool MainWindow::importKWallet()
{//TODO
}

bool MainWindow::importCsv()
{//TODO
}

void MainWindow::print_slot()
{
	PwMDoc *doc = getDoc();
	if (!doc) {
		PWM_ASSERT(false);
		return;
	}
	PwMPrint p(doc, this);
	p.printNow();
}

void MainWindow::config_slot()
{
	conf()->showConfWnd(this);
}

void MainWindow::genPassword_slot()
{
	PwMInit::genPassword(this);
}

void MainWindow::readSmartKeyId_slot()
{
#ifdef CONFIG_PWMANAGER_SMARTCARD
	SmartKey sk(this);
	sk.displayKeyId();
#endif
}

void MainWindow::genSmartKey_slot()
{
#ifdef CONFIG_PWMANAGER_SMARTCARD
	SmartKey sk(this);
	sk.genKey();
#endif
}

void MainWindow::eraseSmartCard_slot()
{
#ifdef CONFIG_PWMANAGER_SMARTCARD
	SmartKey sk(this);
	sk.eraseCard();
#endif
}

void MainWindow::exportSmartKey_slot()
{
#ifdef CONFIG_PWMANAGER_SMARTCARD
	SmartKey sk(this);
	sk.exportToKeyFile();
#endif
}

void MainWindow::importSmartKey_slot()
{
#ifdef CONFIG_PWMANAGER_SMARTCARD
	SmartKey sk(this);
	sk.importFromKeyFile();
#endif
}

void MainWindow::changeMasterKey_slot()
{
	PwMDoc *doc = getDoc();
	if (!doc) {
		PWM_ASSERT(false);
		return;
	}
	doc->changeMasterKey();
}

#include "mainwindow.moc"
