/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004, 2005  by Michael Buesch                     *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#ifndef PWMANAGER_MAINWND_H_
#define PWMANAGER_MAINWND_H_

#include "globalstuff.h"
#include "pwmview.h"

#include <qptrlist.h>

#include <kmainwindow.h>
#include <kiconloader.h>


class KPopupMenu;

class PwMInit;
class PwMDoc;

/** This class implements the main window for PwManager. */
class MainWindow : public KMainWindow
{
	Q_OBJECT

	friend class PwMTray;

public:
	MainWindow(QWidget *parent = 0,
		   const char *name = 0,
		   WFlags f = WType_TopLevel | WDestructiveClose);
	~MainWindow();

	/** Returns a pointer to the MainWindow list.
	  * It contains a pointer to every active MainWindow instance.
	  */
	static QPtrList<MainWindow> * getList()
			{ return &list; }

	/** Initialize the main window, connect to the document
	  * and show the widget.
	  * If _doc is 0, a new PwMDoc instance is created.
	  * This function can not fail, if loadFile is empty.
	  */
	bool initialize(PwMDoc *doc = 0,
			const QString &loadFile = QString(),
			bool showWndMinimized = false);

	/** Returns a pointer to the PwMDoc instance. */
	PwMDoc * getDoc()
		{
			PwMView *view = getView();
			return (view ? view->getDoc() : 0);
		}
	/** Returns a pointer to the PwMView central view widget,
	  * or NULL if there is none.
	  */
	PwMView * getView()
			{ return (dynamic_cast<PwMView *>(central)); }

	/** Display a message in the status bar.
	  * The message disappears after a timeout.
	  * The timeout is in seconds.
	  */
	void statusBarMsg(const QString &msg, int timeout = 10);

protected:
	/** Initialize the menu bar. */
	void initMenuBar();
	/** Helper function to add buttons to the menu bar. */
	void addMButton(QPopupMenu *menu, int id,
			const char *member, const QString &text = QString(),
			const char *icon = 0);
	/** Initialize the tool bar. */
	void initToolBar();
	/** Helper function to add buttons to the tool bar. */
	void addTButton(KToolBar *tb, int id,
			const char *member, const QString &text,
			const char *icon);
	/** Initialize the window metrics. */
	void initMetrics();
	/** Initialize the view widget as central widget. */
	void initView(PwMDoc *doc = 0);
	/** Initialize the splash screen as central widget. */
	void initSplash();
	/** Update the window title bar. */
	void updateCaption();

protected slots:
	/** File / New */
	void new_slot();
	/** File / Open */
	void open_slot();
	/** File / Close */
	void close_slot();
	/** File / (Un)Lock */
	void lockDocument_slot();
	/** File / Quit */
	void quit_slot();
	/** File / Save */
	void save_slot();
	/** File / SaveAs */
	void saveAs_slot();
	/** File / Export / Text */
	void exportText();
	/** File / Export / GPasman */
	void exportGpasman();
	/** File / Export / PwSafe */
	void exportPwSafe();
	/** File / Export / KWallet */
	void exportKWallet();
	/** File / Export / CSV */
	void exportCsv();
	/** File / Import / Text */
	bool importText();
	/** File / Import / GPasman */
	bool importGpasman();
	/** File / Import / PwSafe */
	bool importPwSafe();
	/** File / Import / KWallet */
	bool importKWallet();
	/** File / Import / CSV */
	bool importCsv();
	/** File / Print */
	void print_slot();
	/** Settings / Configure */
	void config_slot();
	/** Tools / Generate Password */
	void genPassword_slot();
	/** Smartcard / Read ID */
	void readSmartKeyId_slot();
	/** Smartcard / Generate Key */
	void genSmartKey_slot();
	/** Smartcard / Erase */
	void eraseSmartCard_slot();
	/** Smartcard / To Keyfile */
	void exportSmartKey_slot();
	/** Smartcard / From Keyfile */
	void importSmartKey_slot();
	/** Manage / Change Masterkey */
	void changeMasterKey_slot();

protected slots:
	/** Update the menu button states and texts. */
	void updateButtons();

protected:
	/** List of all MainWindow instances. */
	static QPtrList<MainWindow> list;
	/** Pointer to the central widget. */
	QWidget *central;
	/** Icon loader for button icons. */
	KIconLoader iconLoader;
	/** "File" popup menu. */
	KPopupMenu *filePopup;
	/** "Manage" popup menu. */
	KPopupMenu *managePopup;
#ifdef CONFIG_PWMANAGER_SMARTCARD
	/** Smartcard popup menu. */
	KPopupMenu *smartcardPopup;
#endif
	/** "View" popup menu. */
	KPopupMenu *viewPopup;
	/** "Tools" popup menu. */
	KPopupMenu *toolsPopup;
	/** "Settings" popup menu. */
	KPopupMenu *settingsPopup;
	/** "Help" popup menu. */
	KPopupMenu *helpPopup;
	/** "Export" popup menu. */
	KPopupMenu *exportPopup;
	/** "Import" popup menu. */
	KPopupMenu *importPopup;
};

class KHTMLPart;

class MainWindowSplash : public QWidget
{
	Q_OBJECT

	friend class MainWindow;

protected:
	MainWindowSplash(QWidget *parent = 0,
			 const char *name = 0,
			 WFlags f = 0);
	~MainWindowSplash();

protected:
	void resizeEvent(QResizeEvent *);

protected:
	/** HTML part of the splash. */
	KHTMLPart *html;
	/** Alternative Widget (if no HTML available). */
	QWidget *alt;
};

#endif // PWMANAGER_MAINWND_H_
