/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004 by Michael Buesch                                  *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "masterkey/smartkey/smartkey_backend.h"
#include "masterkey/smartkey/smartkey.h"
#include "libpwmanager/ipc.h"
#include "libpwmanager/waitwnd.h"
#include "pwmexception.h"
#include "globalstuff.h"
#include "pwminit.h"

#include <chipcard.h>
#include <time.h>
#include <unistd.h>

#include <string>

#include <kapplication.h>
#include <klocale.h>

#include <qeventloop.h>


/** open timeout in seconds */
#define CARD_OPEN_TIMEOUT	30
/** begin reading/writing the card at this offset. */
#define CARD_BEGIN_OFFSET	30

/** show the openWaitWnd. */
#define IPC_MSG_OWW_SHOW	"openWaitWnd_show"
/** hide the openWaitWnd. */
#define IPC_MSG_OWW_HIDE	"openWaitWnd_hide"
/** do timeout progress for openWaitWnd. */
#define IPC_MSG_OWW_PROGRESS	"openWaitWnd_progress"
/** show the accessWaitWnd. */
#define IPC_MSG_AWW_SHOW	"accessWaitWnd_show"
/** hide the accessWaitWnd. */
#define IPC_MSG_AWW_HIDE	"accessWaitWnd_hide"


/* variables for communication with the thread. */
static volatile SmartKey_backend::Task task = SmartKey_backend::task_none;
static volatile SmartKey_backend::Status status = SmartKey_backend::status_none;
static QByteArray *dataBuffer;
static ThreadIpcClient *threadIpc;


SmartKey_backend::SmartKey_backend(SmartKey *_sk)
 : sk (_sk)
 , openWaitWnd (0)
 , accessWaitWnd (0)
 , hostIpc (0)
{
}

SmartKey_backend::~SmartKey_backend()
{
	PWM_ASSERT(openWaitWnd == 0);
	PWM_ASSERT(accessWaitWnd == 0);
	PWM_ASSERT(hostIpc == 0);
}

SmartKey_backend::Status SmartKey_backend::writeCard(const QByteArray &data)
{
	PWM_ASSERT(status == status_none);
	PWM_ASSERT(task == task_none);
	task = task_write;
	// We cast it back to (const QByteArray &) in run() before using it
	dataBuffer = const_cast<QByteArray *>(&data);
	return fireUp();
}

SmartKey_backend::Status SmartKey_backend::readCard(QByteArray *data)
{
	PWM_ASSERT(status == status_none);
	PWM_ASSERT(task == task_none);
	task = task_read;
	dataBuffer = data;
	return fireUp();
}

SmartKey_backend::Status SmartKey_backend::eraseCard()
{
	PWM_ASSERT(status == status_none);
	PWM_ASSERT(task == task_none);
	task = task_erase;
	return fireUp();
}

void SmartKey_backend::cancel()
{
	if (!running())
		return;
	task = task_cancel;
	wait();
}

SmartKey_backend::Status SmartKey_backend::fireUp()
{
	// initialize the IPC
	hostIpc = new ThreadIpcHost(false);
	connect(hostIpc, SIGNAL(receivedMessage(const QByteArray &)),
		this, SLOT(threadMessage(const QByteArray &)));
	threadIpc = new ThreadIpcClient(*hostIpc);

	// create the notification windows.
	openWaitWnd = new WaitWnd(i18n("Accessing SmartCard"),
				  i18n("Please insert the SmartCard into "
				       "your SmartCard Terminal."),
				  true, true, sk->parent);
	openWaitWndCnt = CARD_OPEN_TIMEOUT;
	connect(openWaitWnd, SIGNAL(closing()),
		this, SLOT(cancel()));
	accessWaitWnd = new WaitWnd(i18n("Accessing SmartCard"),
				    i18n("Accessing SmartCard.\n"
					 "Please do not remove the SmartCard."),
				    false, false, sk->parent);

	hostIpc->startLookup();
	start(); // now start the thread
	while (running())
		relax();
	hostIpc->stopLookup();

	delete_and_null(accessWaitWnd);
	delete_and_null(openWaitWnd);
	delete_and_null(threadIpc);
	delete_and_null(hostIpc);

	PWM_ASSERT(status != status_running);
	PWM_ASSERT(status != status_none);
	Status retStatus = status;
	status = status_none;
	task = task_none;
	return retStatus;
}

void SmartKey_backend::relax()
{
	bool inEventLoop = (kapp->eventLoop()->loopLevel() >= 1);
	if (inEventLoop)
		kapp->processEvents();
	QThread::msleep(10);
}

void SmartKey_backend::start(Priority priority)
{
	PWM_ASSERT(task != task_none);
	QThread::start(priority);
}

void SmartKey_backend::threadMessage(const QByteArray &msg)
{
	if (QCString(msg) == QCString(IPC_MSG_OWW_SHOW)) {
		openWaitWnd->show();
		sk->parent->setEnabled(false);
		openWaitWnd->setEnabled(true);
		goto oww_set_timeout;
	} else if (QCString(msg) == QCString(IPC_MSG_OWW_HIDE)) {
		sk->parent->setEnabled(true);
		openWaitWnd->hide();
	} else if (QCString(msg) == QCString(IPC_MSG_OWW_PROGRESS)) {
		openWaitWndCnt--;
		goto oww_set_timeout;
	} else if (QCString(msg) == QCString(IPC_MSG_AWW_SHOW)) {
		accessWaitWnd->show();
		sk->parent->setEnabled(false);
		accessWaitWnd->setEnabled(true);
	} else if (QCString(msg) == QCString(IPC_MSG_AWW_HIDE)) {
		sk->parent->setEnabled(true);
		accessWaitWnd->hide();
	} else
		BUG();
	return;
oww_set_timeout:
	openWaitWnd->setGenericText(i18n("Timeout: %1")
			.arg(QString::number(openWaitWndCnt)));
}

void SmartKey_backend::run()
{
	status = status_running;
	switch (task) {
	case task_read:
		read(dataBuffer);
		break;
	case task_write:
		write(*dataBuffer);
		break;
	case task_erase:
		erase();
		break;
	case task_cancel:
		break;
	case task_none:
		BUG();
	}
	PWM_ASSERT(status != status_running);
}

void SmartKey_backend::write(const QByteArray &data)
{
	CTCardTrader *ctTrader;
	CTCard *ctCard;

	if (!ct_open(&ctTrader, &ctCard)) {
		printDebug("ct_open() failed");
		return;
	}
	if (!ct_write(data, ctCard)) {
		printDebug("ct_write() failed");
		return;
	}
	if (!ct_close(ctTrader, ctCard)) {
		printDebug("ct_close() failed");
		return;
	}
	status = status_ok;
}

void SmartKey_backend::read(QByteArray *data)
{
	CTCardTrader *ctTrader;
	CTCard *ctCard;

	if (!ct_open(&ctTrader, &ctCard)) {
		printDebug("ct_open() failed");
		return;
	}
	if (!ct_read(data, ctCard)) {
		printDebug("ct_read() failed");
		return;
	}
	if (!ct_close(ctTrader, ctCard)) {
		printDebug("ct_close() failed");
		return;
	}
	status = status_ok;
}

void SmartKey_backend::erase()
{
	CTCardTrader *ctTrader;
	CTCard *ctCard;

	if (!ct_open(&ctTrader, &ctCard)) {
		printDebug("ct_open() failed");
		return;
	}
	if (!ct_erase(ctCard)) {
		printDebug("ct_read() failed");
		return;
	}
	if (!ct_close(ctTrader, ctCard)) {
		printDebug("ct_close() failed");
		return;
	}
	status = status_ok;
}

bool SmartKey_backend::ct_open(CTCardTrader **_ctTrader, CTCard **_ctCard)
{
	CTError err;
	CTCardTrader *ctTrader;
	CTCard *ctCard;
	int i;

	threadIpc->sendMessage(QCString(IPC_MSG_OWW_SHOW));

	ctTrader = new CTCardTrader(false,
				    0, 0,
				    CHIPCARD_STATUS_INSERTED,
				    CHIPCARD_STATUS_INSERTED |
				    CHIPCARD_STATUS_LOCKED_BY_OTHER,
				    CHIPCARD_STATUS_INSERTED);
	err = ctTrader->start();
	if (!err.isOk()) {
		printDebug("ctTrader->start() failed");
		status = status_errOpen;
		goto err_delete;
	}
	for (i = 0; i < CARD_OPEN_TIMEOUT; ++i) {
		if (task == task_cancel) {
			status = status_cancel;
			goto err_stop;
		}
		err = ctTrader->getNext(ctCard, 1);
		if (err.isOk())
			break;
		threadIpc->sendMessage(QCString(IPC_MSG_OWW_PROGRESS));
	}
	if (!err.isOk()) {
		printDebug("card insert timeout");
		status = status_errTimeout;
		goto err_stop;
	}
	err = ctCard->openCard();
	if (!err.isOk()) {
		printDebug("ctCard->openCard() failed");
		status = status_errOpen;
		goto err_stop;
	}
	*_ctTrader = ctTrader;
	*_ctCard = ctCard;
	threadIpc->sendMessage(QCString(IPC_MSG_OWW_HIDE));
	return true;

err_stop:
	ctTrader->stop();
err_delete:
	delete ctTrader;
	threadIpc->sendMessage(QCString(IPC_MSG_OWW_HIDE));
	return false;
}

bool SmartKey_backend::ct_close(CTCardTrader *ctTrader, CTCard *ctCard)
{
	ctCard->closeCard();
	ctTrader->stop();
	delete ctTrader;
	return true;
}

bool SmartKey_backend::ct_write(const QByteArray &data, CTCard *ctCard)
{
	bool ret = true;
	CTError err;
	std::string buf;
	threadIpc->sendMessage(QCString(IPC_MSG_AWW_SHOW));
	unsigned int cardSize = ctCard->memorySize();
	if (cardSize < data.size() + CARD_BEGIN_OFFSET) {
		printDebug("card too small");
		status = status_err2small;
		ret = false;
		goto out;
	}
	buf.assign(data.data(), data.size());
	err = ctCard->updateBinaryRaw(buf, CARD_BEGIN_OFFSET);
	if (!err.isOk()) {
		printDebug("ctCard->updateBinaryRaw() failed");
		status = status_errWrite;
		ret = false;
		goto out;
	}
out:
	threadIpc->sendMessage(QCString(IPC_MSG_AWW_HIDE));
	return ret;
}

bool SmartKey_backend::ct_read(QByteArray *data, CTCard *ctCard)
{
	bool ret = true;
	CTError err;
	threadIpc->sendMessage(QCString(IPC_MSG_AWW_SHOW));
	unsigned int cardSize = ctCard->memorySize();
	std::string buf;
	err = ctCard->readBinaryRaw(buf, CARD_BEGIN_OFFSET, cardSize - CARD_BEGIN_OFFSET);
	if (!err.isOk()) {
		printDebug("ctCard->readBinaryRaw() failed");
		status = status_errRead;
		ret = false;
		goto out;
	}
	data->duplicate(buf.c_str(), buf.size());
out:
	threadIpc->sendMessage(QCString(IPC_MSG_AWW_HIDE));
	return ret;
}

bool SmartKey_backend::ct_erase(CTCard *ctCard)
{
	bool ret = true;
	CTError err;
	threadIpc->sendMessage(QCString(IPC_MSG_AWW_SHOW));
	unsigned int cardSize = ctCard->memorySize();
	unsigned int i = 0;
	const unsigned int byteCntEachLoop = 255;
	std::string buf(byteCntEachLoop, static_cast<char>(0));
	while (i < cardSize) {
		if ((cardSize - i) < byteCntEachLoop)
			buf.assign(cardSize - i, static_cast<char>(0));
		err = ctCard->updateBinaryRaw(buf, i);
		if (!err.isOk()) {
			printDebug("ctCard->updateBinaryRaw() failed");
			status = status_errWrite;
			ret = false;
			goto out;
		}
		i += byteCntEachLoop;
	}
out:
	threadIpc->sendMessage(QCString(IPC_MSG_AWW_HIDE));
	return ret;
}

#include "smartkey_backend.moc"
