/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004, 2005  by Michael Buesch                     *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pwmdoc.h"
#include "pwmexception.h"
#include "pwminit.h"

#include <kfiledialog.h>
#include <klocale.h>
#include <kmessagebox.h>


using namespace PwManager;


QPtrList<PwMDoc> PwMDoc::list;
uint64_t PwMDoc::ctxNoFileCnt (1);


PwMDoc::PwMDoc(QObject *parent, const char *name)
 : QObject (parent, name)
 , usageCnt (0)
 , pwfile (0)
 , context (ctx_noFile)
{
	list.append(this);
}

PwMDoc::~PwMDoc()
{
	list.remove(this);
	emit closed(this);
	delete_ifnot_null(pwfile);
}

bool PwMDoc::decUsageCount(bool force)
{
	PWM_ASSERT(usageCnt);
	usageCnt--;
	if (usageCnt)
		return true;
	/* The Usage-Count reached zero.
	 * Try to delete the instance, now. Be careful to save all data.
	 */
	if (pwfile && pwfile->isDirty()) {
		int ret;
		QString msg(i18n("The list \"%1\" has been modified;\n"
				 "do you want to save it?").arg(getTitle()));
		QString title(i18n("Save?"));
		QWidget *pWidget = pwminit->curWidget();
		if (force)
			ret = KMessageBox::questionYesNo(pWidget, msg, title);
		else
			ret = KMessageBox::questionYesNoCancel(pWidget, msg, title);
		switch (ret) {
		case KMessageBox::Yes:
		    try_again:
			if (!save()) {
				if (!force)
					return false;
				ret = KMessageBox::questionYesNo(pWidget,
								 i18n("Failed to save the data. "
								      "Do you want to try again?\n"
								      "All changes will be lost, if you do "
								      "not try again."),
								 i18n("Try Again?"));
				if (ret == KMessageBox::Yes)
					goto try_again;
				printWarn("Document closing forced, "
					  "althought there was dirty data!");
			}
			break;
		case KMessageBox::No:
			break;
		case KMessageBox::Cancel:
			PWM_ASSERT(!force);
			return false;
		default:
			PWM_ASSERT(false);
		}
	}
	delete this;
	return true;
}

void PwMDoc::forceZeroUsageCount()
{
	bool err;
	PWM_ASSERT(usageCnt >= 1);
	usageCnt = 1;
	err = decUsageCount(true);
	PWM_ASSERT(err);
}

PwFile * PwMDoc::getPwFile()
{
	if (!pwfile) {
		pwfile = pwminit->lib()->newFile();
		connectPwFile(pwfile);
	}
	return pwfile;
}

void PwMDoc::connectPwFile(PwManager::PwFile *f)
{
	PWM_ASSERT(f);
	//FIXME: nothing to do here. remove it?
}

QString PwMDoc::getTitle() const
{
	if (pwfile)
		return pwfile->getPath();
	if (getContext() == ctx_locked) {
		PWM_ASSERT(!savedPath.isEmpty());
		return savedPath;
	}
	static uint64_t idnum = ctxNoFileCnt++;
	return i18n("Untitled %1").arg(QString::number(idnum));
}

bool PwMDoc::load(const QString &filepath,
		  QWidget *parentWidget)
{
	PWM_ASSERT(!pwfile);

	if (!parentWidget)
		parentWidget = pwminit->curWidget();
	QString path(filepath);
	if (path.isEmpty()) {
		path = KFileDialog::getOpenFileName(QString(),
						    i18n("*.pwm|PwManager Data File\n"
							 "*|All Files"),
						    parentWidget);
	}
	if (path.isEmpty())
		return false;
	pwfile = pwminit->lib()->openFile(path);
	if (!pwfile) {
		ErrCode err;
		//TODO: get and display error code.
		return false;
	}
	connectPwFile(pwfile);
	return true;
}

bool PwMDoc::save(const QString &filepath,
		  QWidget *parentWidget)
{
	PWM_ASSERT(pwfile);

	if (!parentWidget)
		parentWidget = pwminit->curWidget();
	QString path(filepath);
	if (path.isEmpty()) {
		path = getPwFile()->getPath();
		if (path.isEmpty())
			return saveAs(parentWidget);
		// ok, save under the old name.
	}
	return saveAs(path, parentWidget);
}

bool PwMDoc::saveAs(QWidget *parentWidget)
{
	if (!parentWidget)
		parentWidget = pwminit->curWidget();
	QString path(KFileDialog::getSaveFileName(QString(),
						  i18n("*.pwm|PwManager Data File"),
						  parentWidget));
	if (path.isEmpty())
		return false;
	return saveAs(path, parentWidget);
}

bool PwMDoc::saveAs(const QString &filepath,
		    QWidget *parentWidget)
{
	PWM_ASSERT(!filepath.isEmpty());
	PWM_ASSERT(pwfile);

	if (!pwfile->setPath(filepath)) {
		FIXME();//FIXME
		goto out_err;
	}
	if (!pwfile->sync())
		goto out_err;
	return true;

out_err:
	ErrCode err = pwfile->getError();
	switch (err) {
	//TODO
	}
	return false;
}

bool PwMDoc::lockDocument(bool saveDataToDisk)
{
	switch (getContext()) {
	case ctx_noFile:
		/* We can not save to disk, as we don't have a file
		 * for writing. The only thing we can do is disabling
		 * each entry. No way to delete the data from memory!
		 */
		TODO();//TODO
#if 0
		if (!lockAll(true))
			return false;
#endif
	case ctx_locked:
		// Nothing to do here.
		return true;
	case ctx_file:
		PWM_ASSERT(pwfile);
		PWM_ASSERT(savedPath.isEmpty());
		savedPath = pwfile->getPath();
		PWM_ASSERT(!savedPath.isEmpty());
		if (saveDataToDisk && getDataTree()->count()) {
			if (!getPwFile()->sync())
				return false;
		}
		delete_and_null(pwfile);
	}
	setContext(ctx_locked);
	return true;

}

bool PwMDoc::unlockDocument()
{
	switch (getContext()) {
	case ctx_noFile:
		/* We are not half closed and there is no file on disk
		 * to read from. Just return with success.
		 */
		//FIXME: success correct here?
		return true;
	case ctx_locked: {
		PWM_ASSERT(!savedPath.isEmpty());
		PWM_ASSERT(!pwfile);
		setContext(ctx_file);
		pwfile = pwminit->lib()->openFile(savedPath);
		if (!pwfile) {
			setContext(ctx_locked);
			return false;
		}
		connectPwFile(pwfile);
	} case ctx_file:
		return true;
	}
	return true;
}

void PwMDoc::changeMasterKey()
{TODO();//TODO
}

#include "pwmdoc.moc"
