/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004, 2005  by Michael Buesch                     *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#ifndef PWMDOC_H_
#define PWMDOC_H_

#include "globalstuff.h"
#include "libpwmanager/libpwmanager.h"

#include <qobject.h>
#include <qptrlist.h>

#include <stdint.h>


/** Implementation of the "document" in the Document-View model. */
class PwMDoc : public QObject
{
	Q_OBJECT

public:
	enum Context
	{
		/** The document has no file on disk. */
		ctx_noFile,
		/** We have a file on disk and all data has
		  * loaded/decrypted into memory.
		  */
		ctx_file,
		/** All data is written to the file on disk and
		  * deleted in memory.
		  */
		ctx_locked,
	};

public:
	PwMDoc(QObject *parent = 0, const char *name = 0);
	~PwMDoc();

	/** Returns a pointer to the PwMDoc list.
	  * It contains a pointer to every PwMDoc instance.
	  */
	static QPtrList<PwMDoc> * getList()
			{ return &list; }

	/** Increment the Usage-Counter. */
	void incUsageCount()
			{ usageCnt++; }
	/** Decrement the Usage-Counter.
	  * If the counter reaches zero, this instance is deleted.
	  * If force is true, this will always succeed. You may loose data!
	  */
	bool decUsageCount(bool force = false);
	/** Force the Usage-Counter to zero, delete the document and
	  * finally emit closed() through the dtor.
	  */
	void forceZeroUsageCount();
	/** Returns the current document context. */
	Context getContext() const
			{ return context; }
	/** Returns a pointer to the main PwFile object.
	  * If no object exists, a new one is created.
	  */
	PwManager::PwFile * getPwFile();
	/** Returns a pointer to the data tree root. */
	PwManager::DataRoot * getDataTree()
			{ return getPwFile()->getDataTree(); }
	/** Returns the title for the document.
	  * This is for use in a window title bar for example.
	  */
	QString getTitle() const;
	/** Returns a unique ID string for the document.
	  * This is currently the same as getTitle().
	  */
	QString getId() const
			{ return getTitle(); }

	/** Load a document file.
	  * If filepath is empty, this opens a dialog
	  * window to ask for the file.
	  */
	bool load(const QString &filepath = QString(),
		  QWidget *parentWidget = 0);
	/** Save the document data to disk.
	  * If filepath is empty and a filepath is not already
	  * available in the document, this will ask for a new one.
	  */
	bool save(const QString &filepath = QString(),
		  QWidget *parentWidget = 0);
	/** Save the document data to disk.
	  * This is the same as above, but always asks for the
	  * file name.
	  */
	bool saveAs(QWidget *parentWidget = 0);
	/** Lock the document. (changes Context to ctx_locked). */
	bool lockDocument()
			{ return lockDocument(true); }
	/** Unlock the document. (changes Context to ctx_file). */
	bool unlockDocument();
	/** open a window and request the user to change the master key. */
	void changeMasterKey();

signals:
	/** This signal is emitted, if this document is deleted. */
	void closed(PwMDoc *doc);

protected:
	/** List of all PwMDoc instances. */
	static QPtrList<PwMDoc> list;
	/** Global counter to get a unique name for
	  * documents in ctx_noFile.
	  * uint64_t because unlikely to overflow before
	  * the universum blows up.
	  */
	static uint64_t ctxNoFileCnt;
	/** Usage reference count.
	  * If this counter reaches zero, the object is destroyed.
	  */
	unsigned int usageCnt;
	/** Pointer to the main PwFile object.
	  * This can be NULL.
	  */
	PwManager::PwFile *pwfile;
	/** Document context. */
	Context context;
	/** This is the saved filepath while the document is locked. */
	QString savedPath;

protected:
	/** Set the document context. */
	void setContext(Context ctx)
			{ context = ctx; }
	/** Low level save function, with user interface.
	  * This is only called by the public save() and saveAs().
	  */
	bool saveAs(const QString &filepath,
		    QWidget *parentWidget);
	/** Lock the document. (changes Context to ctx_locked). */
	bool lockDocument(bool saveDataToDisk);
	/** Connect PwFile signals. */
	void connectPwFile(PwManager::PwFile *f);
};

#endif // PWMDOC_H_
