/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "main.h"
#include "pwminit.h"
#include "configuration.h"
#include "pwmexception.h"
#include "pwmtray.h"
#include "pwmdoc.h"
#include "mainwindow.h"
#include "libpwmanager/libpwmanager.h"
#include "pwgenwndimpl.h"

//FIXME: tempoarly disabled now.
#undef CONFIG_KWALLETIF

#ifdef CONFIG_KWALLETIF
# include "kwalletemu.h"
#endif // CONFIG_KWALLETIF

#include <qmessagebox.h>
#include <qeventloop.h>
#include <qclipboard.h>

#include <kmessagebox.h>
#include <kcmdlineargs.h>
#include <kapplication.h>
#include <kiconloader.h>
#include <kwin.h>
#include <kinputdialog.h>
#include <klocale.h>
#include <dcopclient.h>

#include <signal.h>

using namespace PwManager;


static void NOREGPARM sig_handler(int signum)
{
	PwMInit *init = pwminit;
	PWM_ASSERT(init);
	PwMApplication *app = init->getApp();
	PWM_ASSERT(app);
	app->lock();
	bool hardShutdown = false;
	switch (signum) {
		case SIGILL:
#if COMPILE_BUGCHECKING != 0
			if (!_pwmanagerFatalInProgress)
#endif
			{
				printError("Received SIGILL from the Operating System.\n"
					   "This is either a bug in PwManager, your Compiler "
					   "or your Operating System libraries.");
			}
			hardShutdown = true;
		case SIGINT:
		case SIGTERM:
			printInfo(QString("caught signal ") + QString::number(signum));
			init->shutdownApp(PWM_EXIT_SIG + signum, hardShutdown);
			break;
		default:
			printDebug(QString("unhandled signal ")
				   + QString::number(signum));
	}
	app->unlock();
}

PwMInit * PwMInit::myself (0);

PwMInit::PwMInit(PwMApplication *_app)
 : runStatus (unknown)
 , app (_app)
 , _dcopClient (0)
 , _kwalletEmu (0)
 , _tray (0)
{
	BUG_ON(myself);
	myself = this;
}

PwMInit::~PwMInit()
{
	printDebug("~PwMInit");

#ifdef CONFIG_KWALLETIF
	delete_ifnot_null(_kwalletEmu);
#endif // CONFIG_KWALLETIF
	delete_ifnot_null(_tray);
	Configuration::cleanup();
	delete_ifnot_null(_lib);

	BUG_ON(!MainWindow::getList()->isEmpty());
	BUG_ON(!PwMDoc::getList()->isEmpty());
	myself = 0;
}

void PwMInit::copyToClipboard(const QString &text)
{
	QClipboard *cb = QApplication::clipboard();
	if (cb->supportsSelection())
		cb->setText(text, QClipboard::Selection);
	cb->setText(text, QClipboard::Clipboard);
}

void PwMInit::genPassword(QWidget *parent)
{
	if (!parent)
		parent = pwminit->curWidget();
	PwGenWndImpl wnd(parent);
	if (wnd.exec() == 0)
		return;
	bool ok = false;
	QString pw(wnd.getPassword());
	pw = KInputDialog::getText(i18n("Generated Password"),
				   i18n("This is your new Password.\n"
					"Click on OK to copy it to the clipboard."),
				   pw, &ok, parent);
	if (ok)
		copyToClipboard(pw);
}

bool PwMInit::initializeApp()
{
	PWM_ASSERT(app->eventLoop()->loopLevel() >= 1);
	PWM_ASSERT(runStatus == unknown);
	runStatus = init;
	initPosixSignalHandler();
	_lib = LibPwManager::initialize();
	if (!_lib)
		return false;
	Configuration::init();
	initDCOP();
	initKWalletEmu();
	initTray();
	handleCmdLineArgs();

	if (!PwMDoc::getList()->count()) {
		if (conf()->confWndAutoMinimizeOnStart() ||
		    savedCmd.minToTray) {
			PwMDoc *newDoc = new PwMDoc(this);
			if (!newDoc->load(conf()->confGlobAutoStart()))
				delete newDoc;
		} else {
			MainWindow *newMainWnd = new MainWindow;
			bool ok;
			ok = newMainWnd->initialize(0, conf()->confGlobAutoStart(),
						    savedCmd.minimized);
			if (!ok)
				delete newMainWnd;
		}
	}

	runStatus = running;
	return true;
}

static void noinline _hard_shutdown(int exitStatus)
{
	printDebug(QString("_hard_shutdown(")
		   + QString::number(exitStatus) + ")");
	delete pwminit;
	::exit(exitStatus);
}

void PwMInit::shutdownApp(int exitStatus, bool forceHardShutdown)
{
	printDebug(QString("PwMInit::shutdownApp(")
		   + QString::number(exitStatus)
		   + (forceHardShutdown ? ", true" : ", false")
		   + ") called.");

	PWM_ASSERT(runStatus == running ||
		   runStatus == shutdown ||
		   runStatus == init);
	// already shutting down?
	if (runStatus == shutdown) {
		if (forceHardShutdown)
			goto do_hard_shutdown;
		return;
	}
	runStatus = shutdown;

	lib()->setParentWidget(0);

	{
		// close all open mainwnds
		QPtrList<MainWindow> listCopy(*MainWindow::getList());
		QPtrList<MainWindow>::iterator i = listCopy.begin(),
					       end = listCopy.end();
		for ( ; i != end; ++i) {
/*FIXME still need this signal?
			disconnect(*i, SIGNAL(closed(PwM *)),
				   this, SLOT(mainWndClosed(PwM *)));
*/
			delete *i;
		}
	}
	{
		// close all remaining open documents
		QPtrList<PwMDoc> *dl = PwMDoc::getList();
		QPtrList<PwMDoc>::iterator i = dl->begin(),
					   end = dl->end();
		for ( ; i != end; ++i)
			(*i)->forceZeroUsageCount();
	}

	if (forceHardShutdown)
		goto do_hard_shutdown;
	else {
		bool inEventLoop = (app->eventLoop()->loopLevel() >= 1);
		if (inEventLoop)
			goto do_qt_shutdown;
		else {
			/* As we aren't in the event loop,
			 * QApplication::exit() doesn't work, yet.
			 * Just shut down hard.
			 */
			goto do_hard_shutdown;
		}
	}
	PWM_ASSERT(false);
do_qt_shutdown:
	printDebug("calling QApplication::exit()");
	// "this" is deleted in main() after we exited a->exec()
	QApplication::exit(exitStatus);
	return;
do_hard_shutdown:
	_hard_shutdown(exitStatus);
	PWM_ASSERT(false);
}

void PwMInit::initPosixSignalHandler()
{
	struct sigaction act;
	act.sa_handler = sig_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGILL, &act, NULL);
}

void PwMInit::initDCOP()
{
	_dcopClient = app->dcopClient();
	_dcopClient->setNotifications(true);
}

void PwMInit::initKWalletEmu(bool forceDisable, bool forceReload)
{
#ifdef CONFIG_KWALLETIF
	if (!conf()->confGlobKwalletEmu() ||
	    forceDisable) {
		delete_ifnot_null(_kwalletEmu);
		return;
	}
	if (_kwalletEmu && forceReload)
		delete_and_null(_kwalletEmu);
	if (!_kwalletEmu) {
		_kwalletEmu = new KWalletEmu(this);
		if (!_kwalletEmu->initialize())
			delete_and_null(_kwalletEmu);
	}
#else // CONFIG_KWALLETIF
	PARAM_UNUSED(forceDisable);
	PARAM_UNUSED(forceReload);
#endif // CONFIG_KWALLETIF
}

void PwMInit::initTray()
{
	if (!conf()->confGlobTray()) {
		if (!_tray)
			return;
		_tray->hide();
		delete_and_null(_tray);
		return;
	}
	if (_tray)
		return;
	_tray = new PwMTray(this);
	connect(_tray, SIGNAL(quitSelected()),
		this, SLOT(removeTrayAndQuit()));
	connect(_tray, SIGNAL(closed(PwMTray *)),
		this, SLOT(trayIconClosed(PwMTray *)));
	KIconLoader icons;
	_tray->setPixmap(icons.loadIcon(PACKAGE_NAME, KIcon::Small));
	_tray->show();
	// connect the signals of all open documents.
//FIXME: move to PwMTray:
#if 0
	const vector<PwMDocList::listItem> *dl = PwMDoc::getOpenDocList()->getList();
	vector<PwMDocList::listItem>::const_iterator i = dl->begin(),
						     end = dl->end();
	while (i != end) {
		_tray->connectDocToTray((*i).doc);
		++i;
	}
#endif
}

void PwMInit::removeTrayAndQuit()
{
	PWM_ASSERT(_tray);
	// _tray is deleted in ~PwMInit
	shutdownApp();
}

//TODO: remove
#if 0
PwM * PwMInit::createMainWnd(const QString &loadFile,
			     bool virginity,
			     PwMDoc *doc,
			     bool minimized)
{
	PwM *newWnd;
	if (!doc)
		doc = createDoc();
	newWnd = new PwM(this, doc, virginity);
	_mainWndList.push_back(newWnd);
	connect(newWnd, SIGNAL(closed(PwM *)),
		this, SLOT(mainWndClosed(PwM *)));
	connect(newWnd, SIGNAL(gotFocus(PwM *)),
		this, SLOT(setCurWidget(PwM *)));
	connect(newWnd, SIGNAL(lostFocus(PwM *)),
		this, SLOT(resetCurWidget()));
	if (minimized)
		newWnd->showMinimized();
	else
		newWnd->show();
	if (!loadFile.isEmpty())
		newWnd->openDoc(loadFile);
	return newWnd;
}

PwMDoc * PwMInit::createDoc()
{
	PwMDoc *doc = new PwMDoc(this);
	doc->setInit(this);
#ifdef CONFIG_KWALLETIF
	if (kwalletEmu())
		kwalletEmu()->connectDocSignals(doc);
#endif
	if (_tray)
		_tray->connectDocToTray(doc);
	return doc;
}
#endif

#if 0
void PwMInit::mainWndClosed(PwM *wnd)
{
FIXME();
//FIXME
#if 0
	bool doMinimizeToTray = false;
	bool doDeleteDoc = false;
	dcopClient()->suspend();
	dcopClient()->setAcceptCalls(false);
again:
	if (wnd->isForceMinimizeToTray()) {
		if (unlikely(!_tray)) {
			/* This should not happen! If we set forceMinimizeToTray ,
			 * we must be sure that _tray exists.
			 */
			BUG();
			wnd->setForceMinimizeToTray(false);
			goto again;
		}
		doMinimizeToTray = true;
	} else {
		// Ask to minimize to tray. If not, delete doc.
		if (_tray &&
		    runStatus != shutdown &&
		    !wnd->isForceQuit() &&
		    !wnd->curDoc()->isDeleted()) {
			if (conf()->confWndClose())
				doDeleteDoc = true;
			else
				doMinimizeToTray = true;
		} else {
			doDeleteDoc = true;
		}
	}
	if (doMinimizeToTray) {
		PWM_ASSERT(_tray);
		if (conf()->confGlobMinimizeLock())
			wnd->curDoc()->lockDocument();
	} else if (doDeleteDoc) {
		bool allowCancel = (runStatus != shutdown);
		if (!wnd->curDoc()->tryDelete(allowCancel)) {
			if (allowCancel) {
				/* We failed deleting the doc,
				 * so open a new window with it, again.
				 */
				createMainWnd(QString::null, false,
					      false, wnd->curDoc());
			} else {
				// no way out. force it!
				printDebug("delete doc forced in PwMInit::mainWndClosed()");
				delete wnd->curDoc();
			}
		}
	}
	// find the closed window in the "mainWndList" and delete it.
	QValueList<PwM *>::iterator i = _mainWndList.begin(),
				    end = _mainWndList.end();
	while (i != end) {
		if (*i == wnd) {
			_mainWndList.erase(i);
			goto out_success;
		}
		++i;
	}
	BUG();
out_success:
	if (!_mainWndList.size()) {
		/* If there's no main window and no tray icon
		 * left, we have no user interface, so we can
		 * shut down the application.
		 */
		if (!_tray) {
			dcopClient()->setAcceptCalls(true);
			dcopClient()->resume();
			shutdownApp(PWM_EXIT_SUCCESS);
			return;
		}
		/* There is no widget left, so set
		 * _curWidget to 0
		 */
		resetCurWidget();
	}
	dcopClient()->setAcceptCalls(true);
	dcopClient()->resume();
#endif
}
#endif

void PwMInit::trayIconClosed(PwMTray *tray)
{
	if (runStatus != running)
		return;
	PARAM_UNUSED(tray);
	PWM_ASSERT(tray == _tray);
	/* If there's no main wnd left we have to
	 * shutdown the app (same as in mainWndClosed())
	 */
	if (!MainWindow::getList()->count())
		shutdownApp(PWM_EXIT_SUCCESS);
}

void PwMInit::handleCmdLineArgs(bool /* initial */)
{
	KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
	PWM_ASSERT(args);
	int i, numArgs = args->count();
	const char *curArg;

	// read all cmdline options
	savedCmd.minimized = args->isSet("minimized");
	savedCmd.minToTray = args->isSet("mintray");
	if (savedCmd.minimized &&
	    savedCmd.minToTray) {
		printInfo(i18n("Commandline option \"--minimized\" and "
			       "\"--mintray\" selected. These are incompatible. "
			       "\"--mintray\" will be selected.").latin1());
	}
	if (savedCmd.minToTray && !_tray) {
		printInfo(i18n("No tray Icon available. "
			       "Commandline option \"--mintray\" is replaced "
			       "by \"--minimized\".").latin1());
	}
	/* Iterate through all non-option arguments.
	 * Every non-option arg is a filename to open.
	 */
	for (i = 0; i < numArgs; ++i) {
		curArg = args->arg(i);
		PWM_ASSERT(curArg);
		if (savedCmd.minToTray && _tray) {
			PwMDoc *newDoc = new PwMDoc(this);
			if (!newDoc->load(curArg))
				delete newDoc;
		} else {
			MainWindow *newMainWnd = new MainWindow;
			bool ok;
			ok = newMainWnd->initialize(0, QString(), savedCmd.minimized);
			if (!ok) {
//FIXME				newInstance->setForceQuit(true);
				delete newMainWnd;
			}
		}
	}

	if (savedCmd.minToTray) {
		minimizeAllMainWnd(true);
	} else if (savedCmd.minimized) {
		minimizeAllMainWnd(false);
	}
	args->clear();
}

void PwMInit::minimizeAllMainWnd(bool toTray)
{//FIXME
#if 0
	if (!_mainWndList.size())
		return;
	const QValueList<PwM *> *ml = mainWndList();
	QValueList<PwM *>::const_iterator it = ml->begin(),
					  end = ml->end();
	PwM *wnd;
	if (toTray && _tray) {
		/* minimize to tray.
		 * close all mainWnd.
		 */
		while (it != end) {
			wnd = *it;
			wnd->setForceMinimizeToTray(true);
			wnd->close_slot();
			++it;
		}
	} else {
		// normal minimize
		while (it != end) {
			wnd = *it;
			wnd->hide();
			wnd->showMinimized();
			++it;
		}
	}
#endif
}

#include "pwminit.moc"
