/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004 by Michael Buesch                                  *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This file contains code derived from                                  *
 *   pwsafe - commandline tool compatible with Counterpane's Passwordsafe  *
 *   Copyright (C) 2004 Nicolas S. Dade                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pwsafefile.h"
#include "pwmdoc.h"
#include "randomizer/randomizer.h"
#include "pwminit.h"

#include <qfile.h>
#include <qregexp.h>

#include <kmessagebox.h>
#include <klocale.h>

#ifdef BIG_ENDIAN_HOST
# include <algorithm>
#endif // BIG_ENDIAN_HOST


#define SPLIT_CHAR		'\xAD'
#define DEFAULT_USER_CHAR	'\xA0'


PwSafeFile::PwSafeFile()
 : file (0)
{
}

PwSafeFile::~PwSafeFile()
{
	PWM_ASSERT(file == 0);
}

bool PwSafeFile::importData(const QString &filepath,
			    const QString &pw,
			    PwMDoc *doc)
{
	bool ret = true;
	file = new QFile(filepath);
	if (!file->open(IO_ReadOnly)) {
		KMessageBox::error(pwminit->curWidget(),
				   i18n("Could not open file.\n"
					"Does the file exist?"),
				   i18n("Open Error"));
		ret = false;
		goto out_delete;
	}
	Header hdr;
	if (!readHeader(&hdr)) {
		KMessageBox::error(pwminit->curWidget(),
				   i18n("Could not read file."),
				   i18n("Read Error"));
		ret = false;
		goto out_close;
	}
	if (!testPw(hdr, pw)) {
		KMessageBox::error(pwminit->curWidget(),
				   i18n("Wrong password.\n"
					"Import failed."),
				   i18n("Wrong Password"));
		ret = false;
		goto out_close;
	}
	if (!importEntries(doc, hdr, pw)) {
		KMessageBox::error(pwminit->curWidget(),
				   i18n("Could not import data.\n"
					"Is the file corrupt?"),
				   i18n("Import Error"));
		ret = false;
		goto out_close;
	}
out_close:
	file->close();
out_delete:
	delete_and_null(file);
	return ret;
}

bool PwSafeFile::exportData(const QString &filepath,
			    const QString &pw,
			    PwMDoc *doc)
{
	bool ret = true;
	file = new QFile(filepath);
	if (!file->open(IO_ReadWrite)) {
		KMessageBox::error(pwminit->curWidget(),
				   i18n("Could not open file for "
					"writing.\n"
					"Do you have enough permissions?"),
				   i18n("Open Error"));
		ret = false;
		goto out_delete;
	}
	Header hdr;
	genNewHeader(&hdr, pw);
	if (!writeHeader(hdr)) {
		KMessageBox::error(pwminit->curWidget(),
				   i18n("Could not write file."),
				   i18n("Write Error"));
		ret = false;
		goto out_close;
	}
	if (!exportEntries(doc, hdr, pw)) {
		KMessageBox::error(pwminit->curWidget(),
				   i18n("Could not write file."),
				   i18n("Write Error"));
		ret = false;
		goto out_close;
	}
out_close:
	file->close();
out_delete:
	delete_and_null(file);
	return ret;
}

bool PwSafeFile::readHeader(Header *hdr)
{
	if (file->readBlock(reinterpret_cast<char *>(hdr->random), array_size(hdr->random))
	    != array_size(hdr->random))
		return false;
	if (file->readBlock(reinterpret_cast<char *>(hdr->hash), array_size(hdr->hash))
	    != array_size(hdr->hash))
		return false;
	if (file->readBlock(reinterpret_cast<char *>(hdr->salt), array_size(hdr->salt))
	    != array_size(hdr->salt))
		return false;
	if (file->readBlock(reinterpret_cast<char *>(hdr->iv), array_size(hdr->iv))
	    != array_size(hdr->iv))
		return false;
	return true;
}

bool PwSafeFile::writeHeader(const Header &hdr)
{
	if (file->writeBlock(reinterpret_cast<const char *>(hdr.random), array_size(hdr.random))
	    != array_size(hdr.random))
		return false;
	if (file->writeBlock(reinterpret_cast<const char *>(hdr.hash), array_size(hdr.hash))
	    != array_size(hdr.hash))
		return false;
	if (file->writeBlock(reinterpret_cast<const char *>(hdr.salt), array_size(hdr.salt))
	    != array_size(hdr.salt))
		return false;
	if (file->writeBlock(reinterpret_cast<const char *>(hdr.iv), array_size(hdr.iv))
	    != array_size(hdr.iv))
		return false;
	return true;
}

void PwSafeFile::genNewHeader(Header *hdr,
			      const QString &pw)
{
	Randomizer rnd;
	rnd.nextSize(array_size(hdr->random));
	rnd >> hdr->random;
	rnd.nextSize(array_size(hdr->salt));
	rnd >> hdr->salt;
	rnd.nextSize(array_size(hdr->iv));
	rnd >> hdr->iv;
	hashPw(pw, hdr->random, hdr->hash);
}

bool PwSafeFile::testPw(const Header &hdr,
			const QString &pw)
{
	unsigned char hash[SHA1_HASH_LEN_BYTE];
	hashPw(pw, hdr.random, hash);
	if (memcmp(hdr.hash, hash, array_size(hash)))
		return false;
	return true;
}

void PwSafeFile::hashPw(const QString &pw,
			const unsigned char *headerRandom,
			unsigned char *hash)
{
	/* WARNING: This function is ugly!
	 *          It needs to be ugly, because we have to mimic
	 *          some ugly bugs, to be compatible.
	 */

	/* generate test hash from random and passphrase.
	 * I am mystified as to why Bruce uses these extra 2 zero bytes in the hashes.
	 */
	Sha1 sha;
	sha.sha1_write(headerRandom, 8);
	const byte zeros[2] = { 0, 0 };
	sha.sha1_write(zeros, 2);
	sha.sha1_write(reinterpret_cast<const byte *>(pw.latin1()),
		       pw.length());
	QByteArray temp_key;
	sha.sha1_read(&temp_key);

	byte tmp_random[8];
	memcpy(tmp_random, headerRandom, 8);
	makeBlockLE(reinterpret_cast<unsigned char *>(tmp_random));
	Blowfish bf;
	bf.bf_setkey(reinterpret_cast<const byte *>(temp_key.data()),
		     temp_key.size());
	uint32_t xl;
	uint32_t xr;
	memcpy(&xl, tmp_random, 4);
	memcpy(&xr, tmp_random + 4, 4);
	/* to mimic passwordsafe I use do_encrypt() directly,
	 * but that means I have to pretend that I am on a
	 * little-endian machine b/c passwordsafe assumes a i386.
	 */
	register unsigned long i;
	for (i = 0; i < 1000; ++i)
		bf.do_encrypt(&xl, &xr);
	memcpy(tmp_random, &xl, 4);
	memcpy(tmp_random + 4, &xr, 4);
	makeBlockLE(reinterpret_cast<unsigned char *>(tmp_random));

	/* Now comes a sad part: I have to hack to mimic the original passwordsafe which contains what I believe
	 * is a bug. passwordsafe used its own blowfish and sha1 libraries, and its version of SHA1Final()
	 * memset the sha context to 0's. However the passwordsafe code went ahead and performed a
	 * SHA1Update on that zero'ed context. This of course did not crash anything, but it is not
	 * a real sha hash b/c the initial state of a real sha1 is not all zeros. Also we end up only
	 * hashing 8 bytes of stuff, so there are not 20 bytes of randomness in the result.
	 * The good thing is we are hashing something which is already well hashed, so I doubt this
	 * opened up any holes. But it does show that one should always step the program in a debugger
	 * and watch what the variables are doing; sometimes it is eye opening!
	 */
	/* The context already has been 0'ed in sha1_read() */
	sha.sha1_write(tmp_random, array_size(tmp_random));
	sha.sha1_write(zeros, 2);
	sha.sha1_read(&temp_key);
	PWM_ASSERT(temp_key.size() == SHA1_HASH_LEN_BYTE);
	memcpy(hash, temp_key.data(), SHA1_HASH_LEN_BYTE);
}

#ifdef BIG_ENDIAN_HOST
void PwSafeFile::makeBlockLE(unsigned char *block)
{
	int i;
	// this is a big-endian system; put the 8 bytes data in little-endian order
	for (i = 0; i < 2; ++i) {
		std::swap(block[i], block[3 - i]);
		std::swap(block[4 + i], block[7 - i]);
	}
}
#endif // BIG_ENDIAN_HOST

bool PwSafeFile::importEntries(PwMDoc *doc,
			       const Header &hdr,
			       const QString &pw)
{TODO();//TODO
#if 0
	unsigned char ctx[8];
	/* calculate a hash from passphrase and salt.
	 * Set this hash as key in blowfish.
	 */
	Sha1 sha;
	sha.sha1_write(reinterpret_cast<const byte *>(pw.latin1()),
		       pw.length());
	sha.sha1_write(hdr.salt, array_size(hdr.salt));
	QByteArray temp_key;
	sha.sha1_read(&temp_key);

	Blowfish bf;
	bf.bf_setkey(reinterpret_cast<const byte *>(temp_key.data()),
		     temp_key.size());
	memcpy(ctx, hdr.iv, array_size(ctx));
	makeBlockLE(ctx);

	PwMDataItem d;
	while (!file->atEnd()) {
		// now read the entries
		QString name_login;
		if (!readEntry(bf, ctx, &name_login))
			return false;
		QString password;
		if (!readEntry(bf, ctx, &password))
			return false;
		QString notes;
		if (!readEntry(bf, ctx, &notes))
			return false;

		QString name, login;
		// split name_login if it contains the magic split char
		int p = name_login.find(SPLIT_CHAR);
		if (p > 0) {
			name = name_login.left(p);
			login = name_login.right(name_login.length() - p - 1);
		} else {
			p = name_login.find(DEFAULT_USER_CHAR);
			if (p > 0) {
				/* this entry uses the default login.
				 * this is not part of the database;
				 * instead it is part of the configuration,
				 * or in our case, $USER
				 */
				// just skip this.
				continue;
			} else {
				/* no magic split chars; assume this is
				 * a very old database that contains no
				 * login field.
				 */
				name = name_login;
			}
		}
		// and trim any extra whitespace from name and login
		name.remove(QRegExp(" +$"));
		login.remove(QRegExp("^ +"));
		// now add this entry to the document.
		d.clear();
		d.desc = name.latin1();
		d.name = login.latin1();
		d.pw = password.latin1();
		d.comment = notes.latin1();
		doc->addEntry(DEFAULT_CATEGORY,
			      &d, true);
	}
	doc->flagDirty();
	return true;
#endif
return false;
}

bool PwSafeFile::readEntry(Blowfish &bf,
			   unsigned char *ctx,
			   QString *ret)
{
	if (file->atEnd())
		return false;
	unsigned char tmpBlock[8];
	unsigned char tmpBlockCopy[array_size(tmpBlock)];
	uint32_t xl;
	uint32_t xr;
	if (file->readBlock(reinterpret_cast<char *>(tmpBlock), array_size(tmpBlock))
	    != array_size(tmpBlock))
		return false;
	makeBlockLE(tmpBlock);
	memcpy(tmpBlockCopy, tmpBlock, array_size(tmpBlockCopy));

	memcpy(&xl, tmpBlock, 4);
	memcpy(&xr, tmpBlock + 4, 4);
	bf.do_decrypt(&xl, &xr);
	memcpy(tmpBlock, &xl, 4);
	memcpy(tmpBlock + 4, &xr, 4);

	int i, j;
	for (i = 0; i < 8; ++i)
		*(tmpBlock + i) ^= *(ctx + i);
	memcpy(ctx, tmpBlockCopy, array_size(tmpBlockCopy));

	int32_t len;
	// because we are always byte-ordered correctly, we can just do this
	memcpy(&len, tmpBlock, 4);
	memset(tmpBlock, 0, array_size(tmpBlock));
	memset(tmpBlockCopy, 0, array_size(tmpBlockCopy));

	if (len < 0) {
		// invalid file format?
		return false;
	}
	// make sure len isn't completely nuts (we could also compare with remaining file length...)
	if (len > 64 * 1024)
		return false;

	int numblocks = (len + 8 - 1) / 8;
	if (numblocks == 0)
		numblocks = 1;

	*ret = QString::null;
	ret->fill(0, len);
	unsigned char data[array_size(tmpBlock)];
	for (i = 0; i < numblocks; ++i) {
		if (file->readBlock(reinterpret_cast<char *>(tmpBlock), array_size(tmpBlock))
		    != array_size(tmpBlock))
			return false;
		makeBlockLE(tmpBlock);
		memcpy(tmpBlockCopy, tmpBlock, array_size(tmpBlockCopy));
		
		memcpy(&xl, tmpBlock, 4);
		memcpy(&xr, tmpBlock + 4, 4);
		bf.do_decrypt(&xl, &xr);
		memcpy(tmpBlock, &xl, 4);
		memcpy(tmpBlock + 4, &xr, 4);

		for (j = 0; j < 8; ++j)
			*(tmpBlock + j) ^= *(ctx + j);

		memcpy(data, tmpBlock, array_size(data));
		makeBlockLE(data);
		for (j = 0; j < 8 && i * 8 + j < len; ++j)
			(*ret)[i * 8 + j] = data[j];

		memcpy(ctx, tmpBlockCopy, array_size(tmpBlockCopy));
	}
	return true;
}

bool PwSafeFile::exportEntries(PwMDoc *doc,
			       const Header &hdr,
			       const QString &pw)
{TODO();//TODO
#if 0
	unsigned char ctx[8];
	memset(ctx, 0, array_size(ctx));
	/* calculate a hash from passphrase and salt.
	 * Set this hash as key in blowfish.
	 */
	Sha1 sha;
	sha.sha1_write(reinterpret_cast<const byte *>(pw.latin1()),
		       pw.length());
	sha.sha1_write(hdr.salt, array_size(hdr.salt));
	QByteArray temp_key;
	sha.sha1_read(&temp_key);

	Blowfish bf;
	bf.bf_setkey(reinterpret_cast<const byte *>(temp_key.data()),
		     temp_key.size());
	memcpy(ctx, hdr.iv, array_size(ctx));
	makeBlockLE(ctx);

	unsigned int numCat = doc->numCategories();
	unsigned int numEntr;
	unsigned int i, j;
	PwMDataItem d;
	QString name_login, password, notes;

	for (i = 0; i < numCat; ++i) {
		numEntr = doc->numEntries(i);
		for (j = 0; j < numEntr; ++j) {
			doc->getEntry(i, j, &d);
			name_login = QString(d.desc.c_str()) + SPLIT_CHAR + d.name.c_str();
			password = d.pw.c_str();
			notes = d.comment.c_str();
			if (!writeEntry(bf, ctx, name_login))
				return false;
			if (!writeEntry(bf, ctx, password))
				return false;
			if (!writeEntry(bf, ctx, notes))
				return false;
		}
	}
	return true;
#endif
return false;
}

bool PwSafeFile::writeEntry(Blowfish &bf,
			    unsigned char *ctx,
			    const QString &e)
{
//TODO
	return true;
}
