/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "rawfile/rawfile.h"
#include "pwmexception.h"
#include "globalstuff.h"

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>


using namespace PwManager;


RawFile::RawFile()
{
}

RawFile::~RawFile()
{
	PWM_ASSERT(!f.isOpen());
	close();
}

QString RawFile::getDefaultDocPath()
{
	static QString path;
	const char *env_home;

	if (!path.isEmpty()) {
		// the path is already cached.
		return path;
	}
	env_home = getenv("HOME");
	if (!env_home) {
		printWarn("Environment Variable $HOME not set!");
		// hm, fallback.
		env_home = ".";
	}
	path = env_home;
	path += "/.pwmanager";
	if (mkdir(path.ascii(), S_IRWXU)) {
		if (errno != EEXIST) {
			printWarn(QString("Could not create directory \"")
				  + path.ascii()
				  + "\"");
			path = QString();
			goto out;
		}
	}
	path += "/documents";
	if (mkdir(path.ascii(), S_IRWXU)) {
		if (errno != EEXIST) {
			printWarn(QString("Could not create directory \"")
				  + path.ascii()
				  + "\"");
			path = QString();
			goto out;
		}
	}
	// ok, here it is. tada.
out:
	return path;
}

QString RawFile::makePath(const QString &document) const
{
	if (document.isEmpty())
		return QString();
	if (document[0] == '/')
		return document;
	QString ret(getDefaultDocPath());
	ret += '/';
	ret += document;
	return ret;
}

bool RawFile::sync()
{
	PWM_ASSERT(!f.isOpen());
	if (!isValidHeader(header))
		return false;
	if (!isValidData(data))
		return false;
	if (!write())
		return false;
	return true;
}

bool RawFile::read()
{
	PWM_ASSERT(f.isOpen());
	if (!readHeader())
		return false;
	if (!readData())
		return false;
	return true;
}

bool RawFile::readHeader()
{
	int ret;
	unsigned int hashLenByte;
	char tmpBuf[128]; // 128 should be enough.
	PWM_ASSERT(!header.valid);

	// read the "magic" header
	PWM_ASSERT(strlen(PWM_FILE_MAGIC) == PWM_FILE_MAGIC_LENGTH);
	BUG_ON(PWM_FILE_MAGIC_LENGTH > array_size(tmpBuf));
	if (f.readBlock(tmpBuf, PWM_FILE_MAGIC_LENGTH) != PWM_FILE_MAGIC_LENGTH)
		return false;
	if (memcmp(tmpBuf, PWM_FILE_MAGIC, PWM_FILE_MAGIC_LENGTH))
		return false;
	header.magic.duplicate(tmpBuf, PWM_FILE_MAGIC_LENGTH);

	// read/check version number
	ret = f.getch();
	if (ret == -1)
		return false;
	if (ret != PWM_FILE_VER_MAJOR)
		return false;
	header.verMajor = ret;
	ret = f.getch();
	if (ret == -1)
		return false;
	header.verMinor = ret;

	// read key hash id
	ret = f.getch();
	if (ret == -1)
		return false;
	header.keyHashId = ret;
	if (!validHashId(header.keyHashId))
		return false;

	// read data hash id
	ret = f.getch();
	if (ret == -1)
		return false;
	header.dataHashId = ret;
	if (!validHashId(header.dataHashId))
		return false;

	// read crypt algo id
	ret = f.getch();
	if (ret == -1)
		return false;
	header.cryptAlgoId = ret;
	if (!validCryptId(header.cryptAlgoId))
		return false;

	// read compression algo id
	ret = f.getch();
	if (ret == -1)
		return false;
	header.comprAlgoId = ret;
	if (!validComprId(header.comprAlgoId))
		return false;

	// read the master key type
	ret = f.getch();
	if (ret == -1)
		return false;
	header.keyType = ret;

	// read the "RESERVED" bytes
	BUG_ON(PWM_FILE_RESERVED_BYTES > array_size(tmpBuf));
	if (f.readBlock(tmpBuf, PWM_FILE_RESERVED_BYTES) != PWM_FILE_RESERVED_BYTES)
		return false;
	header.reserved.duplicate(tmpBuf, PWM_FILE_RESERVED_BYTES);

	// read the key hash
	hashLenByte = hashToLength(header.keyHashId) / 8;
	BUG_ON(hashLenByte > array_size(tmpBuf));
	if (f.readBlock(tmpBuf, hashLenByte) != static_cast<Q_LONG>(hashLenByte))
		return false;
	header.keyHash.duplicate(tmpBuf, hashLenByte);

	// read the data hash
	hashLenByte = hashToLength(header.dataHashId) / 8;
	BUG_ON(hashLenByte > array_size(tmpBuf));
	if (f.readBlock(tmpBuf, hashLenByte) != static_cast<Q_LONG>(hashLenByte))
		return false;
	header.dataHash.duplicate(tmpBuf, hashLenByte);

	header.valid = 1;
	return true;
}

bool RawFile::readData()
{
	PWM_ASSERT(f.isOpen());
	PWM_ASSERT(!data.valid);
	data.data = f.readAll();
	if (data.data.isEmpty())
		return false;
	data.valid = 1;
	return true;
}

bool RawFile::write()
{
	bool ret = true;

	PWM_ASSERT(!f.isOpen());
	/* Move the existing file to some tmp file.
	 * When saving file succeeds, delete tmp file. Otherwise
	 * move tmp file back. See below.
	 */
	QString oldName(f.name());
	QString newName;
	unsigned int cnt = 0;
	do {
		newName = oldName + "_save_" + QString::number(cnt++);
		if (cnt > 1000) {
			ret = false;
			goto out;
		}
	} while (rename(oldName.ascii(), newName.ascii()));
	if (!f.open(IO_ReadWrite)) {
		ret = false;
		goto out_moveback;
	}
	if (!writeHeader()) {
		ret = false;
		goto out_close_move;
	}
	if (!writeData()) {
		ret = false;
		goto out_close_move;
	}
	if (unlink(newName.ascii()))
		WARN();
	f.close();
	goto out;

out_close_move:
	f.close();
out_moveback:
	if (unlink(oldName.ascii()))
		WARN();
	if (rename(newName.ascii(), oldName.ascii()))
		WARN();
out:
	return ret;
}

bool RawFile::writeHeader()
{
	PWM_ASSERT(header.valid);
	PWM_ASSERT(f.at() == 0);
	if (f.writeBlock(header.magic) != static_cast<Q_LONG>(header.magic.size()))
		return false;
	if (f.putch(header.verMajor) != header.verMajor)
		return false;
	if (f.putch(header.verMinor) != header.verMinor)
		return false;
	if (f.putch(header.keyHashId) != header.keyHashId)
		return false;
	if (f.putch(header.dataHashId) != header.dataHashId)
		return false;
	if (f.putch(header.cryptAlgoId) != header.cryptAlgoId)
		return false;
	if (f.putch(header.comprAlgoId) != header.comprAlgoId)
		return false;
	if (f.putch(header.keyType) != header.keyType)
		return false;
	if (f.writeBlock(header.reserved) != static_cast<Q_LONG>(header.reserved.size()))
		return false;
	if (f.writeBlock(header.keyHash) != static_cast<Q_LONG>(header.keyHash.size()))
		return false;
	if (f.writeBlock(header.dataHash) != static_cast<Q_LONG>(header.dataHash.size()))
		return false;
	return true;
}

bool RawFile::writeData()
{
	PWM_ASSERT(f.isOpen());
	PWM_ASSERT(data.valid);
	if (f.writeBlock(data.data) != static_cast<Q_LONG>(data.data.size()))
		return false;
	return true;
}

int RawFile::open(const QString &file)
{
	int ret = 0;
	bool isEmpty;

	close();
	QString filePath(makePath(file));
	f.setName(filePath);
	if (!f.open(IO_ReadWrite)) {
		ret = -1;
		goto out;
	}
	isEmpty = (f.size() == 0);
	if (isEmpty) {
		// file does not exist, yet. Create a new one.
		printDebug("new file");
		header.clear();
		data.clear();
	} else {
		// file exists. Read it.
		printDebug("read file");
		if (!read()) {
			ret = -2;
			goto out_close;
		}
	}
out_close:
	f.close();
out:
	return ret;
}

void RawFile::close()
{
	PWM_ASSERT(!f.isOpen());
	/* We do NOT sync here.
	 * The user has to sync by himself, before he
	 * closes this.
	 */
	header.clear();
	data.clear();
}

bool RawFile::isValidHeader(const FileHeader_t &hdr)
{
	if (!hdr.valid)
		return false;
	if (qstrncmp(hdr.magic, PWM_FILE_MAGIC, hdr.magic.size()))
		return false;
	if (hdr.verMajor != PWM_FILE_VER_MAJOR)
		return false;
	if (!validHashId(hdr.keyHashId))
		return false;
	if (!validHashId(hdr.dataHashId))
		return false;
	if (!validCryptId(hdr.cryptAlgoId))
		return false;
	if (!validComprId(hdr.comprAlgoId))
		return false;
	if (hdr.reserved.size() != PWM_FILE_RESERVED_BYTES)
		return false;
	if (hdr.keyHash.size() != (hashToLength(hdr.keyHashId) / 8))
		return false;
	if (hdr.dataHash.size() != (hashToLength(hdr.dataHashId) / 8))
		return false;
	return true;
}

bool RawFile::isValidData(const FileData_t &dta)
{
	if (!dta.valid)
		return false;
	if (dta.data.isEmpty())
		return false;
	return true;
}
