/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004, 2005  by Michael Buesch                           *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#ifndef PWMANAGER_RAWFILE_H_
#define PWMANAGER_RAWFILE_H_

#include "pwmexception.h"

#include <qcstring.h>
#include <qfile.h>


#define PWM_FILE_MAGIC  	"PWM_PASSWORD_FILE"
#define PWM_FILE_MAGIC_LENGTH	17 /* byte */
#define PWM_FILE_VER_MAJOR	(static_cast<unsigned char>(0x06))
#define PWM_FILE_VER_MINOR	(static_cast<unsigned char>(0x00))
#define PWM_FILE_RESERVED_BYTES	64


#define PWM_HASH_SHA1		(static_cast<unsigned char>(0x01))
#define PWM_HASH_SHA256		(static_cast<unsigned char>(0x02))
#define PWM_HASH_SHA384		(static_cast<unsigned char>(0x03))
#define PWM_HASH_SHA512		(static_cast<unsigned char>(0x04))
#define PWM_HASH_MD5		(static_cast<unsigned char>(0x05))
#define PWM_HASH_RMD160		(static_cast<unsigned char>(0x06))
#define PWM_HASH_TIGER		(static_cast<unsigned char>(0x07))

#define PWM_HASH_MIN		PWM_HASH_SHA1
#define PWM_HASH_MAX		PWM_HASH_TIGER


#define PWM_CRYPT_BLOWFISH	(static_cast<unsigned char>(0x01))
#define PWM_CRYPT_AES128	(static_cast<unsigned char>(0x02))
#define PWM_CRYPT_AES192	(static_cast<unsigned char>(0x03))
#define PWM_CRYPT_AES256	(static_cast<unsigned char>(0x04))
#define PWM_CRYPT_3DES		(static_cast<unsigned char>(0x05))
#define PWM_CRYPT_TWOFISH	(static_cast<unsigned char>(0x06))
#define PWM_CRYPT_TWOFISH128	(static_cast<unsigned char>(0x07))

#define PWM_CRYPT_MIN		PWM_CRYPT_BLOWFISH
#define PWM_CRYPT_MAX		PWM_CRYPT_TWOFISH128


#define PWM_COMPRESS_NONE	(static_cast<unsigned char>(0x00))
#define PWM_COMPRESS_GZIP	(static_cast<unsigned char>(0x01))
#define PWM_COMPRESS_BZIP2	(static_cast<unsigned char>(0x02))

#define PWM_COMPRESS_MIN	PWM_COMPRESS_NONE
#define PWM_COMPRESS_MAX	PWM_COMPRESS_BZIP2


namespace PwManager
{

class FileHeader_t
{
public:
	FileHeader_t()
	 : valid (0)
		{ }
	/** clear all data */
	void clear()
		{
			valid = 0;
			magic.resize(0);
			verMajor = 0;
			verMinor = 0;
			keyHashId = 0;
			dataHashId = 0;
			cryptAlgoId = 0;
			comprAlgoId = 0;
			keyType = 0;
			reserved.resize(0);
			keyHash.resize(0);
			dataHash.resize(0);
		}
	/** does this class contain valid data?
	  * 0 == invalid
	  * 1 == valid
	  */
	unsigned char valid;
public:
	/** Magic string */
	QByteArray magic;
	/** FILEVERSION */
	unsigned char verMajor;
	unsigned char verMinor;
	/** KEY-HASH-ALGORITHM-ID */
	unsigned char keyHashId;
	/** DATA-HASH-ALGORITHM-ID */
	unsigned char dataHashId;
	/** CRYPT-ALGORITHM-ID */
	unsigned char cryptAlgoId;
	/** COMPRESSION-ALGORITHM-ID */
	unsigned char comprAlgoId;
	/** MASTER-KEY-TYPE-FLAG */
	unsigned char keyType;
	/** RESERVED BYTES FOR FUTURE USE */
	QByteArray reserved;
	/** KEY-HASH */
	QByteArray keyHash;
	/** DATA-HASH */
	QByteArray dataHash;
};

class FileData_t
{
public:
	FileData_t()
	 : valid (0)
		{ }
	/** clear all data */
	void clear()
		{
			valid = 0;
			data.resize(0);
		}
	/** does this class contain valid data?
	  * 0 == invalid
	  * 1 == valid
	  */
	unsigned char valid;
public:
	/** ENCRYPTED DATA BODY */
	QByteArray data;
};

} // namespace PwManager


class DaemonInit;

/** This is the representation of a PWM_PASSWORD_FILE in memory */
class RawFile
{
public:
	RawFile();
	~RawFile();

	/** check if the hash id is in the allowed range. */
	static inline
	bool validHashId(unsigned char hash)
		{
			if (hash >= PWM_HASH_MIN &&
			    hash <= PWM_HASH_MAX)
				return true;
			return false;
		}
	/** check if the crypt id is in the allowed range. */
	static inline
	bool validCryptId(unsigned char crypt)
		{
			if (crypt >= PWM_CRYPT_MIN &&
			    crypt <= PWM_CRYPT_MAX)
				return true;
			return false;
		}
	/** check if the compression algo id is in the allowed range. */
	static inline
	bool validComprId(unsigned char compr)
		{
			/* The "plus" shuts up this warning:
			 * "comparison is always true due to limited range of data type"
			 */
			const unsigned char plus = (PWM_COMPRESS_MIN == 0) ? 1 : 0;
			if (compr + plus >= PWM_COMPRESS_MIN + plus &&
			    compr <= PWM_COMPRESS_MAX)
				return true;
			return false;
		}
	/** map the hash id to the hash length in bit */
	static inline
	unsigned int hashToLength(unsigned char hash)
		{
			const unsigned int map[] = {
				160,	// SHA1
				256,	// SHA256
				384,	// SHA384
				512,	// SHA512
				128,	// MD5
				160,	// RMD160
				192	// TIGER
			};
			PWM_ASSERT(validHashId(hash));
			PWM_ASSERT(hash - PWM_HASH_MIN <= static_cast<unsigned char>(array_size(map)) - 1);
			return map[hash - PWM_HASH_MIN];
		}
	/** check if the given header data is valid. */
	static bool isValidHeader(const PwManager::FileHeader_t &hdr);
	/** check if the given file data is valid. */
	static bool isValidData(const PwManager::FileData_t &dta);
	/** get the default document path. */
	static QString getDefaultDocPath();

	/** check if this RawFile is for the document. */
	bool operator==(const QString &document) const
		{
			return (makePath(document) == filename());
		}

	/** open a new or existing file */
	int open(const QString &file);
	/** close this file */
	void close();
	/** returns the filename or an empty string if not open. */
	QString filename() const
			{ return f.name(); }
	/** get a pointer to the file header. */
	PwManager::FileHeader_t * getHeader()
			{ return &header; }
	/** get a pointer to the file data. */
	PwManager::FileData_t * getData()
			{ return &data; }
	/** synchronize the data in memory with the disk */
	bool sync();

protected:
	/** creates a path from the document name. */
	QString makePath(const QString &document) const;
	/** read the data from disk into memory */
	bool read();
	/** read the header from disk into memory */
	bool readHeader();
	/** read the data from disk into memory */
	bool readData();
	/** write the data from memory to disk */
	bool write();
	/** write the header from memory to disk */
	bool writeHeader();
	/** write the data from memory to disk */
	bool writeData();

protected:
	/** file access */
	QFile f;
	/** header for this file */
	PwManager::FileHeader_t header;
	/** data for this file */
	PwManager::FileData_t data;
};

#endif // PWMANAGER_RAWFILE_H_
